/***************************************************************************
                           application.hpp
                           -------------------
    copyright            : (C) 2001,2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: application.hpp,v 1.12 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef APPLICATION_HPP
#define APPLICATION_HPP

#include <string>
#include <map>
#include <fstream>
extern "C" {
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
}

#include "cmdlinekey.hpp"
#include "config.h"

/*
 * concerning signal handling:
 * ^^^^^^^^^^^^^^^^^^^^^^^^^^
 * i could not find any other concept than using c functions
 * and in them call the appropriate Application memberfunction.
 * if there is any better solution please send it to the ml.
 */

// call this functions only if you know what you are doing
void sigIntHandler( int signum );
void sigHupHandler( int signum );
void sigQuitHandler( int signum );
void sigAbrtHandler( int signum );
void sigSegvHandler( int signum );


class Application 
{
public:
  typedef char**   arg_type;
  typedef int      int_type;
  typedef int_type signal_type;
  typedef pid_t    pid_type;
  typedef std::map<const CMDLineKey, std::string> ConfigMap;

  typedef enum { APP_RUNNING, APP_IDLE } status_e;

  friend void sigIntHandler( int );
  friend void sigHupHandler( int );
  friend void sigQuitHandler( int );
  friend void sigAbrtHandler( int );
  friend void sigSegvHandler( int );

  static const int_type    SUCCESS;
  static const int_type    ERROR;
  static const int_type    IDLE; // nothing affected by func call
  static const signal_type SIG_HUP  = SIGHUP;
  static const signal_type SIG_INT  = SIGINT;
  static const signal_type SIG_QUIT = SIGQUIT;
  static const signal_type SIG_ABRT = SIGABRT;
  static const signal_type SIG_SEGV = SIGSEGV;

  Application( const std::string& name, arg_type args, int_type argNum );

  /**
   * calls 'stop'
   */
  virtual ~Application();

  /*
   * might return NULL when no application has been created before
   */
  static Application* getLastApplication();

  int start();
  /**
   * @return fork successull: Application::SUCCESS, else Application::ERROR
   */
  int fork();

  /**
   * @return stop successfull / application was running: Application::SUCCESS,
   * Application:ERROR else
   */
  int  stop();
  /**
   * will emit an signal if this signal is one of the defined signals above
   * this function will only emit a signal if the application is running
   *
   * @return Application::ERROR if the application is not running or the signal is unknown
   *         Application::SUCCESS if the signal has been emitted successfully
   */
  virtual int kill( signal_type signum = SIG_INT );

  bool isStarted() const
  { return m_status == APP_RUNNING; }
  bool isStopped() const
  { return m_status == APP_IDLE; }
  bool isIdle() const
  { return isStopped(); }
  
  const std::string& getName() const
  { return m_name; }
  const arg_type getArguments() const
  { return m_args; }
  const int_type getArgumentNum() const
  { return m_argNum; }

  pid_type getPID() const
  { return getpid(); }

  status_e getStatus() const
  { return m_status; }

protected:
  const std::string m_name;
  const arg_type    m_args;
  const int_type    m_argNum;

private:
  status_e   m_status;
  ConfigMap  m_argMap;
  
protected:
  /**
   * will 'parse' the argument vector and copy results into 'm_argMap'
   */
  void fillConfigMap();
  
  ConfigMap& getArgMap()
  { return m_argMap; }

  /**
   * @return pidfile written
   */
  bool writePIDFile( const std::string& filename );

  /**
   * @return 0 - no pidfile found / not readable;
   *         != 0 pidfile found, read and content given back
   */
  pid_type readPIDFile( const std::string& filename );

  void erasePIDFile( const std::string& filename );

  void redirectConsoleOutputStreams( const char* stdoutFilename,
				     const char* stderrFilename );

  void restoreConsoleOutputStreams();

  // hooks  
  /**
   * called whenever 'start()' gets called and the application is
   * not running already
   */
  virtual void onStart()
  {}
  /**
   * called whenever 'stop()' gets called and the application is 
   * running ( and has not been stopped ).
   * the default behaviour is to terminate the process
   */
  virtual void onStop()
  { exit( 0 ); }

  /**
   * is a hook of function 'fork()'. 
   * This function gets called after an successfull fork() call by
   * the <b>mother</b>.
   * the default behaviour is to delete 'this'
   */
  virtual void onFork()
  { delete this; }

  /**
   * is a hoook of function 'fork()'
   * This function gets called after an successull fork() call
   * by the process <b>child</b>
   * the default behaviour is to do nothing
   */
  virtual void onForked()
  {}

  /**
   * calls c-function 'signal()'
   * @return successfull operation
   */
  void registerSignal( signal_type signum );
  void deregisterSignal( signal_type signum );

  virtual void onSigInt()
  { exit( 1 ); }
  virtual void onSigHup()
  { exit( 1 ); }
  virtual void onSigAbrt()
  { exit( 1 ); }
  virtual void onSigSegv()
  { exit( 1 ); }
  virtual void onSigQuit()
  { exit( 1 ); }

private:
  Application();
  Application( const Application& );
  const Application& operator=( const Application& );

  static bool parseArgKey( std::string& key, std::string& value );

  static const int IDX_STDOUT = 0;
  static const int IDX_STDERR = 1;

  std::ofstream m_redirectedStreams[2];
  /* gets initialized by the ctor */
  std::streambuf* m_origStreams[2];
};

#endif

