#ifndef CGI_HPP
#define CGI_HPP


#include <map>
#include <algorithm>
#include "httpstring.hpp"
#include "hostserv.hpp"
#include "config.h"

namespace httpd
{

/*
 * command object
 */
class CGIScriptHandler
{
   public:
      virtual ~CGIScriptHandler()
      {}

      virtual void execute( const http::HTTPString& key, const http::HTTPString& value ) = 0;
};

template<class C>
class MemberFunctionExecutor : public CGIScriptHandler
{
   public:
      typedef void (C::*function_type)();

      MemberFunctionExecutor( C* c, function_type func )
	 : m_c( c ),
	   m_func( func )
      {}

      virtual ~MemberFunctionExecutor()
      {}

      virtual void execute( const http::HTTPString& key, const http::HTTPString& value )
      { (m_c->*m_func)(); }

   private:
      MemberFunctionExecutor();
      MemberFunctionExecutor( const MemberFunctionExecutor& );

      C* const m_c;
      function_type const m_func;
};

/*
 * manages a collection of cgiscripthandler objects. 
 * unifies access to cgiscripthandler instances.
 */
class CGIScriptHandlerKit : public CGIScriptHandler
{
   private:
      typedef std::multimap<const http::HTTPString, CGIScriptHandler*> handlermap_type;
      typedef handlermap_type::iterator iterator_type;
      typedef pair<handlermap_type::key_type,handlermap_type::value_type> pair_type;


      handlermap_type m_handlers;
      iterator_type   m_begin;
      iterator_type   m_end;

   public:
      CGIScriptHandlerKit()
	 : m_handlers(),
	   m_begin(),
	   m_end()
      {}

      virtual ~CGIScriptHandlerKit()
      {}

      void registerScriptHandler( const http::HTTPString& handlerName, CGIScriptHandler* handler )
      { m_handlers.insert( make_pair( handlerName, handler ) ); }

	 /**
	  * @return handler with name `handlerName' found and removed
	  */
      bool removeScriptHandler( const http::HTTPString& handlerName, const CGIScriptHandler* handler )
      {
	 initIterators( handlerName );

	 for( ; m_begin != m_end && m_begin->second != handler; ++m_begin );

	 if( m_begin != m_end )
	 {
	    m_handlers.erase( m_begin );
	    return true;
	 }

	 return false;
      }

      /*
       * @return number of affected CGIScriptHandler objects
       */
      void execute( const http::HTTPString& key, const http::HTTPString& value )
      {
	 initIterators( key );

	 if( m_begin != m_end )
	 {
#ifdef DEBUG
	    cout << "CGIScriptHandlerKit::execute(): executing " << key << endl;
#endif

	    do {
	       m_begin->second->execute( key, value );
	       ++m_begin;
	    } while( m_begin != m_end );
	 }
	 else
	    onCommandNotFound( key, value );
      }

   protected:
      /** 
       * gets called by execute() whenever a command with a certain could not be found
       */
      virtual void onCommandNotFound( const http::HTTPString& key, const http::HTTPString& value )
      {}

   private:
      CGIScriptHandlerKit( const CGIScriptHandlerKit& );
      CGIScriptHandlerKit& operator=( const CGIScriptHandlerKit& );
      
      inline void initIterators( const http::HTTPString& key )
      {
	 m_begin = this->m_handlers.lower_bound( key );
	 m_end   = this->m_handlers.upper_bound( key );
      }

//       class ExecFuncOp
//       {
// 	 private:


// 	 public:
// 	    ExecFuncOp( const http::HTTPString& key, const http::HTTPString& value )
// 	       : m_key ( key ),
// 		 m_value( value )
// 	    {}
	    
// 	    void operator()( pair_type& p )
// 	    { 
// 	       CGIScriptHandler* handler = p.second;
// 	       handler->execute( m_key, m_value );
// 	    }

// 	 private:
// 	    const http::HTTPString& m_key;
// 	    const http::HTTPString& m_value;
//       };

//       class ExecFuncPredicate 
//       {
// 	 public:
// 	    ExecFuncPredicate( const CGIScriptHandler* handler )
// 	       : m_handler( handler )
// 	    {}

// 	    template< class pairType>
// 	    bool operator()( const pairType& p )
// 	    { return p.second == m_handler; }
	    
// 	 private:
// 	    const CGIScriptHandler* const m_handler;
//       };
};

} // EON httpd


// Forward declaration for POSTCGIMethodKit
class HostServ;


namespace httpd {

class POSTCGIScriptHandlerKit : public CGIScriptHandlerKit
{
   public:
   static const http::HTTPString DIRECTIVE_SERVERNAME;
   static const http::HTTPString DIRECTIVE_SERVERDESCRIPTION;
   static const http::HTTPString DIRECTIVE_REMOTEHOST;
   static const http::HTTPString DIRECTIVE_REMOTEPASSWORD;
   static const http::HTTPString DIRECTIVE_HOSTSERVNICKNAME;
   static const http::HTTPString DIRECTIVE_HOSTSERVUSERNAME;
   static const http::HTTPString DIRECTIVE_HOSTSERVADDRESS;
   static const http::HTTPString DIRECTIVE_HOSTSERVREALNAME;
   static const http::HTTPString DIRECTIVE_HOSTSERVMODES;
   static const http::HTTPString DIRECTIVE_HOSTSERVCHANNEL;
   static const http::HTTPString DIRECTIVE_CLIENTDEFAULTPERMISSIONS;
   static const http::HTTPString DIRECTIVE_VHOSTPUBLICPOLICY;
   static const http::HTTPString DIRECTIVE_HOSTSERVCONFIGFILENAME;
   static const http::HTTPString DIRECTIVE_VHOSTSCONFIGFILENAME;
   static const http::HTTPString DIRECTIVE_USERSCONFIGFILENAME;
   static const http::HTTPString DIRECTIVE_RELOAD;
   static const http::HTTPString DIRECTIVE_RECONNECT;
   static const http::HTTPString DIRECTIVE_SHUTDOWN;
   
   POSTCGIScriptHandlerKit( HostServ* hostserv );
   
   
   protected:
   
   /*
    * gets called by CGIScriptHandlerKit whenever no command named 'key' was found.
    */
   virtual void onCommandNotFound( const http::HTTPString& key, const http::HTTPString& value )
   { m_staticKit->execute( key, value ); }
   
   private:
   POSTCGIScriptHandlerKit( const POSTCGIScriptHandlerKit& );
   POSTCGIScriptHandlerKit& operator=( const POSTCGIScriptHandlerKit& );
   
   void initKit();
   
   static void initStaticKit();
   
   HostServ*        const      m_hostserv;
   static CGIScriptHandlerKit* m_staticKit;
   MemberFunctionExecutor<HostServ> m_reloadHandler;
   MemberFunctionExecutor<HostServ> m_reconnectHandler;
   MemberFunctionExecutor<HostServ> m_shutdownHandler;
};


// forward declaration for CGIScript
class Response;
   
class CGIScript 
{
   public:
      /*
       * a map of key - value associations
       * cgi 'arguments'
       */
      typedef std::map<const http::HTTPString, http::HTTPString> execenv_type;

      virtual ~CGIScript()
      {}

      virtual const Response* makeResponse( const execenv_type& env ) = 0;
};

// forward declarations for SimpleCGIScript
class Request;
class GETRequest;

/*
 * What this `script' does is to call the appropriate CGIScriptHandlers
 * for every submitted data segment and then call the GETMethod method with
 * an appropriate Request objects.
 */
class SimpleCGIScript : public CGIScript
{
   public:
      SimpleCGIScript( const Request* req, CGIScriptHandlerKit& kit );
      
      virtual ~SimpleCGIScript();

      virtual const Response* makeResponse( const execenv_type& env );

      const CGIScriptHandlerKit& getKit() const
      { return m_kit; }

      /*
       * caller is authorative for returned request instance
       */
      const Request* getRequest() const
      { return m_request; }

   private:
      SimpleCGIScript( const SimpleCGIScript& );
      SimpleCGIScript& operator=( const SimpleCGIScript& );

      const Request* const  m_request;
      CGIScriptHandlerKit&  m_kit;
      GETRequest*           m_getRequest;
};

} // eon httpd
#endif
