/***************************************************************************
                           filebuf.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: filebuf.hpp,v 1.13 2002/04/26 21:25:22 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FILEBUF_HPP
#define FILEBUF_HPP

#include <iostream>
#include "file.hpp"
extern "C" {
#include "assert.h"
}
#include "config.h"

/**
 * manages a file reference and a read cache
 * 
 * m_readPos marks the current character which would get read next
 * m_readMax points to the end of the read data
 */
class FileInBuf
{
public:
  typedef File::len_type       len_type;
  typedef File::data_type      data_type;
  typedef File::data_reference data_reference;
  typedef int                  int_type;


  /**
   * @param f a file from which data has to be read
   * @param buflen length of the cache
   */
  FileInBuf( File& f, len_type buflen );
  virtual ~FileInBuf();

  /**
   * reads max. <tt>len</tt> characters from
   * the intern buffer and never more than
   * available characters.
   *
   * If stateEOF == true nothing will be read.
   * returned will be File::eof
   *
   * This function will fill the buffer if
   * the buffer has been read completely.
   */
  len_type read( data_reference buf,
		 len_type len );

  len_type fill();
  
  /**
   * moves data around if the buffer is not full
   * makes space for new data to be read;
   */
  void flush();

  bool isEmpty()
  { return m_readMax == m_readBuffer; }
  bool isFull()
  { return m_readMax == m_readBufferEnd; }
  bool readCompletely()
  { return m_readPos >= m_readMax; }

protected:
  /**
   * fills buffer completly and sets flags
   * - called by 'read' if buffer is empty
   * - will not check flags!
   *
   * if there was an read error return 'File::eof'
   * flag good has to be updated then
   */
  int_type fillEmptyBuffer();

  /**
   * if there was an read error return 'File::eof'
   *
   * @param bufPos a memory adresse where data has to be written
   * @param numBytes the number of bytes to read maximal
   * @return the actual number of read bytes
   */
  virtual int_type underflow( data_reference bufPos, 
			      len_type numBytes  );

  data_reference getReadPos() const
  { return m_readPos; }
  data_reference getReadMax() const 
  { return m_readMax; }
  data_reference getRBufBase() const
  { return m_readBuffer; }
  len_type getRBufLen() const
  { return m_rBufLen; }
  File& getInputFD() const
  { return m_inFd; }
  void setRState( bool eofNotReceived )
  { m_rStateGood = eofNotReceived; }
  
  void setReadPos( const data_reference readPos )
  { m_readPos = readPos; }

private:
  File&                m_inFd;         // the file
  data_reference       m_readBuffer; // base pointer to the memory segment of the buffer
  const data_reference m_readBufferEnd; // => m_readBuffer+m_rBufLen
  data_reference       m_readPos;    // current read position
  data_reference       m_readMax;    // boundary. marks the end of read data in buffer
  const len_type       m_rBufLen;    // max. capacity of the buffer
  bool                 m_rStateGood; // -> EOF not received

public:
  /**
   * @return false if 'EOF' has been read
   */
  bool stateGood()
  { return m_rStateGood; };
  
  /**
   * @return true if 'EOF' has been read
   */
  bool stateEof()
  { return !m_rStateGood; };

  /**
   * call this function when File points to another location / 
   * changed it's state.
   */
  void resetFlags()
  { m_rStateGood = true; };
}; // EOF FileInBuf

/**
 * output buffer with explicit flushing via 'doFlush()'.
 * uses a <b>reference</b> to given File to do writing.
 *
 */
class FileOutBuf : public std::streambuf
{
public:
  typedef File::len_type       len_type;
  typedef File::data_type      data_type;
  typedef File::data_reference data_reference;
  typedef int                  int_type;
  typedef int_type             stl_int_type;
  typedef std::streamsize      streamsize;

  /** 
   * @param f file. will be used to write to
   * @param buflen the size of the buffer
   */
  FileOutBuf( File& f, len_type buflen );
  virtual ~FileOutBuf();

  /**
   * flush function has been overwritten.
   * to flush the stream explicitely this
   * function is introduced.
   *
   *@return - on error: File::eof
   *         - else number of written bytes
   */
  len_type doFlush();

  bool isFilled() const
  { return pbase() < pptr(); }

  bool isEmpty() const
  { return pbase() == pptr(); }

  bool isFull() const
  { return epptr() == pptr(); }
  
protected:
  virtual stl_int_type overflow( int_type );
  
  /**
   * delayed flushing
   */
  virtual streamsize xsputn( const char* buf,
			     streamsize num );

  /**
   * gets called by ~'std::endl()'.
   * - does nothing but returning 0
   * -> explicit flushing
   * @see #doFlush
   */
  virtual int sync();

  /**
   * gets called by doFlush()
   */
  virtual len_type flushBuffer();
  

private:
  File&          m_outFd;
  data_reference m_writeBuffer;
  len_type       m_writeBufLen;
};

#endif
