/***************************************************************************
                           hsclient.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: hsclient.hpp,v 1.24 2002/03/20 16:45:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _HSCLIENT_HPP
#define _HSCLIENT_HPP

class HSClient;
#include <vhost.hpp>
#include <ircclient.hpp>
#include <dereferencecomparator.hpp>
#include <config.h>

using namespace std;

class HSClient 
{
public:
  typedef unsigned int flag_type;

  static const flag_type flagAdmin        = 1;       // flag: is hostserv admin
  static const flag_type flagDisabled     = 1 << 1;  // user disabled ( e.g. after too many failed logins )
  static const flag_type flagUpdateOnAuth = 1 << 2;  // vhost may not be applied on auth up
  static const flag_type flagUpdateUser   = 1 << 3;
  static const flag_type flagChangeVHost  = 1 << 4;  // user may change his / her associated vhost
  static const flag_type flagVHostDBAdmin = 1 << 5;
  static const flag_type flagUserDBAdmin  = 1 << 6;
  static const flag_type flagNetworkAdmin = 1 << 7;
  // ^
  // '- these flags should be saved in config file

  
  static const flag_type flagOnline       = 1 << 8; // flag: user online / logedin

  static const flag_type flagUnauthed     = 1 << 9; // this is a flag which is used by 'HSIRCClient'.
                                                    // it is used as a workaround for the permission
                                                    // problem in class 'HostServ' when a user is not
                                                    // authed but in the irc network

  // used for flags string creation
  // sets the number of lowest bits to be stored in a config file
  static const flag_type MAX_FLAGS            = 8;

private:
  // default flags for a user if not other defined
  // this flag can be set on runtime via setDefaultFlags()
  static flag_type defaultFlags; 

  // flags automatically included with the admin flag
  static const flag_type IMPLICIT_ADMIN_FLAGS = flagChangeVHost | flagVHostDBAdmin |
  flagUserDBAdmin | flagNetworkAdmin;
      
public:
  HSClient( const string &loginName,
	    const string &loginPassword,
	    VHost * Vhost,
	    flag_type flags = defaultFlags );

  HSClient( const IRCToken& loginName,
	    const IRCToken& loginPassword,
	    VHost* Vhost,
	    flag_type flags = defaultFlags );

  HSClient( const HSClient& hcl );

  virtual ~HSClient();

  /**
   * @return ircclient is oper
   */
  inline bool isAdmin() const 
  { return (flags & flagAdmin ) == flagAdmin; };
  /**
   * @return ircclient is in irc
   */
  inline bool isOnline() const
  { return (flags & flagOnline ) == flagOnline; };
  /**
   * @return update hostname / username on auth up
   */ 
  inline bool updateOnAuth() const
  { return (flags & flagUpdateOnAuth ) == flagUpdateOnAuth; };
  /**
   * @return user may change his / her associated vhost
   */
  inline bool mayChangeVHost() const
  { return ( flags & flagChangeVHost ) == flagChangeVHost; };
  /**
   * @return client is disabled
   */
  inline bool isDisabled() const
  { return ( flags & flagDisabled ) == flagDisabled; };
  /**
   * @return username should be updated on auth up
   */
  inline bool updateUsername() const
  { return ( flags & flagUpdateUser ) == flagUpdateUser; };
  /**
   * @return user may change vhost <b>db</b> specific data
   */
  inline bool isVHostDBAdmin() const
  { return ( flags & flagVHostDBAdmin ) == flagVHostDBAdmin; };
  /**
   * @return user may control the user db
   */
  inline bool isUserDBAdmin() const
  { return ( flags & flagUserDBAdmin ) == flagUserDBAdmin; };
  /**
   * @return user may use sethost / setident and can view real hostname of user
   */
  inline bool isNetworkAdmin() const
  { return ( flags & flagNetworkAdmin ) == flagNetworkAdmin; };

  const IRCToken& getUsername() const
  { return loginName; };
  const IRCToken& getPassword() const
  { return loginPassword; };

  /**
   * @return vhost corresponding to user
   */ 
  const VHost* getVHost() const
  { return vhost; };

  flag_type getFlags() const
  { return flags; }

  inline void setOnline( bool b)
  { flags = b ? flags | flagOnline : flags & ~flagOnline; };
  inline void setAdmin( bool b )
  { flags = b ? flags | flagAdmin | IMPLICIT_ADMIN_FLAGS : flags & ~flagAdmin; };
  inline void setDisabled( bool b )
  { flags = b ? flags | flagDisabled : flags & ~flagDisabled; };
  inline void setUpdateOnAuthUp( bool b )
  { flags = b ? flags | flagUpdateOnAuth : flags & ~flagUpdateOnAuth; };
  inline void setUpdateUsername( bool b)
  { flags = b ? flags | flagUpdateUser : flags & ~flagUpdateUser; };
  inline void setChangeVHost( bool b )
  { flags = b ? flags | flagChangeVHost : flags & ~flagChangeVHost; };
  inline void setVHostDBAdmin( bool b )
  { flags = b ? flags | flagVHostDBAdmin : flags & ~flagVHostDBAdmin; };
  inline void setUserDBAdmin( bool b )
  { flags = b ? flags | flagUserDBAdmin : flags & ~flagUserDBAdmin; };
  inline void setNetworkAdmin( bool b )
  { flags = b ? flags | flagNetworkAdmin : flags & ~flagNetworkAdmin; };
  void setVHost( VHost* Vhost );

  void setPassword( const string& newPassword )
  { loginPassword = newPassword; };

  void setPassword( const IRCToken& newPassword )
  { loginPassword = newPassword; };

  /**
   * sets new default ircclient flags
   * @param defFlags new default flags
   */
  static void setDefaultFlags( int defFlags )
  { defaultFlags = defFlags; }

  string createFlagsString() const;

  /**
   * makes a string suitable as an ctor argument for HSClient
   * @param s string containing 0 and 1 ( length should be 3 at least )
   * @return binary representation of s
   */
  static flag_type parseFlagsString( const string &s );

  bool operator == ( const IRCToken& token ) const
  { return loginName == token; }

  bool operator ==( const HSClient &c ) const
  { return this == &c || loginName == c.loginName; };

  bool operator< ( const IRCToken& username ) const
  { return loginName < username; };

  bool operator < ( const HSClient &c ) const
  { 
    return loginName < c.loginName ; 
  };

  bool operator > ( const IRCToken& username ) const
  { return loginName > username; };

  bool operator > ( const HSClient &c ) const
  { return loginName > c.loginName; };

private:
  HSClient( );
  const HSClient & operator = ( const HSClient & );

  const IRCToken loginName;
  IRCToken loginPassword;
  flag_type flags;
  VHost *vhost;
};

ostream& operator<<( ostream& out, const HSClient& client );

#endif
