#ifndef HSCOMMANDS_HPP
#define HSCOMMANDS_HPP

#include <list>
#include <map>
#include <iostream>
#include <vector>
#include "hsircclient.hpp"
#include "hsircserverconnection.hpp"
#include "config.h"

namespace hostserv
{

namespace commands
{

class AbstractHSCommandState
{};

/**
 * Abstract hostserv command
 */
class AbstractHSCommand
{
   public:
      virtual ~AbstractHSCommand()
      {}

      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args, 
					       int argNum )                   = 0;
      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state ) = 0;
};

class CommandSet;

class AbstractCommandSetHSCommand : public virtual AbstractHSCommand
{
      friend class CommandSet;

   public:
      virtual ~AbstractCommandSetHSCommand()
      {}
      
   private:
      /*
       * gets called whenever a command was unregistered
       * from a CommandSet instance
       * 
       * we have introduced this hook to ensure destruction of
       * instances of AbstractHSCommand which are on the heap
       */
      virtual void onUnregistered() = 0;
};


/*
 * clients which derive from this class have to have a at least `:protected'
 * standard constructor
 */
class CommandSetHSCommandImpl : public virtual AbstractCommandSetHSCommand
{
   public:
      virtual ~CommandSetHSCommandImpl()
      {}

      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args,
					       int argNum )
      {
	 if( argNum < m_minArgs )
	 {
#ifdef DEBUG
	    cout << "CommandSetHSCommandImpl::execute(): not enough params" << endl;
	    cout << "CommandSetHSCommandImpl::execute(): received: " << argNum << " need: " << m_minArgs << endl;
#endif	   
	    return onHelp( conn, client, command, args, argNum );
	 }

	 return onExecute( conn, client, command, args, argNum );
      }

   protected:
      CommandSetHSCommandImpl( int minArgs )
	 : m_minArgs( minArgs )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command,
						 const IRCToken* args,
						 int argNum ) = 0;
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum ) = 0;
      
      void setMinArgs( int minimumArgs )
      { m_minArgs = minimumArgs; }

      static void printClientNotFound( HSIRCServerConnection& conn,
				       const HSIRCClient& target,
				       const IRCToken& clientNick );

      static void printSuccessfullOperation( HSIRCServerConnection& conn,
					     const HSIRCClient& target );
      static void printSuccessfullOperation( HSIRCServerConnection& conn,
					     const HSIRCClient& target,
					     const IRCToken& arg );

      static void printVHostNotFound( HSIRCServerConnection& conn,
				      const HSIRCClient& target,
				      const IRCToken& vhostName );

      static void printHSIRCClientDetails( HSIRCServerConnection& conn,
					   const HSIRCClient& sender,
					   const HSIRCClient* target,
					   bool printRealHostname );
      static void printHSClientDetails( HSIRCServerConnection& conn,
					const HSIRCClient& sender,
					const HSClient* target );

      /*
       * looks hostserv logins up 
       * @param username a token. beginning with '#' means the algorithm will look for `username' in
       *        the list of clients in irc network. else the regular login list will be searched
       */
      static HSClient* lookupHSClient( HSIRCServerConnection& conn, const IRCToken& username );

   private:
      virtual void onUnregistered()
      {}

      int m_minArgs;
};

/*
 * a command object that gets created by CommandSet.
 * it is a list of unfinished AbstractHSCommand* instances.
 * execute() ignores the command param and iterates through 
 * the list. for every object in the list we will call
 * resume with the appropriate AbstractHSCommandState object
 * Objects which resume as result_type not EXEC_UNFINISHED will
 * get removed from the list.
 * execute function will call assert( false ). Means this class
 * is only able to resume operation in logical meaning.
 */
class CommandStateMap : public AbstractCommandSetHSCommand, public AbstractHSCommandState
{
   public:
      static CommandStateMap* makeMap()
      { return new CommandStateMap(); }

      virtual ~CommandStateMap()
      {
#ifdef VERBOSE_MEMORY
	 cout << "CommandStateMap] destroyed" << endl;
#endif	 
      }

      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args,
					       int argNum )
      { assert( false ); return NULL; }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state );

      void addCommand( AbstractHSCommand* command, 
		       AbstractHSCommandState* state )
      { m_list.push_back( ListElem( command, state ) ); }

   protected:
      CommandStateMap()
	 : m_list()
      {
#ifdef VERBOSE_MEMORY
	 cout << "CommandStateMap] created" << endl;
#endif
      }

   private:
      struct ListElem
      {
	    ListElem( AbstractHSCommand* command,
		      AbstractHSCommandState* state )
	       : m_command( command ),
		 m_state( state )
	    {}

	    AbstractHSCommand*      m_command;
	    AbstractHSCommandState* m_state;
      };
      typedef std::vector<ListElem> list_type;

      list_type m_list;

      virtual void onUnregistered()
      { delete this; }
};

/*
 * a sorted collection of AbstractCommandSetHSCommand instances.
 * multiple usage of command names are allowed.
 * Notice: CommandSet is also an AbstractCommandSetHSCommand. Means we can
 *         composite CommandSet instances recursively.
 */
class CommandSet : public virtual AbstractCommandSetHSCommand
{
   public:
      CommandSet()
	 : m_commands()
      {}

      virtual ~CommandSet()
      {}

      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args,
					       int argNum );
      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state );

      /**
       * adds command `command' named `name' into the set
       */
      void registerCommand( const IRCToken& name,
			    AbstractCommandSetHSCommand* command )
      { m_commands.push_back( make_pair( name, command ) ); }
      
      /** 
       * unregisters _ALL_ commands associated with `name'
       */
      void unregisterCommand( const IRCToken& name );

      /**
       * removes instance at address `command' from the set
       * 
       * @return number of removed occurencies of `command'
       */
      int unregisterCommand( const AbstractCommandSetHSCommand* command );

   protected:
      /**
       * gets called by execute whenever a command named `command' was not found.
       */
      virtual void onCommandNotFound( HSIRCServerConnection& conn,
				      HSIRCClient& client,
				      const IRCToken& command,
				      const IRCToken* args,
				      int argNum )
      {}

   private:
      typedef std::pair<IRCToken, AbstractCommandSetHSCommand*> commandsetentry_type;
      typedef std::vector<commandsetentry_type> commandset_type;

      static bool commandSetPredicate( const commandsetentry_type& p, const IRCToken& cmd )
      { return p.first == cmd; }

      commandset_type         m_commands;

      class CSetPredicate
      {
	 public:
	    CSetPredicate( const IRCToken& name )
	       : m_name( name )
	    {}
	    
	    bool operator()( const commandsetentry_type& p )
	    { return p.first == m_name; }
	    
      private:
	    const IRCToken& m_name;
      };



      virtual void onUnregistered()
      {}
};

// /** 
//  * merges several commandset instances to one monolithic AbstractCommandSetHSCommand
//  *
//  */
// class CommandSetComposition : public AbstractCommandSetHSCommand, public CommandSet
// {
//    public:
//       CommandSetComposition()
// 	 : m_commandSets()
//       { 
// 	 // insert this / instance as part of the composition
// 	 m_commandSets[0] = const_cast<CommandSetComposition*>( this );
//       }

//       virtual ~CommandSetComposition()
//       {}

//       virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
// 					       HSIRCClient& client,
// 					       const IRCToken& command, 
// 					       const IRCToken* args, 
// 					       int argNum );

//       virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
//       { return NULL; }

//       // as CommandSet::resume() just calls CommandStateMap::resume() and the CommandStateMap supports resuming
//       // we do not have to overwrite anything

//       void addCommandSet( CommandSet* commandSet );
//       bool removeCommandSet( CommandSet* commandSet );

//    private:
//       typedef std::vector<CommandSet*> commandsetvector_type;
      
//       commandsetvector_type m_commandSets;

//       virtual void onUnregistered()
//       {}
// };


/*
 * Combines CommandSet and CommandSetHSCommandImpl into 1 object.
 * We will use this object to automatize argNum looking. 
 * As every command set is a set of commands and expects therefore
 * at least 1 paramter, it makes no sense to implement the same
 * code again and again. That is basically what this class does.
 *
 * This class overwrites onCommandNotFound from CommandSet. 
 * HSCommandSet::onCommandNotFound calls onHelp();
 *
 * Furthermore this class does not check in execute() if the given 
 * command is correct. in onExecute we will make the recursive calls
 * working by ignoring the command by using as the command the first
 * argument.
 * This also means that classes which derive from HSCommandSet should
 * check if they want this behaviour. E.g. HS[Un,A]uthedCommands 
 * overwrites onExecute to make checks to implement checks on the given
 * command.
 */
class HSCommandSet : public CommandSet, public CommandSetHSCommandImpl
{
   public:
      // g++ does not seem to care about these two lines ... :/
//       using CommandSetHSCommandImpl::execute;
//       using CommandSet::resume;

      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args,
					       int argNum )
      { return CommandSetHSCommandImpl::execute( conn, client, command, args, argNum ); }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return CommandSet::resume( state ); }


   protected:
      HSCommandSet( HSClient::flag_type minFlags )
	 : CommandSet(),
	   CommandSetHSCommandImpl( 1 ),
	   m_minFlags( minFlags )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum )
      { 
	 if( client.getHSClient()->getFlags() & m_minFlags )
	    return CommandSet::execute( conn, client, *args, args+1, argNum-1 );
	 return 
	    onPermissionViolation( conn, client, *args, args+1, argNum-1 );
      }

      virtual void onCommandNotFound( HSIRCServerConnection& conn,
				      HSIRCClient& client,
				      const IRCToken& command,
				      const IRCToken* args,
				      int argNum )
      { onHelp( conn, client, command, args, argNum ); }

      virtual AbstractHSCommandState* onPermissionViolation( HSIRCServerConnection& conn,
							     HSIRCClient& client,
							     const IRCToken& command,
							     const IRCToken* args,
							     int argNum )
      { return NULL; }

      void setMinAccess( HSClient::flag_type minimumAccess )
      { m_minFlags = minimumAccess; }

   private:
      virtual void onUnregistered()
      {}

      HSClient::flag_type m_minFlags;
};

/*
 * command set of all the `admin *' commands
 */
class HSAdminCommandSet : public HSCommandSet
{
   public:
      static const IRCToken CMD_SHUTDOWN;
      static const IRCToken CMD_RECONNECT;
      static const IRCToken CMD_RELOAD;

      static const IRCToken SUPER_CMD;

      virtual ~HSAdminCommandSet()
      {}

      static HSAdminCommandSet* getSet();

      static const IRCToken& getName()
      { return SUPER_CMD; }
      

   protected:
      HSAdminCommandSet()
	 : HSCommandSet( HSClient::flagAdmin )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command,
					      const IRCToken* args,
					      int argNum );

   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

/*
 * 'admin shutdown [yes]' command
 */
class HSAdminShutdownCommand : public CommandSetHSCommandImpl
{
   public:
      static HSAdminShutdownCommand* getCommand()
      { 
	 static HSAdminShutdownCommand* command = new HSAdminShutdownCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSAdminShutdownCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'admin reconnect [yes]' command
 */
class HSAdminReconnectCommand : public CommandSetHSCommandImpl
{
   public:
      static HSAdminReconnectCommand* getCommand()
      {
	 static HSAdminReconnectCommand* command = new HSAdminReconnectCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }


   protected:
      HSAdminReconnectCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'admin reload [yes]' command
 */
class HSAdminReloadCommand : public CommandSetHSCommandImpl
{
   public:
      static HSAdminReloadCommand* getCommand()
      {
	 static HSAdminReloadCommand* command = new HSAdminReloadCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSAdminReloadCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};







/*
 * command set of all the `net *' commands
 */
class HSNetCommandSet : public HSCommandSet
{
   public:
      static const IRCToken CMD_SETHOST;
      static const IRCToken CMD_SETIDENT;

      static const IRCToken SUPER_CMD;

      static HSNetCommandSet* getSet();
      
      static const IRCToken& getName()
      { return SUPER_CMD; }

      virtual ~HSNetCommandSet()
      {}

   protected:
      HSNetCommandSet()
	 : HSCommandSet( HSClient::flagNetworkAdmin )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

   private:
      void initSet();
};

/*
 * 'net sethost [username] [hostname]' command
 */
class HSNetSethostCommand : public CommandSetHSCommandImpl
{
   public:
      static HSNetSethostCommand* getCommand()
      {
	 static HSNetSethostCommand* command = new HSNetSethostCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSNetSethostCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'net setident [victim] [ident]' command
 */
class HSNetSetidentCommand : public CommandSetHSCommandImpl
{
   public:
      static HSNetSetidentCommand* getCommand()
      {
	 static HSNetSetidentCommand* command = new HSNetSetidentCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSNetSetidentCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/**
 * this class provides a generic mechanism to manipulate 
 * hsclient flags through member functions.
 *
 * `onHelp' has to be implemented by deriving classes
 */
class HSClientPermissionChangingCommand : public CommandSetHSCommandImpl
{
   protected:
      typedef void (HSClient::*memFunc)( bool );
      
      HSClientPermissionChangingCommand( int minArgs,
					 memFunc func )
	 : CommandSetHSCommandImpl( minArgs ),
	   m_memberFunction( func )
      {}

      HSClientPermissionChangingCommand( memFunc func )
	 : CommandSetHSCommandImpl( 1 ),
	   m_memberFunction( func )
      {}
      
      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );

      inline void updatePermission( HSClient* cl, bool val )
      { (cl->*m_memberFunction)( val ); }

   private:
      memFunc const m_memberFunction;
};

/** 
 * this class provides parametrized hsclient-flags-setting with client lookup
 * 
 * `onHelp' has to be implemented by deriving classes
 */
class HSClientLookingUpPermissionChangingCommand : public HSClientPermissionChangingCommand
{
   protected:
      HSClientLookingUpPermissionChangingCommand( memFunc func )
	 : HSClientPermissionChangingCommand( 2, func )
      {}

      /*
       * looks up client username and tries to get an hsclient instance
       * of this client. if this fails 'clientNotFound()' gets called
       */
      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
};


/*
 * command set of all the `me *' commands
 */
class HSMeCommandSet : public HSCommandSet
{
   public:
      static const IRCToken CMD_SETVHOST;
      static const IRCToken CMD_UPDATE;
      static const IRCToken CMD_LISTPUBLIC;
      static const IRCToken CMD_SETPASSWORD;
      static const IRCToken CMD_DETAILS;
      static const IRCToken CMD_RESTORE;
      static const IRCToken CMD_SETUPDATEONAUTH;
      static const IRCToken CMD_SETAFFECTUSERNAME;
      
      static const IRCToken SUPER_CMD;

      virtual ~HSMeCommandSet()
      {}

      static HSMeCommandSet* getSet();
      
      static const IRCToken& getName()
      { return SUPER_CMD; }

   protected:
      HSMeCommandSet()
	 : HSCommandSet( HSClient::flagOnline )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

/*
 * 'me setvhost <public vhost name>' command
 */
class HSMeSetVHostCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeSetVHostCommand* getCommand()
      {
	 static HSMeSetVHostCommand* command = new HSMeSetVHostCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeSetVHostCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'me update' command
 */
class HSMeUpdateCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeUpdateCommand* getCommand()
      {
	 static HSMeUpdateCommand* command = new HSMeUpdateCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeUpdateCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'me listpublic <criteria>' command
 *
 * supports operation resuming
 * for every processed vhost onPrintVHost will get called. 
 */
class HSMeListpublicCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeListpublicCommand* getCommand()
      {
	 static HSMeListpublicCommand* command = new HSMeListpublicCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state );

   protected:
      HSMeListpublicCommand()
	 : CommandSetHSCommandImpl( 0 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

      /*
       * @return vhost printed
       */
      virtual bool onPrintVHost( HSIRCServerConnection& conn, 
				 HSIRCClient& client,
				 const VHost* vhost,
				 const IRCToken& criteria );

      struct CommandState : public AbstractHSCommandState
      {
	    const static int                MAX_VHOSTS_PER_CALL = 10;
	    const static int                MAX_VHOSTS_PRINT    = 600;

	    HSIRCServerConnection&          m_conn;
	    HSIRCClient&                    m_client;
	    const IRCToken                  m_criteria;
	    VHostList::const_iterator       m_pos;
	    int                             m_numPrinted;

	    CommandState( HSIRCServerConnection& conn,
			  HSIRCClient& client,
			  const IRCToken& criteria,
			  const VHostList::const_iterator iter )
	       : m_conn( conn ),
		 m_client( client ),
		 m_criteria( criteria ),
		 m_pos( iter ),
		 m_numPrinted( 0 )
	    {}
      };
};

/*
 * 'me setpassword <oldpassword> <newpassword>' command
 */
class HSMeSetPasswordCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeSetPasswordCommand* getCommand()
      {
	 static HSMeSetPasswordCommand* command = new HSMeSetPasswordCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

      /**
       * @return -1 new password contains illegal characters
       *          1 password changed 
       *          0 old and real password do not match
       */
      static IRCToken::size_type changePassword( HSClient* target,
						 const IRCToken& newPassword );
   protected:
      HSMeSetPasswordCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'me details' command
 */
class HSMeDetailsCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeDetailsCommand* getCommand()
      {
	 static HSMeDetailsCommand* command = new HSMeDetailsCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeDetailsCommand()
	 : CommandSetHSCommandImpl( 0 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum )
      { return NULL; }
};

/*
 * 'me setupdateonauth <1|0>' command
 */
class HSMeRestoreCommand : public CommandSetHSCommandImpl
{
   public:
      static HSMeRestoreCommand* getCommand()
      {
	 static HSMeRestoreCommand* command = new HSMeRestoreCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeRestoreCommand()
	 : CommandSetHSCommandImpl( 0 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum )
      { return NULL; }
};

/*
 * 'me setupdateonauth <1|0>' command
 */
class HSMeSetUpdateonauthCommand : public HSClientPermissionChangingCommand
{
   public:
      static HSMeSetUpdateonauthCommand* getCommand()
      {
	 static HSMeSetUpdateonauthCommand* command = new HSMeSetUpdateonauthCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeSetUpdateonauthCommand()
	 : HSClientPermissionChangingCommand( &HSClient::setUpdateOnAuthUp )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * 'me setaffectusername <1|0>' command
 */
class HSMeSetAffectusernameCommand : public HSClientPermissionChangingCommand
{
   public:
      static HSMeSetAffectusernameCommand* getCommand()
      {
	 static HSMeSetAffectusernameCommand* command = new HSMeSetAffectusernameCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSMeSetAffectusernameCommand()
	 : HSClientPermissionChangingCommand( &HSClient::setUpdateUsername )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};



/*
 * command set of all the 'client *' commands
 */
class HSClientCommandSet : public HSCommandSet
{
   public:
      static const IRCToken CMD_ADD;
      static const IRCToken CMD_DEL;
      static const IRCToken CMD_SETPASSWORD;
      static const IRCToken CMD_DISABLE;
      static const IRCToken CMD_UPDATEONAUTH;
      static const IRCToken CMD_UPDATEUSERNAME;
      static const IRCToken CMD_ADMIN;
      static const IRCToken CMD_VHOSTADMIN;
      static const IRCToken CMD_USERADMIN;
      static const IRCToken CMD_NETWORKADMIN;
      static const IRCToken CMD_SETVHOST;
      static const IRCToken CMD_LIST;
      static const IRCToken CMD_DETAILS;

      static const IRCToken SUPER_CMD;

      virtual ~HSClientCommandSet()
      {}

      static HSClientCommandSet* getSet();

      static const IRCToken& getName()
      { return SUPER_CMD; }

   protected:
      HSClientCommandSet()
	 : HSCommandSet( HSClient::flagUserDBAdmin | HSClient::flagAdmin )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command,
					      const IRCToken* args,
					      int argNum );

   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

/*
 * 'client details [username]' command
 */
class HSClientDetailsCommand : public virtual CommandSetHSCommandImpl
{
   public:
      static HSClientDetailsCommand* getCommand()
      {
	 static HSClientDetailsCommand* command = new HSClientDetailsCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientDetailsCommand()
	 : CommandSetHSCommandImpl( 0 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};


/*
 * 'clients list [criteria]' command
 */
class HSClientListCommand : public CommandSetHSCommandImpl
{
   public:
      static HSClientListCommand* getCommand()
      {
	 static HSClientListCommand* command = new HSClientListCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state );

   protected:
      HSClientListCommand()
	 : CommandSetHSCommandImpl( 0 )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum )
      { return NULL; }

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum )
      {
	 if( argNum > 0 )
	    return resume( new CommandState( conn, client, *args, conn.getHSClients()->begin() ) );
	 else
	    return resume( new CommandState( conn, client, IRCToken( "*" ), conn.getHSClients()->begin() ) );
      }

      virtual bool onPrintClient( HSIRCServerConnection& conn,
				  HSIRCClient& client,
				  const HSClient* hsclient,
				  const IRCToken& criteria );



      struct CommandState : public AbstractHSCommandState
      {
	    const static int MAX_ROWS_PER_CALL   = 3;
	    const static int MAX_CLIENTS_PER_ROW = 10;
	    const static int MAX_NUM_CALLED      = 4;

	    HSIRCServerConnection&       m_conn;
	    HSIRCClient&                 m_client;
	    const IRCToken               m_criteria;
	    HSClientList::const_iterator m_pos;
	    int                          m_numCalled;

	    CommandState( HSIRCServerConnection& conn,
			  HSIRCClient& client,
			  const IRCToken& criteria,
			  const HSClientList::const_iterator iter )
	       : m_conn( conn ),
		 m_client( client ),
		 m_criteria( criteria ),
		 m_pos( iter ),
		 m_numCalled( 0 )
	    {} 
      };
};


/*
 * 'clients setvhost [username] [vhostname]' command
 */
class HSClientSetvhostCommand : public CommandSetHSCommandImpl
{
   public:
      static HSClientSetvhostCommand* getCommand()
      {
	 static HSClientSetvhostCommand* command = new HSClientSetvhostCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientSetvhostCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientNetworkadminCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientNetworkadminCommand* getCommand()
      {
	 static HSClientNetworkadminCommand* command = new HSClientNetworkadminCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientNetworkadminCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setNetworkAdmin )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};


class HSClientUseradminCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientUseradminCommand* getCommand()
      {
	 static HSClientUseradminCommand* command = new HSClientUseradminCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientUseradminCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setUserDBAdmin )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};


class HSClientVHostadminCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientVHostadminCommand* getCommand()
      {
	 static HSClientVHostadminCommand* command = new HSClientVHostadminCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientVHostadminCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setVHostDBAdmin )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientAdminCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientAdminCommand* getCommand()
      {
	 static HSClientAdminCommand* command = new HSClientAdminCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientAdminCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setAdmin )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientUpdateonauthCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientUpdateonauthCommand* getCommand()
      {
	 static HSClientUpdateonauthCommand* command = new HSClientUpdateonauthCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientUpdateonauthCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setUpdateOnAuthUp )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientUpdateusernameCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientUpdateusernameCommand* getCommand()
      {
	 static HSClientUpdateusernameCommand* command = new HSClientUpdateusernameCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientUpdateusernameCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setUpdateUsername )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

};

class HSClientDisableCommand : public HSClientLookingUpPermissionChangingCommand
{
   public:
      static HSClientDisableCommand* getCommand()
      {
	 static HSClientDisableCommand* command = new HSClientDisableCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientDisableCommand()
	 : HSClientLookingUpPermissionChangingCommand( &HSClient::setDisabled )
      {}

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientDelCommand : public CommandSetHSCommandImpl
{
   public:
      static HSClientDelCommand* getCommand()
      {
	 static HSClientDelCommand* command = new HSClientDelCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientDelCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientAddCommand : public CommandSetHSCommandImpl
{
   public:
      static HSClientAddCommand* getCommand()
      {
	 static HSClientAddCommand* command = new HSClientAddCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientAddCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSClientSetPasswordCommand : public CommandSetHSCommandImpl
{
   public:
      static HSClientSetPasswordCommand* getCommand()
      {
	 static HSClientSetPasswordCommand* command = new HSClientSetPasswordCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSClientSetPasswordCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};


class HSDeauthCommand : public CommandSetHSCommandImpl
{
   public:
      static void logOut( HSIRCServerConnection& conn,
			  HSIRCClient& client )
      {
	 client.logout();

	 static const IRCToken MSG1( "You have been loged out" );

	 conn.printNotice( conn.getHostservNick(), client.getNickname(), 
			   MSG1 );
      }	

      static HSDeauthCommand* getCommand()
      {
	 static HSDeauthCommand* command = new HSDeauthCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSDeauthCommand()
	 : CommandSetHSCommandImpl( -1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum )
      {
	 logOut( conn, client );
	 return NULL;
      }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum )
      { 
	 logOut( conn, client );
	 return NULL; 
      }
};


class HSAuthCommand : public CommandSetHSCommandImpl
{
   public:
      static HSAuthCommand* getCommand()
      {
	 static HSAuthCommand* command = new HSAuthCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSAuthCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSVHostAddCommand : public CommandSetHSCommandImpl
{
   public:
      static HSVHostAddCommand* getCommand()
      {
	 static HSVHostAddCommand* command = new HSVHostAddCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSVHostAddCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSVHostListCommand : public HSMeListpublicCommand
{
   public:
      static HSVHostListCommand* getCommand()
      {
	 static HSVHostListCommand* command = new HSVHostListCommand();

	 return command;
      }

   protected:
      HSVHostListCommand()
	 : HSMeListpublicCommand()
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

      virtual bool onPrintVHost(  HSIRCServerConnection& conn, 
				  HSIRCClient& client,
				  const VHost* vhost,
				  const IRCToken& criteria );
};

class HSVHostPublicCommand : public CommandSetHSCommandImpl
{
   public:
      static HSVHostPublicCommand* getCommand()
      {
	 static HSVHostPublicCommand* command = new HSVHostPublicCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSVHostPublicCommand()
	 : CommandSetHSCommandImpl( 2 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSVHostReferencesCommand : public CommandSetHSCommandImpl
{
   public:
      static HSVHostReferencesCommand* getCommand()
      {
	 static HSVHostReferencesCommand* command = new HSVHostReferencesCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSVHostReferencesCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

class HSVHostDelCommand : public CommandSetHSCommandImpl
{
   public:
      static HSVHostDelCommand* getCommand()
      {
	 static HSVHostDelCommand* command = new HSVHostDelCommand();

	 return command;
      }

      virtual AbstractHSCommandState* resume( AbstractHSCommandState* state )
      { return NULL; }

   protected:
      HSVHostDelCommand()
	 : CommandSetHSCommandImpl( 1 )
      {}

      virtual AbstractHSCommandState* onExecute( HSIRCServerConnection& conn,
						 HSIRCClient& client,
						 const IRCToken& command, 
						 const IRCToken* args, 
						 int argNum );
      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );
};

/*
 * command set of all the 'vhost *' commands
 */
class HSVHostCommandSet : public HSCommandSet
{
   public:
      static const IRCToken CMD_ADD;
      static const IRCToken CMD_LIST;
      static const IRCToken CMD_PUBLIC;
      static const IRCToken CMD_REFERENCES;
      static const IRCToken CMD_DEL;
      
      static const IRCToken SUPER_CMD;

      virtual ~HSVHostCommandSet()
      {}

      static HSVHostCommandSet* getSet();

      static const IRCToken& getName()
      { return SUPER_CMD; }

   protected:
      HSVHostCommandSet()
	 : HSCommandSet( HSClient::flagVHostDBAdmin | HSClient::flagAdmin )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command,
					      const IRCToken* args,
					      int argNum );

   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

/** 
 * implements the command checking behaviour as mentioned in the documentation
 * for HSCommandSet. 
 * Means we will not ignore the incoming command but check whether the command
 * is part of the Set instead of the first argument checks what HSCommandSet uses
 *
 * Should be used for _TOPLEVEL_ commandsets.
 */
class HSCommandSetComposition : public virtual HSCommandSet
{
   public:
      virtual AbstractHSCommandState* execute( HSIRCServerConnection& conn,
					       HSIRCClient& client,
					       const IRCToken& command,
					       const IRCToken* args, 
					       int argNum )
      { return CommandSet::execute( conn, client, command, args, argNum ); }

   protected:
      HSCommandSetComposition()
	 : HSCommandSet( HSClient::flagOnline )
      {}
};

/*
 * composition of commandsets for authed hsircclients
 */
class HSAuthedCommands : public HSCommandSetComposition
{
   public:
      static const IRCToken CMD_DEAUTH;

      static HSAuthedCommands* getSet()
      { 
	 static HSAuthedCommands* set = new HSAuthedCommands();

	 return set;
      }

   protected:
      HSAuthedCommands()
	 : HSCommandSetComposition(),
	   HSCommandSet( HSClient::flagOnline )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command,
					      const IRCToken* args,
					      int argNum );
   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

/*
 * composition of HSCommandSetHSCommand instances for unauthed clients
 */
class HSUnauthedCommands : public HSCommandSetComposition
{
   public:
      static const IRCToken CMD_AUTH;

      static HSUnauthedCommands* getSet()
      {
	 static HSUnauthedCommands* set = new HSUnauthedCommands();

	 return set;
      }

   protected:
      HSUnauthedCommands()
	 : HSCommandSetComposition(),
	   HSCommandSet( HSClient::flagUnauthed )
      { initSet(); }

      virtual AbstractHSCommandState* onHelp( HSIRCServerConnection& conn,
					      HSIRCClient& client,
					      const IRCToken& command, 
					      const IRCToken* args, 
					      int argNum );

   private:
      void initSet();

      virtual void onUnregistered()
      {}
};

} //EON hostserv::commands
} //EON hostserv



#endif
