/***************************************************************************
                           httpclientconnection.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: httpclientconnection.hpp,v 1.14 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HTTPCLIENTCONNECTION_HPP
#define HTTPCLIENTCONNECTION_HPP


#include "httpd.hpp"
#include "request.hpp"
#include "response.hpp"
#include "requestfilebuffer.hpp"
#include "tcpclientconnection.hpp"
#include "responsefilebuffer.hpp"


namespace http
{
class Request;
}

namespace httpd
{

class HTTPd;

  /**
   * this class receives extern stimuli.
   * After an instance of it has been created 
   * HTTPClientConnection will only receive input
   * events and will try to form a request object.
   * Done this request object will be 'transmitted'
   * to the http server. The http server does something
   * with the request object and then send a response
   * object back to the httpclientconnection.
   * Then the output event will registered for the
   * same clientconnection while the input event is
   * unregistered.
   *
   * Note: a request object <b>MUST NOT</b> be transmitted
   * to the HTTPd server.
   * <br><br>
   *
   * Furthermore: this class will delete itself whenever
   * it got removed from the poller queue.
   * <br><br>
   */
class HTTPClientConnection : virtual public Pollable, protected TCPClientConnection
{
public:
  typedef unsigned int id_type;

  HTTPClientConnection( HTTPd* server, 
			TCPSocket::fd_type Fd,
			id_type id );

  virtual ~HTTPClientConnection();

  virtual const File* getPollableFile() const
  { return this; }

  void writeResponse( const Response* resp );

  const HTTPd* getServer() const
  { return m_server; }

  id_type getID() const
  { return m_id; }

protected:
  using File::getFD;
  using File::close;

  static const len_type DEF_INBUFLEN  = 1024;
  /*
   * this should be at least == ResponseBody::maxJunkSize
   */
  static const len_type DEF_OUTBUFLEN = 2060;

  /**
   * @see Pollable#onInput
   */
  virtual void onInput();
  /**
   * gets called after a request has been read and send to the httpd
   * @see Pollable#onOutput
   */
  virtual void onOutput();
  /**
   * @see Pollable#Error
   */
  virtual void onError();
  /**
   * @see Pollable#onHangup
   */
  virtual void onHangup();
  /**
   * used to autodestruct 'this'
   * @see Pollable#onRemoved
   */
  virtual void onRemoved()
  { delete this; }

private:
  HTTPClientConnection();
  HTTPClientConnection( const HTTPClientConnection& );
  const HTTPClientConnection& operator=( const HTTPClientConnection& );
  
  RequestFileBuffer     m_inputBuffer;
  ResponseFileBuffer    m_outputBuffer;
  ostream               m_ostream;
  HTTPd*                m_server;
  Request*              m_request;
  const int             m_id;

  /**
   * gets set by 'write'
   * gets deleted by the appropriate method
   */
  const Response*   m_response;
  
  enum { PRINTING_HEADER = 0, PRINTING_BODY, PRINTING_DONE } m_printStatus;
  enum { RECV_STATUSLINE = 0, RECV_HEADER,   RECV_BODY }     m_recvStatus;

  /**
   * call if after a response has been printed
   */
  void printingDone();
};

}//EON httpd

#endif
