/***************************************************************************
                           httpd.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: httpd.hpp,v 1.9 2002/05/19 17:54:03 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HTTPD_HPP
#define HTTPD_HPP


#include <vector>
#include "httpserverport.hpp"
#include "httpclientconnection.hpp"
#include "request.hpp"
#include "pollerclient.hpp"
#include "fifo.hpp"


class HostServ;

namespace httpd
{

class HTTPServerPort;
class HTTPClientConnection;

//FIXME: use an array of HTTPClientConnections
//       to make getPollerClient() be more efficient

/**
 * This is the central IO http data datastructure.<br>
 * HTTPd gets 'stimulted' ( receives messages ) from HTTPClientConnections.
 * They read and parse requests which get sended here to member function
 * 'processRequest'.<br>
 * If HTTPd receives such an event he will process the data someday and
 * send a response back to the client. Whenever this response got processed
 * the client will call another hook: responsePrinted. All this mechanism
 * has been implemented to remove a PollerClient object from the HTTPClient-
 * connection. PollerClient eventregistration is done via this class.
 */
class HTTPd
{
public:
  typedef vector<MeasuringPollerClient*> httpclientlist_type;
  typedef unsigned int id_type;
  
  /**
   * creates a new HTTPd instance
   * - opens a tcpport at 'bind':'port'
   * - registers this port at 'poller'
   * - makes the port be in listening state for incoming connections
   *
   * @param maxTimeoutTime number of seconds which a connection is allowed to idle
   * @param maxClients maximum number of simultaneously served clients
   */
  HTTPd( const HostServ* app,
	 Poller* poller, 
	 const InetHostAddress& bind, TCPSocket::port_type port,
	 int maxClients = 5,
	 long maxTimeoutTime = 30 );
  
  virtual ~HTTPd();

  /**
   * @return number of removed clients
   */
  int runRemoveTimeoutdClients();

  void addClientConnection( File::fd_type fd );

  /*
   * gets called by HTTPClientConnection objects.
   * Removes conn if found in queue from the queue.
   * - Does not delete `conn'.
   */
  void removeClientConnection( HTTPClientConnection* conn );

  void setMaxTimeout( long maxTimeout )
  { m_maxTimeout = maxTimeout; }

  void processRequest( HTTPClientConnection* from,
		       Request* request ); 

  void responsePrinted( HTTPClientConnection* who,
			const Response& what );

  void run();

  const HostServ* getApplication() const
  { return m_hostserv; }

private:
  const HostServ* const m_hostserv;
  Poller*               m_poller;
  HTTPServerPort*       m_tcpPort;
  PollerClient          m_tcpPortPollerClient;

  /**
   * list of currently available clients
   */
  httpclientlist_type m_clients;

  typedef pair<HTTPClientConnection*, Request*> queueelem_type;
  typedef Fifo<queueelem_type> queuecontainer_type;

  queuecontainer_type m_requestQueue;

  /*
   * current number of httpd clients
   */
  int m_clientNum;
  
  /*
   * maximum number of simultaneous clients
   */
  const int m_maxClientNum;

  /**
   * the time interval a client is allowed not to send any
   * data before get disconnected
   */
  long m_maxTimeout;

  id_type getNextFreeID();
};

}//EON httpd


#endif
