/***************************************************************************
                           ircserverconnection.hpp
                           -------------------
    copyright            : (C) 2001,2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: ircserverconnection.hpp,v 1.22 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef _IRCSERVERCONNECTION_HPP
#define _IRCSERVERCONNECTION_HPP

extern "C" {
#include <stdarg.h>
}

#include "ircclientconnection.hpp"
#include "config.h"

using namespace std;

class IRCServerConnection;

/**
 * implements irc- server to irc- server connections
 * const unsigned char* format arguments can be formatted as known for printf():
 * %s - print string 
 * %% - print '%'
 * %i / %d : print integer
 * %c : print single characters
 * %l : print long
 * furthermore: 
 * %B : bold
 * %U : underlined
 * IMPORTANT: you should not terminate the messages by hand using \n. That
 * is done automatically.
 */
class IRCServerConnection : public IRCClientConnection {
public:
  IRCServerConnection( const IRCToken &servname, 		       
		       const IRCToken &Password,
		       const IRCToken &description,
		       Poller* poller,
		       len_type inputBufLen,
		       len_type outputBufLen );
  virtual ~IRCServerConnection();

  const IRCToken& getDescription() const
  { return myDescription; }
  const IRCToken& getServername() const
  { return hostname; }
  const IRCToken& getPassword() const
  { return password; }

  static const IRCToken CMD_PASS;
  static const IRCToken CMD_SERVER;
  static const IRCToken CMD_PING;
  static const IRCToken CMD_PONG;
  static const IRCToken CMD_NICK;
  static const IRCToken CMD_MODE;
  static const IRCToken CMD_QUIT;
  static const IRCToken CMD_WHOIS;
  static const IRCToken ERR_433;
  static const IRCToken ERR_436;
  static const IRCToken CMD_PRIVMSG;
  static const IRCToken CMD_ERROR;
  static const IRCToken CMD_NOTICE;
  static const IRCToken CMD_SJOIN;
  static const IRCToken CMD_JOIN;
  static const IRCToken CMD_SQUIT;
  static const IRCToken CMD_KILL;
#if SERVER_TYPE == SERVER_HYBRID || SERVER_TYPE == SERVER_BAHAMUT
  static const IRCToken CMD_SVINFO;
#else
  static const IRCToken CMD_NETINFO;
#endif
  static const char CTCP_FORMAT_FLAG = (char) 6;
  static const IRCToken RPL_WHOISUSER;
  static const IRCToken RPL_WHOISSERVER;
  static const IRCToken RPL_ENDOFWHOIS;
  

  static const int MAX_MSG_LEN = 512;
  
protected:
  /**
   * introduces a new user to the irc network
   * @param nick nickname of the new user
   * @param ident ident of the new user
   * @param address address of the new user
   * @param modes initial user modes
   * @param realname user's realname in irc
   */
  void introduceUser( const IRCToken &nick, const IRCToken &ident, 
		      const IRCToken &address, const IRCToken &modes, 
		      const IRCToken &realname );  
  /**
   * virtually disconnectes a user from the irc network
   * @param nick nickname of the user to disconnect
   */
  void removeUser( const IRCToken &nick );

  /**
   * lets a user join a channel
   * @param nick nickname of a local user
   * @param chan name of the channel to join
   */
  void joinChannel( const IRCToken &nick, const IRCToken &chan );

  /**
   * called after the server to server connection has been established
   */
  virtual void onDoLogin();
  
  /**
   * called after a connection has been established with the remote server
   */
  virtual void onLogedin() 
  {}

  /**
   * called when a privmsg has been received
   * @param m irc privmsg
   */
  virtual void onPrivmsg(const IRCToken &from, const IRCToken &to,
			       const IRCToken* msg, int numArgs ) 
  {};

  /**
   * called when a notice has been received
   * @param m irc notice message
   *
   */
  virtual void onNotice(const IRCToken &from, const IRCToken& to,
			const IRCToken* msg, int numArgs ) 
  {};

  /**
   * [not implemented yet] called when a ctcp privmsg has been received
   * @param m irc ctcp message
   */
  virtual void onCtcp(const IRCToken& from, const IRCToken& to,
		      const IRCToken* msg, int numArgs ) 
  {}; // not implemented yet

  /**
   * called when a mode change comand has been received
   * @param m irc mode message
   */
  virtual void onMode( const IRCServerMessage& msg ) 
  {};

  /**
   * called when an error message has been received
   * @param m irc error message
   */
  virtual void onServerError( const IRCServerMessage &m ) 
  {};

  /**
   * called when an client does a remote whois
   * @param from whoiser
   * @param who  who gets whoised
   */
  virtual void onWhois( const IRCToken& from, const IRCToken& who )
  {};
  /**
   * called when a user has changed his nickname
   * @param oldNick previous nickname 
   * @param newNick current nickname
   * @param timestamp unix timestamp when the user changed his nick
   */
  virtual void onNickChange( const IRCToken &oldNick, const IRCToken &newNick )
  {};

  /**
   * called after a new user has been introduced to the irc network
   * @param nick nickname of the new user
   * @param ident ident of the new user
   * @param address address of the new user
   * @param realname realname of the new user
   * @param modes user modes of the new user
   * @param timestamp since when the user is part of the network
   */
  virtual void onNickIntroduce( const IRCToken &nick, const IRCToken &ident, 
				const IRCToken &address, const IRCToken &realname )
  {};
			       
  
  virtual void onNicknameInUse( const IRCToken &nicknameInUse ) 
  {};

  virtual void onNicknameCollision( const IRCToken &collidedNick ) 
  {};

  /**
   * called when a client has quit irc 
   * @param irc user quit message
   */ 
  virtual void onQuit(const IRCToken& from, 
		      const IRCToken* msg, int tokenNum )
  {};

  /**
   * called when a KILL event has been received
   * @param target who has been killed
   * @param cause why target has been killed
   */
  virtual void onKill( const IRCToken& from,
		       const IRCToken &target, 
		       const IRCToken *cause,
		       int numArgs )
  {};

public:
  void printPong( const IRCToken& msg );
  
  /**
   * prints a message
   * @param sender virtual sender of the message
   * @param target nickname of the message receiver
   * @param msg message text
   */
  void printMsg( const IRCToken &sender, const IRCToken &target, 
		 const IRCToken &msg);
  void printMsg( const IRCToken &sender, 
		 const IRCToken &target, 
		 const unsigned char* msg, ... );

  /**
   * print a notice
   * @param sender virtual sender of the message
   * @param target nickname of the message receiver
   * @param msg message text
   */
  void printNotice( const IRCToken &sender, const IRCToken &target, const IRCToken &msg);
  void printNotice( const IRCToken &sender, const IRCToken &target, const unsigned char *msg, ...);
  void printNotice( const IRCToken& sender, const IRCToken& target, const char* format, ... );

  /**
   * sets the mode of a user
   * @param nick nickname of the user which changes his modes. 
   * @param modes new modes of the user
   */
  void printSetUserMode( const IRCToken &nick, const IRCToken &modes );

  void printSquit( const IRCToken &nick, const IRCToken &cause );
  void printSquit( const IRCToken &nick, const unsigned char* format, ... );

  void printKill( const IRCToken &from, const IRCToken &nickname, const IRCToken &cause );
  void printKill( const IRCToken &from, const IRCToken &nickname, const unsigned char*  format, ... );

  void printQuit( const IRCToken &from, const IRCToken &cause );
  void printQuit( const IRCToken &from, const unsigned char* format, ... );

  void servprintf( const unsigned char* format, ... );

private:
  // not implemented
  IRCServerConnection();
  IRCServerConnection(const IRCServerConnection &);
  const IRCServerConnection & operator=(const IRCServerConnection &);


  const IRCToken password;
  const IRCToken hostname;
  IRCToken myDescription;


  virtual void onDisconnected();
  virtual void onMessage( const IRCServerMessage& msg );
  void real_servprintf( const unsigned char *format, va_list &ap );

  // these functions will call the virtual hook functions
  void cmdPingHandler( const IRCServerMessage& msg );
  // nick introduces
  void cmdServerNickHandler( const IRCServerMessage& msg );
  // nick changes
  void cmdClientNickHandler( const IRCServerMessage& msg );
  void cmdPrivmsgHandler( const IRCServerMessage& msg );
  void cmdNoticeHandler( const IRCServerMessage& msg );
  void cmdModeHandler( const IRCServerMessage& msg );
  void cmdQuitHandler( const IRCServerMessage& msg );
  void cmdKillHandler( const IRCServerMessage& msg );
  void cmd433Handler( const IRCServerMessage& msg ); // nickname in use
  void cmd436Handler( const IRCServerMessage& msg ); // nickname collision
  void cmdErrorHandler( const IRCServerMessage& msg );
  void cmdWhoisHandler( const IRCServerMessage& msg );
#if SERVER_TYPE == SERVER_HYBRID || SERVER_TYPE == SERVER_BAHAMUT
  void cmdSvinfoHandler( const IRCServerMessage& msg );
#elif SERVER_TYPE == SERVER_UNREAL
  void cmdNetinfoHandler( const IRCServerMessage& msg );
#endif

  struct CmdHandler
  {
    typedef void (IRCServerConnection::*handler_type)( const IRCServerMessage& msg );
    
    const IRCToken     name;
    handler_type const handler;
  };

  static const CmdHandler servMsgHandlers[];
  static const CmdHandler clientMsgHandlers[];
};



#endif
