/***************************************************************************
                           irctoken.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: irctoken.hpp,v 1.12 2002/05/26 20:21:23 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef IRCTOKEN_HPP
#define IRCTOKEN_HPP

#include <iostream>
#include <string>

/**
 * performance optimized string class
 * - substring 8 times faster than the stl implementation
 * - creation 3 times faster ( even with memory authorative flag set )
 * 
 * performance is gained by using the 'memory authorative' flag.
 * Most operations on hostserv do not need any memory copying 
 * as they stimulate an operation only but are not mandatory for
 * it. E.G. the 'auth' command.
 * this class reduces the allocation traffic as most IRCToken operate
 * on the buffers. I nice side effect is that memory consumption will
 * be reduced.
 * 
 */ 
// FIXME: ! implement reference counting (which should result in the
//          final bootup imho) !
class IRCToken {
public:
  typedef unsigned int   size_type;
  typedef unsigned char  data_type;
  typedef unsigned char* reference_type;

  IRCToken();
  IRCToken( const char* msg, bool authorative = false );
  IRCToken( const std::string& str );
  IRCToken( const reference_type msg );
  IRCToken( const IRCToken&, bool authorative = true );
  virtual ~IRCToken()
  { cleanup(); }
      
  static const size_type npos = ~0;

  bool equalsIgnoreCase( const IRCToken& tok ) const;

  IRCToken substr( size_type begin ) const;
  IRCToken substr( size_type begin, size_type len ) const;
  data_type operator[]( const size_type ) const;

  const reference_type c_str() const
  { return m_strbuf; }
  reference_type c_str()
  { return m_strbuf; }

  size_type length() const
  { return m_length; };
  size_type size() const
  { return m_length; };

  bool empty() const
  { return m_length == 0; }
  
  /**
   * @param str a 0-terminated c string
   */
  const IRCToken& operator=( const reference_type str );
  const IRCToken& operator=( const IRCToken& tok );
  const IRCToken& operator=( const std::string& str );
  const IRCToken& operator=( const char* msg );

  /**
   * appends a null-terminated string 
   */
  const IRCToken& operator+=( const reference_type str );
  const IRCToken& operator+=( const IRCToken& tok );
  const IRCToken& operator+=( const std::string& str );

  bool operator==( const IRCToken& ) const ;
  bool operator==( const reference_type str ) const ;
  bool operator==( const std::string& str ) const;
  bool operator==( const char* ) const;

  bool operator!=( const IRCToken& tok ) const
  { return !this->operator==( tok ); }
  bool operator!=( const reference_type str ) const
  { return !this->operator==( str ); }
  bool operator!=( const std::string& str ) const
  { return !this->operator==( str ); }
  bool operator!=( const char* str ) const
  { return !this->operator==( str ); }


  IRCToken operator+( const reference_type str );
  IRCToken operator+( const IRCToken& tok );
  IRCToken operator+( const std::string& str );

  bool operator<( const IRCToken& ) const;
  bool operator<( const reference_type str ) const;
  bool operator<( const std::string& str ) const;

  bool operator>( const IRCToken& ) const;
  bool operator>( const reference_type str ) const;
  bool operator>( const std::string& str ) const;

      
  /**
   * this function will set the intern string reference 
   * to <tt>buf</tt>. If the string is authorative for 
   * a string on the heap the string will be removed.
   * IRCToken is not authorative for <tt>buf</tt>
   * The string must not be terminated.
   * 
   * @param buf a char array ( must not be terminated by a '\0'
   * @length the length of <tt>buf</tt>
   */
  const IRCToken& assign( const IRCToken& buf );
  const IRCToken& assign( const reference_type, const size_type len );

  // FIXME
  const IRCToken& erase( size_type from );

  void collapse();

protected:
  const IRCToken& assign( const reference_type buf, 
			  const size_type length,
			  bool authorative );

  /**
   * optimized strcpy function
   */
  static void fast_strcpy( reference_type dest,
			   reference_type src,
			   size_type len );
  /**
   * optimized strlen function
   */
  static size_type strlen( const unsigned char* str );
  IRCToken( const reference_type buf, const size_type len, bool authorative );
  /**
   * used by += operators
   */
  void strcat( const reference_type addChars, size_type len );
  
  static int strcmp( const reference_type l, const reference_type r,
		     size_type len );

private:
  /**
   * if IRCToken is memory authorative memory will be freed
   * after the call m_authorative is always false;
   */
  void cleanup();

  /**
   * used by copy, substr, etc. whenever a new (temporary) object
   * needs to be created
   */

  reference_type m_strbuf;
  size_type      m_length;
  bool           m_authorative;
};

inline std::ostream& operator<<( std::ostream& out, const IRCToken& tok )
{ return out.write( (char*)tok.c_str(), tok.length() ); }


#endif
