#ifndef METHOD_HPP
#define METHOD_HPP

/***************************************************************************
                           method.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: method.hpp,v 1.16 2002/05/23 18:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "util.hpp"
#include "response.hpp"
#include "responsestatusline.hpp"
#include "fileresponsesetting.hpp"
#include "config.h"


namespace httpd
{
  class Response;
  class Request;
  class HEADRequest;
  class GETRequest;
  class OPTIONSRequest;
  class POSTRequest;
  class PUTRequest;
  class DELETERequest;
  class CONNECTRequest;
  class TRACERequest;
  class IncompleteRequest;
  class IllegalRequest;
  class NotImplementedRequest;


  class Method
  {
  public:
    void processRequest()
    { m_error = !onProcessRequest(); }

    virtual const Response* getResponse() const = 0;

    bool encounteredError() const
    { return m_error; }

    virtual ~Method()
    {}

  protected:
    Method()
      : m_error( false )
    {}

    /**
     * actual process making request handler
     * gets called by 'processRequest'. The return value
     * of this function gets evaluated. 
     * @return <u>false</u>: error encountered<br><u>true</u>: flawless run
     */
    virtual bool onProcessRequest() = 0;
    

  private:
    Method( const Method& );

    bool m_error;
  };

  template<class requestType>
  class GenericMethod : public Method
  {
  public:
    typedef requestType              request_type;
    typedef const requestType* const const_pointer_type;

    GenericMethod( const_pointer_type request )
      : m_request( request ),
	m_response( NULL )
    {}

    virtual ~GenericMethod()
    { delete m_response; }

    virtual const Response* getResponse() const
    { return m_response; }

  protected:
    const_pointer_type m_request;
    Response*          m_response;

  private:
    GenericMethod();
    GenericMethod( const GenericMethod& );
  };

  class UnimplementedMethod : public GenericMethod<Request>
  {
  public:
    UnimplementedMethod( Request* req )
      : GenericMethod<Request>( req )
    {}

    virtual ~UnimplementedMethod()
    {}

  protected:
    virtual bool onProcessRequest();
  };

  class OPTIONSMethod : public GenericMethod<OPTIONSRequest>
  {
  public:
    OPTIONSMethod( const OPTIONSRequest* req )
      : GenericMethod<OPTIONSRequest>( req )
    {}

  protected:
    virtual bool onProcessRequest()
    { return true; }

  private:
    OPTIONSMethod();
    OPTIONSMethod( const OPTIONSMethod& );
  };

  // forward declaration for POSTMethod
  class CGIScript;
   class POSTCGIScriptHandlerKit;
  
  class POSTMethod : public Method
  {
  public:
    POSTMethod( const POSTRequest* req )
       : Method(),
	 m_request( req ),
	 m_script( NULL )
    { 
#ifdef DEBUG
       cout << "POSTMethod] created" << endl;
#endif
       makeHandlerKit();
    }

    virtual ~POSTMethod();

    const Response* getResponse() const 
    { return m_response; }

  protected:
    virtual bool onProcessRequest();
    
  private:
    POSTMethod();
    POSTMethod( const POSTMethod& );

    static void makeHandlerKit();

    const POSTRequest* const m_request;

    /*
     * will get intialized by onProcessRequest
     */
    static POSTCGIScriptHandlerKit* m_handlerKit;
    CGIScript*               m_script;

   /**
    * m_script is memory authorative for m_response
    */
   Response*  m_response;
  };

  class PUTMethod : public Method
  {
  public:
    PUTMethod( const PUTRequest* );

  protected:
    virtual bool onProcessRequest()
    { return true; }

  private:
    PUTMethod();
    PUTMethod( const PUTMethod& );
  };
  
  class DELETEMethod : public Method
  {
  public:
    DELETEMethod( const DELETERequest* );

  protected:
    virtual bool onProcessRequest()
    { return true; }

  private:
    DELETEMethod();
    DELETEMethod( const DELETEMethod& );
  };
  
  class TRACEMethod : public Method
  {
  public:
    TRACEMethod( const TRACERequest* );

  protected:
    virtual bool onProcessRequest()
    { return true; }

  private:
    TRACEMethod();
    TRACEMethod( const TRACEMethod& );
  };

  class CONNECTMethod : public Method
  {
  public:
    CONNECTMethod( const CONNECTRequest* );

  protected:
    virtual bool onProcessRequest()
    { return true; }

  private:
    CONNECTMethod();
    CONNECTMethod( const CONNECTMethod& );
  };


  class HEADMethod : public GenericMethod<HEADRequest>
  {
  public:
    HEADMethod( const HEADRequest* req )
      : GenericMethod<HEADRequest>( req )
    {}

  protected:
    virtual bool onProcessRequest();

    virtual bool onSuccessfullRequest( 
				      const http::HTTPString& m_requestDocument,
				      const util::FileStats& stats,
				      const FileResponseSetting& responseSetting )
    { return true; }


  private:
    HEADMethod();
    HEADMethod( const HEADMethod& );

    void addFileDocumentHeaders( const http::HTTPString& filename,
				 const util::FileStats& stats,
				 const FileResponseSetting& responseSetting,
				 const bool defaultSetting );
    void addHTTP1xDocumentHeaders( const http::HTTPString& filename,
				   const util::FileStats& stats,
				   const FileResponseSetting& responseSetting,
				   const bool defaultSetting );

  };


  class GETMethod : public HEADMethod
  {
  public:
    GETMethod( const GETRequest* req )
      : HEADMethod( (HEADRequest*) req ),
	m_responseBody( NULL )
    {
#ifdef VERBOSE_MEMORY
      cout << "GETMethod] created" << endl;
#endif
    }

    virtual ~GETMethod()
    { 
      if( m_responseBody )
	delete m_responseBody;
#ifdef VERBOSE_MEMORY
      cout << "GETMethod] destroyed" << endl;
#endif
    }

  protected:
    virtual bool onSuccessfullRequest( const http::HTTPString& m_requestDocument,
 				       const util::FileStats& stats,
 				       const FileResponseSetting& responseSetting );


  private:
    GETMethod();
    GETMethod( const GETMethod& );

    FileResponseBody* m_responseBody;
  };

  class IncompleteMethod : public GenericMethod<Request>
  {
  public:
    IncompleteMethod( const Request* req )
      : GenericMethod<Request>( req )
    {
#ifdef DEBUG
      cout << "IncompleteMethod] created" << endl;
#endif
    }

#ifdef DEBUG
    virtual ~IncompleteMethod()
    {
      cout << "IncompleteMethod] destroyed" << endl;
    }
#endif

  protected:
    virtual bool onProcessRequest();
  };

  class IllegalMethod : public GenericMethod<IllegalRequest>
  {
  public:
    IllegalMethod( const IllegalRequest* req )
      : GenericMethod<IllegalRequest>( req )
    {
#ifdef DEBUG
      cout << "IllegalMethod] created" << endl;
#endif
    }

#ifdef DEBUG
    virtual ~IllegalMethod() 
    {
      cout << "IllegalMethod] destroyed" << endl;
    }
#endif

  protected:
    virtual bool onProcessRequest();

  private:
    IllegalMethod();
    IllegalMethod( const IllegalMethod& );
  };

  class NotImplementedMethod : public GenericMethod<NotImplementedRequest>
  {
    public:
    NotImplementedMethod( const NotImplementedRequest* req )
      : GenericMethod<NotImplementedRequest>( req )
    {}

    virtual ~NotImplementedMethod()
    {}
    
  protected:
    virtual bool onProcessRequest();

  private:
    NotImplementedMethod();
    NotImplementedMethod( const NotImplementedMethod& );
  };

}

#endif
