/***************************************************************************
                           offsetvector.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: hostserv.hpp,v 1.32 2002/03/19 10:58:10 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef OFFSETVECTOR_HPP
#define OFFSETVECTOR_HPP


#include <vector>

template<typename T>
class basic_offsetVector : protected vector<vector<T>*>
{
public:
  typedef vector<T>*   data_type;
  typedef data_type    junk_type;
  typedef data_type    reference_type;
  typedef T            junk_data_type;
  typedef T&           junk_reference_type;
  typedef unsigned int id_type;
  using data_type::push_back;

  basic_offsetVector( int junkSize, int numJunks )
    : vector<vector<T>*>( numJunks ),
      m_junkSize( junkSize ),
      m_useableJunk( NULL ),
      m_lastUseableJunk( NULL ),
      m_full( false ),
      m_numChunksInUse( 0 ),
      m_junkElemNum( 0 )
  {
    assert( numJunks > 0 );
    assert( junkSize > 0 );

    initJunks( numJunks );
  }
    
  virtual ~basic_offsetVector()
  {
    destroyJunks();
  }

  junk_reference_type operator[] ( id_type idx )
  { 
    if( index > size() ) 
      return false;
    
    const int junkNum = index / m_junkSize;
    const int elemNum = index % m_junkSize;
    
    reference_type junk = vector<junk_type>::operator[]( junkNum );
    
    return (*junk)[elemNum];
  }

  /**
   * @return valid object identifier
   */    
  id_type push_back( junk_reference_type obj )
  {
    // last junk in the vector is full
    // -> resize
    if( m_full )
      m_useableJunk = resizeFullOffsetVector();
    
    m_useableJunk->push_back( elem );
    
    // check junk is full
    if( m_useableJunk->size() == m_junkSize ) {
      // check whether this buffer is full
      if( m_useableJunk == m_lastUseableJunk )
	full = true;
      else {
	// update useablejunk and number of junks in use
	m_useableJunk = (*this)[m_numChunksInUse++];
      }
    }
    
    return m_junkElemNum++;
  }

  /**
   * this function can only be used to <b>overwrite</b>
   * already inserted elements.
   */
  bool insert( id_type idx, junk_reference_type junkElem )
  {
    if( idx > size() )
      return false;
    
    const int junkNum = idx / m_junkSize;
    const int elemNum = idx % m_junkSize;
    
    reference_type junk = vector<junk_type>::operator[]( junkNum );
    
    junk->insert( elemNum, junkElem );
    
    return true;
  }

  /**
   * removes last inserted element
   */
  junk_data_type pop_back()
  {
    if( m_useableJunk.empty() )
      {
	// we can only pop_back() if the vector has at least 1 element
	assert( m_numJunksInUse > 0 );
	
	m_useableJunk = vector<junk_type>::operator[]( ( --m_numChunksInUse ) - 1 );
      }
    
    --m_junkElemNum;
    
    return m_useableJunk.pop_back();
  }

  id_type size()
  { return m_junkElemNum; }

  bool empty()
  { return m_junkElemNum == 0; }

  id_type getNextID() const
  { return m_junkElemNum; }
    
private:
  basic_offsetVector();
  basic_offsetVector( const basic_offsetVector& );
  const basic_offsetVector& operator=(const basic_offsetVector& );

  void initJunks( int numJunks )
  {
    while( numJunks-- )
      push_back( new junk_type( NULL, m_junkSize ) );
    
    m_useableJunk = vector<junk_type>::front();
  }

  void destroyJunks()
  {
    junk_type beg = vector<junk_type>::front(),
              last = vector<junk_type>::back();
      
    do {
      delete beg++;
    } while ( beg != last );
  }

  junk_reference_type getElem( id_type idx );

  id_type insertElem( junk_reference_type );

  /**
   * @return next free junk
   */
  reference_type resizeFullOffsetVector()
  {
#ifdef DEBUG
    cout << "VHostoffsetVector():: resizing" << endl;
#endif
    
    // grow by 1/4 only. 
    // ( more than enough imho, but as we are only moving some 
    // pointers around on resizing ... )
    const int plusChunks = m_numChunksInUse / 4;
    
    vector< junk_type >::resize( m_numChunksInUse + plusChunks );
    
    for( int i = 0; i < plusChunks; ++i )
	vector<junk_type>::push_back( new junk_type( NULL, m_chunkSize ) );
	
    m_full = false;
    
    return (*this)[m_numChunksInUse++];
  }
    
  const int m_junkSize;
  // gets updated by insertElem()
  reference_type m_useableJunk; // a junk where we can safely insert new elems

  // gets updated by resizeFullOffsetVector()
  reference_type m_lastUseableJunk; // last junk where we can safely insert new elems

  // is set to true by insertElem() and set to false by 'resizeFullOffsetVector'
  bool m_full;

  // gets updated by insertElem()
  int m_numChunksInUse;

  // contains the number of inserted elems
  // gets incremented by insertElem; 
  id_type m_junkElemNum;
};

  
#endif
