/***************************************************************************
                           poller.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: poller.hpp,v 1.10 2002/05/19 11:56:03 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef POLLER_HPP
#define POLLER_HPP

extern "C" {
#include <poll.h>
}
#include <iostream>
#include "pollable.hpp"
#include "pollerfd.hpp"
#include "config.h"

/**
 * this class is a wrapper around the 'poll' 'select', etc. 
 * system calls
 * it manages arrays of clients. These arrays can be managed
 * via 'addTCPSocket' and 'removeTCPSocket'.
 *
 */
class Poller {
public:

  /**
   * @param maxCapacity maximum number of files to monitor
   *        - Warning: this number cannot be changed afterwards
   */
  Poller( int maxCapacity );
  virtual ~Poller();

  /**
   * runs function 'poll()' in an endless loop
   *
   * @param timeout sets the amout of time until function 'poll()' will return
   *        and the function terminates
   * @return -2 : no client in list of polling sockets
   *         -1 : polling error
   *          0 : no IO while 'timeout'
   *         >0 : number of sockets which have been affected
   */
  int runSyscallPoll( int timeout = -1 );

  /**
   * adds a tcp socket into the intern datatstructures
   * - does <b>not</b> check if <tt>s</tt> is already monitored
   * by poller.
   *
   * @param s TCPSocket which should be monitored by Poller
   * @return NULL - if no slot is free. else the used slot
   */
  PollerFD* addPollableClient( Pollable *s );

  /**
   * makes <tt>s</tt> no more monitored by Poller
   *
   * This function will call 'onRemoved' on a given 
   * TCPSocket object if <tt>s<tt> has been monitored by Poller
   *
   * @return TCPSocket found and removed
   */ 
  bool removePollableClient( const Pollable *s );

  /**
   * unregisters all clients from the queue
   */
  void removeAllClients();

  /**
   * @return the maximum amount of monitoreable files with this 
   * poller instance
   */
  int getMaxCapacity()
  { return maxCapacity; }

  /**
   * @return the actual amount of monitored files by this Poller
   */
  int getClientNum()
  { return m_clientNum; }

  bool isFull()
  { return maxCapacity == m_clientNum; }

protected:
  /**
   * Will be called when a socket has changed it's state
   *
   * @param idx number of changed file descriptors
   */
  virtual void onPolling( int idx );
  
  /**
   * maximum number of files to monitor
   */
  const int maxCapacity;

private:
  /**
   * actual number of monitored files
   */
  int m_clientNum;
  
  /*
   * maximum arrayindice where a used socket can be found
   */
  int m_maxSocketIndex;

  /**
   * array of monitored files
   */
  PollerFD*   pollFDs;

  /** 
   * array of monitored pollable objects
   */
  Pollable** pollSockets;
  
  void findNewMaxSocketIndex();
  void unregisterClient( int index );
  

  Poller();
  Poller( const Poller & );
  const Poller& operator=( const Poller& );
};


#endif
