/***************************************************************************
                           pollerclient.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: pollerclient.hpp,v 1.14 2002/05/09 19:28:54 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef POLLERCLIENT_HPP
#define POLLERCLIENT_HPP

class PollerClient;
#include <ctime>
#include "pollable.hpp"
#include "pollerfd.hpp"
#include "poller.hpp"
#include "config.h"

/**
 * this class gets used by the class 'Poller'
 * Adds a nicer interface to Poller
 * Removes itself from poller on destruction
 */
class PollerClient : public Pollable
{
public:
  typedef PollerFD::event_type event_type;
  friend class Poller;

  PollerClient( Pollable* socket );

  PollerClient( Poller* poller, Pollable* socket,
		bool autoRegister = false );

  virtual ~PollerClient()
  { unregister(); }

  bool registerClient();
  bool registerClient( Poller* poller );
  /**
   * unregisters a client from the poll queue.
   */
  void unregister();

  const Poller* getPoller() const
  { return m_poller; }
  const Pollable* getPollable() const
  { return m_socket; }
  const PollerFD* getPollerFD() const
  { return m_pollerFD; }
  
  void registerEvent( event_type e )
  { m_pollerFD->registerEvent( e ); }
  void unregisterEvent( event_type e )
  { m_pollerFD->unregisterEvent( e ); }

  void unregisterAllEvents()
  { m_pollerFD->unregisterAllEvents(); }

  const File* getPollableFile() const
  { return m_socket->getPollableFile(); }

  bool isRegistered() const
  { return m_pollerFD != NULL; }

protected:
  virtual void onInput() 
  { m_socket->onInput(); }
  virtual void onOutput()
  { m_socket->onOutput(); }
  /**
   * calls 'onError' at the associated socket.
   * calls 'unregister' too
   */
  virtual void onError()
  { m_socket->onError(); }

  /**
   * calls 'onHangup' at the associated socket.
   * calls 'unregister' too
   */
  virtual void onHangup()
  { m_socket->onHangup(); }

  /**
   * calls 'onRemoved' at the associated socket
   * will get called after an successfull 'unregister()' call
   * by the poller
   */
  virtual void onRemoved()
  { m_pollerFD = NULL; }
  
private:
  PollerClient();
  const PollerClient& operator=( const PollerClient& );

  /**
   * can be NULL
   */
  Poller*   m_poller;
  
  /**
   * is never NULL
   */
  Pollable* const m_socket;

  /**
   * whenever this variable is not NULL we can be sure we
   * are registered at m_poller. Means also that whenever
   * this variable is not NULL m_poller is also not null.
   */ 
  PollerFD* m_pollerFD; // pollerFD == NULL signalizes 'm_socket' is not listed at m_poller
};


/**
 * This class saves the time whenever an io event occures.
 */
class MeasuringPollerClient : public PollerClient
{
public:
  /**
   * will insert itself into 'clist' at front.
   */
  MeasuringPollerClient( Poller* poller, Pollable* socket,
			 bool autoRegister = false );

  virtual ~MeasuringPollerClient()
  {}
    
  virtual void onInput();
  virtual void onOutput();

  time_t getLastIOEventPointInTime() const
  { return m_lastIOEvent; }

private:
  MeasuringPollerClient();
  MeasuringPollerClient( const PollerClient& );
  const MeasuringPollerClient& operator=( const MeasuringPollerClient& );

  time_t                      m_lastIOEvent;

  /**
   * calls gettimeofday
   */
  void updateLastIOEvent();
};

#endif
