/***************************************************************************
                           request.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: request.hpp,v 1.12 2002/05/21 19:55:27 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef REQUEST_HPP
#define REQUEST_HPP


#include "httpd.hpp"
#include "http_common.hpp"
#include "requeststartline.hpp"
#include "messageheader.hpp"
#include "method.hpp"
#include "requestbody.hpp"
#include "config.h"


namespace httpd
{
  class HTTPClientConnection;

  /**
   * Before requesting one of the stored values call the appropriate 
   * '*Set' function 
   *<br>
   */
  class Request
  {
  public:
    virtual ~Request()
    {
#ifdef VERBOSE_MEMORY
      cout << "Request] destroyed" << endl;
#endif
    }

    bool requiresResponseBody() const
    { return m_requiresBody; }

    virtual Method* getMethod() = 0;
 
     /*
      * this method may return NULL. therefore always check return value.
      * -> requiresResponseBody() == true and getRequestBody() == NULL
      *    signalizes error
      */
    virtual RequestBody* getRequestBody() = 0;

    virtual const RequestBody* getRequestBody() const = 0;


    bool appendMessageHeader( const http::MessageHeader& msgHeader );
    
    /**
     * this function also affects the responseencoding type
     */
    void setStartline( const http::RequestStartline& startline );

    bool refererSet() const
    { return !m_referer.empty(); }
    bool hostSet() const 
    { return !m_host.empty(); }
    bool userAgentSet() const
    { return !m_userAgent.empty(); }
    bool contentTypeSet() const
    { return !m_contentType.empty(); }

    const http::HTTPString& getReferer() const
    { return m_referer; }
    const http::HTTPString& getHost() const
    { return m_host; }
    const http::HTTPString& getUserAgent() const
    { return m_userAgent; }
    const http::RequestStartline& getRequestStartline() const
    { return m_requestStartline; }
    const http::HTTPString& getContentType() const
    { return m_contentType; }
    const HTTPClientConnection* getFrom() const
    { return m_from; }
    http::httpresponseconnectiontype_type getConnectionType() const
    { return m_connectionType; }
     
    void setReferer( const http::HTTPString& referer )
    { m_referer = referer; }
    void setHost( const http::HTTPString& host )
    { m_host = host; }
    void setUserAgent( const http::HTTPString& userAgent )
    { m_userAgent = userAgent; }
    void setContentType( const http::HTTPString& contentType )
    { m_contentType = contentType; }
    void setConnectionType( http::httpresponseconnectiontype_type cType )
    { m_connectionType = cType; }

  protected:
    /**
     * creates a new Request instance
     * - assigns 'HTTPCONN_UNKNWON' to the connection type value     
     */
    Request( const HTTPClientConnection* from, bool requiresBody );

    const HTTPClientConnection* const m_from;
    const bool                        m_requiresBody;
    http::RequestStartline m_requestStartline;
    http::HTTPString       m_referer;
    http::HTTPString       m_host;
    http::HTTPString       m_userAgent;
    http::HTTPString       m_contentType;
    http::httpresponseconnectiontype_type m_connectionType;

  private:
    Request();
    Request( const Request& );
    Request& operator=( const Request& );

    static const struct HeaderTab {
      typedef void (Request::*valueFunc)(const http::HTTPString& );
      typedef void (Request::*valueOmitFunc)();

    const http::HTTPString fieldName;
	  valueFunc valueHandler;
	  valueOmitFunc valueOmitHandler;
    } MESSAGEHEADERTAB[];


    void setConnectionType( const http::HTTPString& connectionType )
    { 
      m_connectionType =
	http::responseConnectionStringToType( connectionType ); 
    }
  };


  /*
   * warning: derived classes which expect a response body must overwrite
   * getResponseBody()!
   */
  template<class methodType>
  class GenericRequest : public Request
  {
  public:
    typedef methodType  method_type;
    typedef methodType* pointer_type;

    template<class superType>
    GenericRequest( const superType* super, const HTTPClientConnection* from, bool requiresBody )
      : Request( from, requiresBody ),
	m_method( new method_type( super ) )
    {
#ifdef VERBOSE_MEMORY
      cout << "GenericRequest] created(1)" << endl;
#endif
}

    GenericRequest( const HTTPClientConnection* from )
      : Request( from ),
	m_method( NULL )
    {
#ifdef VERBOSE_MEMORY
      cout << "GenericRequest] created(2)" << endl;
#endif
    }
    
    virtual ~GenericRequest()
    { 
       delete m_method; 
#ifdef VERBOSE_MEMORY
    cout << "GenericRequest] destroyed" << endl;
#endif    
    }

    virtual Method* getMethod()
    { return m_method; }

    virtual RequestBody* getRequestBody()
    { return NULL; }

    virtual const RequestBody* getRequestBody() const
    { return NULL; }

  protected:
    pointer_type m_method;

  private:
    GenericRequest( const GenericRequest& );
  };

  class NotImplementedRequest : public GenericRequest<NotImplementedMethod>
  {
  public:
    NotImplementedRequest( const HTTPClientConnection* from )
      : GenericRequest<NotImplementedMethod>( this, from, false )
    {}

    virtual ~NotImplementedRequest()
    {}

  private:
    NotImplementedRequest( const NotImplementedRequest& );
  };

  class UnknownRequest : public NotImplementedRequest
  {
  public:
    UnknownRequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}

  private:
    UnknownRequest( const UnknownRequest& );
  };

  class HEADRequest : public virtual GenericRequest<HEADMethod>
  {
  public:
    HEADRequest( const HTTPClientConnection* from )
      : GenericRequest<HEADMethod>( this, from, false )
    {}

    virtual ~HEADRequest()
    {}

  private:
    HEADRequest( const HEADRequest& );
  };

  class GETRequest : public HEADRequest,
		     public virtual GenericRequest<HEADMethod>
  {
  public:
    GETRequest( const HTTPClientConnection* from )
      : HEADRequest( from ),
	GenericRequest<HEADMethod>( this, from, false )
    {
       delete m_method;
       m_method = new GETMethod( this );
#ifdef VERBOSE_MEMORY
      cout << "GETRequest] created" << endl;
#endif    
    }
    virtual ~GETRequest()
    {
#ifdef VEROBSE_MEMORY
      cout << "GETRequest] destroyed" << endl;
#endif      
    }
  };

  class OPTIONSRequest : public NotImplementedRequest
  {
  public:
    OPTIONSRequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}

  private:
    OPTIONSRequest( const OPTIONSRequest& );
  };

  class POSTRequest : public GenericRequest<POSTMethod>
  {
  public:
    POSTRequest( const HTTPClientConnection* from )
       : GenericRequest<POSTMethod>( this, from, true ),
	 m_requestBody( NULL )
    {}

   /*
    * @return NULL or instance of PlainTextPostRequestBody
    */
    virtual ~POSTRequest()
    { delete m_requestBody; }

    virtual RequestBody* getRequestBody();

    virtual const RequestBody* getRequestBody() const
    { return m_requestBody; }

  private:
    POSTRequest( const POSTRequest& );

    RequestBody* m_requestBody;
  };

  class PUTRequest : public NotImplementedRequest
  {
  public:
    PUTRequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}

  private:
    PUTRequest( const PUTRequest& );
  };

  class DELETERequest : public NotImplementedRequest
  {
  public:
    DELETERequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}

  private:
    DELETERequest( const DELETERequest& );
  };

  class TRACERequest : public NotImplementedRequest
  {
  public:
    TRACERequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}    

  private:
    TRACERequest( const TRACERequest& );
  };

  class CONNECTRequest : public NotImplementedRequest
  {
  public:
    CONNECTRequest( const HTTPClientConnection* from )
      : NotImplementedRequest( from )
    {}

  private:
    CONNECTRequest( const CONNECTRequest& );
  };

  class IncompleteRequest : public GenericRequest<IncompleteMethod>
  {
  public:
    IncompleteRequest( const HTTPClientConnection* from )
      : GenericRequest<IncompleteMethod>( this, from, false )
    {
#ifdef VERBOSE_MEMORY
      cout << "IncompleteRequest] created" << endl;
#endif
    }

    virtual ~IncompleteRequest()
    {
#ifdef VERBOSE_MEMORY
      cout << "IncompleteRequest] destroyed" << endl;
#endif
    }

  private:
    IncompleteRequest( const IncompleteRequest& );
  };

  class IllegalRequest : public GenericRequest<IllegalMethod>
  {
  public:
    IllegalRequest( const HTTPClientConnection* from )
      : GenericRequest<IllegalMethod>( this, from, false )
    {
#ifdef VERBOSE_MEMORY
      cout << "IllegalRequest] destroyed" << endl;
#endif
    }

    virtual ~IllegalRequest()
    {
#ifdef VERBOSE_MEMORY
      cout << "IllegalRequest] created" << endl;
#endif
    }

  private:    
    IllegalRequest( const IllegalRequest& );
  };

  Request* make_request( http::httpmethod_type method,
			 const HTTPClientConnection* from );

}//EON httpd


#endif
