/***************************************************************************
                           requestbody.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: requestbody.hpp,v 1.2 2002/05/25 16:15:36 d3cline Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#ifndef REQUESTBODY_HPP
#define REQUESTBODY_HPP

#include <map>
#include "httptraits.hpp"
#include "filebuf.hpp"
#include "httpstring.hpp"
#include "config.h"


namespace httpd
{
   class RequestBody
   {
      public:
	 /*
	  * reads from f
	  * @return < 0 error
	  *         = 0 requestbody has been read completely
	  *         > 0 successfull read
	  */
	 virtual File::len_type read( FileInBuf& f ) = 0;

	 virtual ~RequestBody()
	 {
#ifdef VERBOSE_MEMORY
	    std::cout << "RequestBody] destroyed" << std::endl;
#endif	    
	 }

      protected:
	 RequestBody()
	 {
#ifdef VERBOSE_MEMORY
	    std::cout << "RequestBody] created" << std::endl;
#endif
	 }
	 RequestBody( const RequestBody& );
	 RequestBody& operator=( const RequestBody& );
   };

   /*
    * Has got a buffer for reading. I
    * Implements read function from RequestBody.
    * Has got an own readbuffer. Reads from FileInBuf
    * an stores data in intern buffer. calls 
    * `addData' then in case of no read error.
    */
   class RequestBodyImpl : public RequestBody
   {
      public:
	 typedef File::len_type len_type;

	 static const len_type MAX_CHUNKSIZE = 2048;


	 virtual ~RequestBodyImpl()
	 { delete[] m_inputBuffer; }


	 /*
	  * in case of no read error addData will get called with appropriate arguments
	  */
	 virtual File::len_type read( FileInBuf& f );

      protected:
	 RequestBodyImpl( len_type bufferLen = MAX_CHUNKSIZE );


	 /**
	  * @param data characters received from client
	  * @param length number of read characters. length is always <= MAX_CHUNKSIZE
	  * @return actually added data. < 0 means error
	  *                              = 0 responsebody has been read completely
	  */
	 virtual len_type addData( HTTPTraits::char_type* data, len_type length ) = 0;

      private:
	 RequestBodyImpl( const RequestBodyImpl& );
	 RequestBodyImpl& operator=( const RequestBodyImpl& );
	 
	 HTTPTraits::char_type* const m_inputBuffer;
	 const len_type               m_inBufLen;
   };

   /*
    * does actually no data processing :-)
    */
   class RequestBodyDummy : public RequestBodyImpl
   {
      public:
	 RequestBodyDummy()
	    : RequestBodyImpl()
	 {
#ifdef VERBOSE_MEMORY
	    std::cout << "RequestBodyDummy] created" << std::endl;
#endif	    
	 }

	 virtual ~RequestBodyDummy()
	 {
#ifdef VERBOSE_MEMORY
	    std::cout << "RequestBodyDummy] destroyed" << std::endl;
#endif	    
	 }

      protected:
	 virtual len_type addData( HTTPTraits::char_type* data, len_type length )
	 { return length; }
   };

   /*
    * implements addData() which is derived from RequestBodyImpl.
    * Reads a line and calls 'onReadline()'. Incomplete lines will 
    * be buffered. The maximum line length for buffered lines is 
    * `MAX_LINELENGTH'. Overflows are catched and '-1' gets returned.
    * This will make 'RequestBodyImpl::read' return -1 which signalizes
    * read error. HTTPClientConnections should catch this and disconnect.
    */
   class LineRequestBody : public RequestBody
   {
      public:
	 typedef File::len_type len_type; 

	 static const len_type MAX_LINELENGTH = RequestBodyImpl::MAX_CHUNKSIZE;

	 virtual ~LineRequestBody()
	 { delete[] m_lineBuffer; }

	 virtual len_type read( FileInBuf& buf );
	 
      protected:
	 LineRequestBody();
	  

	 /**
	  * calls 'onReadLine' for every read line
	  * 
	  * @return process error -> stop processing
	  */
	 virtual bool onInput( HTTPTraits::char_type* data, len_type length );

	 virtual void onReadLine( const HTTPTraits::char_type* data, len_type length ) = 0;

      private:
	 /*
	  * we will use this buffer whenever addData() has been called and received
	  * an incomplete line.
	  */
	 HTTPTraits::char_type* const m_lineBuffer;
	 /*
	  * number of characters in use in m_buffer.
	  * -> if m_usedLen == 0 then we have received a complete line before
	  */
	 len_type                     m_usedLen;
   };

   /*
    * gets used by POSTMethod.
    * reads linebased and parses well-formed entries: <key>=<value>
    * Parsed entries get stored in a 'linemap_type' instance.
    * getDataMap() returns stored data
    */
   class PlainTextPOSTRequestBody : public LineRequestBody
   {
      public:
	 typedef std::map<const http::HTTPString, http::HTTPString> linemap_type;

	 static const HTTPTraits::char_type DIV = '=';

	 PlainTextPOSTRequestBody()
	    : m_dataMap()
	 {}

	 virtual ~PlainTextPOSTRequestBody()
	 {}

	 const linemap_type& getDataMap() const
	 { return m_dataMap; }

      protected:
	 virtual void onReadLine( const HTTPTraits::char_type* data, len_type length );

	 linemap_type m_dataMap;
   };
}


#endif
