/***************************************************************************
                           responsebody.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: responsebody.hpp,v 1.11 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RESPONSEBODY_HPP
#define RESPONSEBODY_HPP

#include <iostream>
#include <strstream>
#include "entitybody.hpp"
#include "http_common.hpp"
#include "hsconfiguration.hpp"
#include "config.h"

// forward declarations for DynamicFileResponseBody
class HSClientList;
class HSIRCClientList;
class VHostList;

namespace httpd
{

/**
 * according to rfc 2616 response-bodies do the connection
 * encoding. so we do here
 */
class ResponseBody 
{
public:
  typedef enum {
    ENCODING_ALL,
    ENCODING_CHUNKED_ONLY,
    ENCODING_IDENTITY_ONLY
  } encoding_type;

  static const int MAX_JUNKSIZE = 2048;

  ResponseBody( int maxChunkSize = MAX_JUNKSIZE )
    : m_maxChunkSize( maxChunkSize )
  {
#ifdef VERBOSE_MEMORY
    cout << "ResponseBody] created" << endl;
#endif
  }

  virtual ~ResponseBody()
  {
#ifdef VERBOSE_MEMORY
    cout << "ResponseBody] destroyed" << endl;
#endif
  }

  /**
   * @return number of printed bytes. 0 if end of response body has been reached
   */
  virtual int printJunk( ostream& out ) = 0;

  virtual encoding_type getSupportedEncoding() const
  { return ENCODING_ALL; }

  int getMaxChunkSize() const
  { return m_maxChunkSize; }

private:
  const int m_maxChunkSize;

  ResponseBody( const ResponseBody& );
  ResponseBody& operator=( const ResponseBody& );
};

/**
 * buffer size is ResponseBody::MAX_JUNKSIZE
 */
class FileResponseBody : public ResponseBody
{
public:
  FileResponseBody( const http::HTTPString& filename )
    : ResponseBody(),
      m_entity( filename ),
      m_buffer( new HTTPTraits::char_type[ResponseBody::getMaxChunkSize()] )
  {}

  virtual ~FileResponseBody()
  { delete[] m_buffer; }

  /**
   * returns whether the file can be read or not
   */
  bool good() const
  { return m_entity.getEntityInputStream().good(); }

  virtual int printJunk( ostream& out )
  {
    // in later revisions we will use self-written buffers to make
    // this junked printing with shared buffers more efficient

    int readLen = fillBuffer();

    /**
     * write will perform nothing whenever readLen == 0 
     */
    out.write( (char*)m_buffer, readLen );
    
#ifdef DEBUG
     cout << "FileResponseBody::printJunk(): printed bytes: " << readLen << endl;
#endif

     return readLen;
  }

protected:
  ReadOnlyFileEntityBody       m_entity;
  HTTPTraits::char_type* const m_buffer;

  int fillBuffer()
  {
    std::istream& in = m_entity.getEntityInputStream();

    in.read( (char*)m_buffer, ResponseBody::getMaxChunkSize() );
    
    return in.gcount();
  }

private:
  FileResponseBody();
  FileResponseBody( const FileResponseBody& );
  FileResponseBody& operator=( const FileResponseBody& );
};

class HTTPd;

class DynamicFileResponseBody : public FileResponseBody
{
public:
  static const HTTPTraits::char_type FORMAT_PREFIX   = '%';
  static const HTTPTraits::char_type FORMAT_OUTQUOTE = '\\';

  DynamicFileResponseBody( const HSConfiguration* config,
			   const HSIRCClientList* ircClients,
			   const HSClientList* hsClients,
			   const VHostList* vhosts,
			   const http::HTTPString& filename );

  virtual ~DynamicFileResponseBody()
  { delete[] m_streamBuffer; }


  virtual int printJunk( ostream& out );

private:
  DynamicFileResponseBody();
  DynamicFileResponseBody( const DynamicFileResponseBody& );
  DynamicFileResponseBody& operator=( const DynamicFileResponseBody& );

  typedef void (DynamicFileResponseBody::*printfunc_type)();

  static const printfunc_type PRINTFUNC_TAB[];
  
  const HTTPTraits::char_type* const m_streamBuffer;
  std::ostrstream                    m_sout;
  const HTTPd* const                 m_httpd;
  const HSConfiguration* const       m_configuration;
  const HSIRCClientList* const       m_ircclients;
  const HSClientList* const          m_hsclients;
  const VHostList* const             m_vhosts;

  /**
   * - is a pointer into m_streamBuffer's memory.
   * - marks process position of printJunk<br>
   *
   * whenever this member is NOT NULL we assume the buffer
   * has not been processed completely now.
   * we need this pointer as a member variable as the algorithm
   * does not guarantee that the whole derived buffer content will be
   * processed by every 'printJunk()' call. Therefore we have to store
   * the last process position to introduce printJunk() resume capabilities
   */
  HTTPTraits::char_type*       m_lastReadPos;
  /**
   * together with m_lastReadMax points fill boundary of m_streamBuffer.
   * -> [m_streamBuffer, m_lastReadMax) build partition
   */
  const HTTPTraits::char_type* m_lastReadMax;

  /**
   * true whenever the last character is 'FORMAT_OUTQUOTE'
   */
  bool                         m_lastCharOutquote;


  /**
   * this function does all the dynamic printing.
   * 'num' is an index into an array of member functions
   * (->PRINTFUNC_TAB)
   */
  void print( int num );

  void printIgnore()
  {}

  /** 
   * OPCode: 00
   * Prints HostServ's servername in the IRC network
   */
  void printServerName()
  { m_sout << m_configuration->getServerName(); }
  /**
   * OPCode: 01
   * Prints HostServ's serverdescription
   */
  void printServerDescription()
  { m_sout << m_configuration->getServerDescription(); }
  /**
   * OPCode: 02
   * Prints the hostname of the irc server which hostserv connected
   */
  void printRemoteHost()
  { m_sout << m_configuration->getRemoteHost(); }
  /**
   * OPCode: 03
   * Prints the password to connect remote irc server
   */
 void printRemotePassword()
 { m_sout << m_configuration->getRemotePassword(); }

  /**
   * OPCode: 04
   * Prints the nickname of HostServ in the IRC Network
   */
  void printHostservNickname()
  { m_sout << m_configuration->getHostservNickname(); }
  /**
   * OPCode: 05
   * Prints the username of HostServ in the IRC Network
   */
  void printHostservUsername()
  { m_sout << m_configuration->getHostservUsername(); }
  /**
   * OPCode: 06
   * Prints the address of HostServ in the IRC Network
   */
  void printHostservAddress()
  { m_sout << m_configuration->getHostservAddress(); }
  /**
   * OPCode: 07
   * Prints the realname of HostServ in the IRC Network
   */
  void printHostservRealname()
  { m_sout << m_configuration->getHostservRealname(); }
  /**
   * OPCode: 08
   * Prints HostServ's usermodes as mentioned in configurationfile
   */
  void printHostservModes()
  { m_sout << m_configuration->getHostservModes(); }
  /**
   * OPCode: 09
   * Prints the name of the channel hostserv has joined
   */
  void printHostservChannel()
  { m_sout << m_configuration->getHostservChannel(); }
  /**
   * OPCode: 0A
   * Prints the default permission string for newly created clients
   */
  void printClientDefaultPermissions()
  { m_sout << m_configuration->getClientDefaultPermissions(); }
  /**
   * OPCode: 0B
   * Prints the public policy for newly created vhosts
   */
  void printVHostPublicPolicy()
  { m_sout << m_configuration->getVHostPublicpolicy(); }
  /**
   * OPCode: 0C
   * Prints the name of the hostserv configurationfile inclusive
   * path
   */
  void printHostservConfigFilename()
  { m_sout << m_configuration->getHostservConfigfilename(); }
  /**
   * OPCode: 0D
   * Prints the name of the vhost dbfile
   */
  void printVHostsConfigFilename()
  { m_sout << m_configuration->getVHostsConfigfilename(); }
  /**
   * OPCode: 0E
   * Prints the name of the client dbfile
   */
  void printUsersConfigfilename()
  { m_sout << m_configuration->getUsersConfigfilename(); }
  /**
   * OPCode: 0F
   * Prints the document root of the httpd server
   */
  void printHTTPdDocumentRoot()
  { m_sout << m_configuration->getHTTPdDocumentRoot(); }
  /**
   * OPCode: 10
   * Prints the name of document which will get read whenever
   * no filename is explicitely mentioned in client's request
   * ( -> http://<domain>/ )
   */
  void printHTTPdDefaultDocument()
  { m_sout << m_configuration->getHTTPdDefaultDocument(); }
  /**
   * OPCode: 11
   * Prints the name / ip of the interface on which hostserv
   * is listening for incoming http client connections
   */
  void printHTTPdBindInterface()
  { m_sout << m_configuration->getHTTPdBindInterface(); }
  /**
   * OPCode: 12
   * Prints the port number on which hostserv's http server
   * is listening for incoming connections
   */
  void printHTTPdPort()
  { m_sout << m_configuration->getHTTPdPort(); }
  /**
   * OPCode: 13
   * Prints the maximum amount of time a client is permitted
   * to input nothing until it gets disconnected
   */
  void printHTTPdMaxClientTimeout()
  { m_sout << m_configuration->getHTTPdMaxClientTimeout(); }
  /**
   * OPCode: 14
   * Prints the maximum number of simultaneously served http 
   * clients
   */
  void printHTTPdMaxClientNum()
  { m_sout << m_configuration->getHTTPdMaxClientNum(); }

  /**
   * Opcode: 30
   * Prints number of current clients in irc network
   */ 
  void printIRCClientNum();

  /**
   * Opcode: 31
   * Prints number of clients in db
   */
  void printHSClientNum();

  /**
   * Opcode: 32
   * Prints number of vhosts in db
   */
  void printVHostNum();
};

}//EON httpd
#endif
