/***************************************************************************
                           file.hpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: util.hpp,v 1.9 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef UTIL_HPP
#define UTIL_HPP

#include <string>
#include <sstream>
#include <ctime>
#include <iostream>

extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
}

#include "httptraits.hpp"


namespace util
{

template<typename T, typename K>
bool equalsIgnoreCase( const std::basic_string<T,K>& lhs,
		       const std::basic_string<T,K>& rhs )
{
  typedef typename std::basic_string<T,K> string_type;
 
  if( lhs.length() == rhs.length() )
  {
    string_type::const_iterator                  rIter = rhs.begin();
    string_type::const_iterator       lIter = lhs.begin();
    string_type::const_iterator lEnd  = lhs.end();

    while( lIter < lEnd ) {
      if( std::tolower( *lIter++ ) != std::tolower( *rIter++ ) ) {
	return false;
      }
    }

    return true;
  }
  return false;
}

/**
 * counts the number of directory steps.
 * Whenever the returned value is negative then
 * path steps <b>up</b> in directory hierarchy.<br>
 * <u>Example</u>: '../index.html'
 * Whenever the returned value is greater than 0
 * then path steps down in directory hierarchy.<br>
 * <u<>Example</u>: '/a_directory/index.html'
 * If the return value is '0' no directory is part of the path.
 * <br><br>
 * All what this function does is to count the number of '/'.
 * Means whenever a '/' is part of filename then this function
 * will <b>_NOT_</b> work correctly.
 */
template<typename T, typename K>
int countDirectoryJumps( const std::basic_string<T,K>& pathToFile )
{
  typedef typename std::basic_string<T,K> string_type;
  typedef typename string_type::value_type character_type;

  //     '.' -.   '..' -.   <any char> -.  '/' - .
  //          v         v               v        v
  enum { LAST_DOT = 0, LAST_DIR_DOWN = 1, LAST_FILE = 2, LAST_DIR = 3 }
                                    lastRead = LAST_DIR;
  int                               dirJumps = 0;
  string_type::const_iterator       iter     = pathToFile.begin();
  const string_type::const_iterator end      = pathToFile.end();
  character_type                    dif      = (character_type)'/';
  character_type                    dot      = (character_type)'.';


  for( ; iter != end; ++iter )
  {
    if( *iter == dif )
    {
      /*
       * '/'
       */
      switch( lastRead )
      {
      case LAST_DOT:
	// let's optmize it by the compiler
	break;
      case LAST_DIR_DOWN:
	// previous 3 chars are "../"
	// therefore we step up in hierarchy
	--dirJumps;
	break;

      case LAST_FILE:
	// previous 4 chars are ".../" or
	// "abc/". ( abc ^= any 3 char string )
	// Therefore we step down in hierarchy
	++dirJumps;
	break;

      case LAST_DIR:
	break;
      }
      lastRead = LAST_DIR;
    }
    else if( *iter == dot )
    {
      /*
       * '.'
       */
      switch( lastRead )
      {
      case LAST_DOT:
	// previous char is: '.': -> mark it as a '..' substring
	lastRead= LAST_DIR_DOWN;
	break;

      case LAST_DIR_DOWN:
	// we get here whenever we have something like that:
	//  '...' or more dots. We will treat that as a filename
	lastRead = LAST_FILE;
	break;

      case LAST_FILE:
	lastRead = LAST_FILE;
	break;

      case LAST_DIR:
	lastRead = LAST_DOT;
	break;

      }//eosw
    }
    else
      lastRead = LAST_FILE;

  }//eofl

  if( lastRead == LAST_DIR_DOWN )
    --dirJumps;

  return dirJumps;
}

/**
 * counts the number of digits of <num> for number system <system>
 */
int countDigits( register int num, register int system = 10 );

template<class T>
std::string typeToString( const T& type )
{ 
  std::ostringstream stream;
  stream << type;
  return stream.str();
}

template<class T>
void typeToString( std::string& str, const T& type )
{
  std::ostringstream stream;
  stream << type;
  stream.str( str );
}

/**
 * @return actual number of written characters
 */
int numToHexString( register char* cache, 
		    register unsigned int num );

int hexStringToNum( const register HTTPTraits::char_type* cache,
		    register int cacheLen );

/**
 * fills 'target' with a rfc 1123 compliant string
 */
std::string& timeToString( std::string& target, 
			   time_t t = time( NULL ) );

/**
 * This function treats 'day's > 6 exactly the same as if day were 6
 * @param weekday ( a number from 0-6 )
 */
std::ostream& printWeekDay( std::ostream& stream, 
			    unsigned int day );

/**
 * This function treats 'month's > 11 exactly the same as if month were 11 
 * 
 * @param month month ( a number from 0 - 11 )
 */
std::ostream& printMonth( std::ostream& stream,
			  unsigned int month );

class FileStats
{
public:
  typedef off_t filesize_type;
  typedef time_t time_type;


  FileStats()
    : m_stat(),
      m_canRead( false ),
      m_canWrite( false )
  {}

  FileStats( const char* filename )
    : m_stat( new struct stat() ),
      m_canRead( false ),
      m_canWrite( false )
  { update( filename ); }

  virtual ~FileStats()
  { delete m_stat; }

  void setFilename( const char* filename )
  { 
    if( m_stat ) 
      delete m_stat;
    update( filename );
  }

  filesize_type getFileSize() const
  { return m_stat ? m_stat->st_size : 0; }

  time_type getLastModificationTime() const
  { return m_stat ? m_stat->st_mtime : 0; }

  bool exists()
  { return m_stat != NULL; }

  /*
   * untested
   */
  bool canRead()
  { return m_canRead; }
  /*
   * untested
   */
  bool canWrite()
  { return m_canWrite; }
  /*
   * untested
   */
  bool isRegularFile()
  { return m_stat ? S_ISREG(m_stat->st_mode) : false; }
  /*
   * untested
   */
  bool isDirectory()
  { return m_stat ? S_ISDIR(m_stat->st_mode) : false; }
  

private:
  /**
   * can be NULL whenever a stat request failed ( no such file )
   * or a new FileStats object has been created without setting
   * a filename
   */
  struct stat* m_stat;
  bool    m_canRead;
  bool    m_canWrite;

  void update( const char* filename );
};

}
#endif
