/***************************************************************************
                           application.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: application.cpp,v 1.15 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdio>
#include "application.hpp"

using std::cout;
using std::cerr;
using std::endl;
using std::string;

const int Application::SUCCESS = 1;
const int Application::ERROR   = -1;
const int Application::IDLE    = 0;
static    Application* lastApp = NULL;

pid_t _myFork()
{
  return fork();
}

void sigIntHandler( int signum )
{
#ifdef DEBUG
   cout << "Program received sig int" << endl;
#endif

   assert( lastApp );

   lastApp->onSigInt();
}

void sigHupHandler( int signum )
{
#ifdef DEBUG
   cout << "Program received sig hup" << endl;
#endif

   assert( lastApp );

   lastApp->onSigHup();
}

void sigQuitHandler( int signum )
{
#ifdef DEBUG
   cout << "Program received sig quit" << endl;
#endif

   assert( lastApp );
   
   lastApp->onSigQuit();
}

void sigAbrtHandler( int signum )
{
#ifdef DEBUG
   cout << "Program received sig abrt" << endl;
#endif

   assert( lastApp );

   lastApp->onSigAbrt();
}

void sigSegvHandler( int signum )
{
#ifdef DEBUG
   cout << "Program received sig segv" << endl;
#endif

   assert( lastApp );

   lastApp->onSigSegv();
}

Application*
Application::getLastApplication()
{ return lastApp; }

Application::Application( const std::string& name, 
			  char** args,
			  int argNum )
   : m_name( name ),
     m_args( args ),
     m_argNum( argNum ),
     m_status( APP_IDLE ),
     m_argMap() 
{
   lastApp = const_cast<Application*>( this );
   m_origStreams[IDX_STDOUT] = cout.rdbuf();
   m_origStreams[IDX_STDERR] = cerr.rdbuf();
}


Application::~Application()
{
   stop();

   exit( 0 );
}

int
Application::start()
{
   if( m_status == APP_RUNNING ) return ERROR;

   m_status = APP_RUNNING;

   onStart();

   m_status = APP_IDLE;

   return SUCCESS;
}


int
Application::stop()
{
   if( m_status == APP_IDLE ) return ERROR;

   onStop();

   m_status = APP_IDLE;

   return SUCCESS;
}

int
Application::kill( signal_type signum )
{
   if( m_status == APP_IDLE )
      return ERROR;
   
   switch( signum )
   {
      case SIG_HUP:
	 onSigHup();
	 break;

      case SIG_INT:
	 onSigInt();
	 break;
	 
      case SIG_QUIT:
	 onSigQuit();
	 break;

      case SIG_ABRT:
	 onSigAbrt();
	 break;

      case SIG_SEGV:
	 onSigSegv();
	 break;

#ifdef DEBUG
      default:
	 cout << "Application::kill(): unknown signal type: " << signum << endl;
	 return ERROR;
#endif	 
   }
   return SUCCESS;
}

int
Application::fork()
{
   pid_t pid = _myFork();

   switch( pid )
   {
      case -1:
	 // error
	 return ERROR;

      case 0:
	 // child
	onForked();
	 break;
	 
      default:
	 // mother
#ifdef DEBUG
	 cout << "Application::fork(): new child created with pid : " << pid << endl;
#endif
 	 onFork();
   }
   
   return SUCCESS;
}

void
Application::redirectConsoleOutputStreams( const char* stdoutFilename,
					   const char* stderrFilename )
{
   m_redirectedStreams[IDX_STDOUT].open( stdoutFilename, std::ios::out | std::ios::trunc );
   m_redirectedStreams[IDX_STDERR].open( stderrFilename, std::ios::out | std::ios::trunc );
   
   if( m_redirectedStreams[IDX_STDOUT].good() )
      cout.rdbuf( m_redirectedStreams[IDX_STDOUT].rdbuf() );
   else
   {
      cerr << "Application::fork(): Could not open " << stdoutFilename << endl;
      cout << "Standard out channel rediraction disabled" << endl;
   }
   
   if( m_redirectedStreams[IDX_STDERR].good() )
      cerr.rdbuf( m_redirectedStreams[IDX_STDERR].rdbuf() );
   else
   {
      cerr << "Application::fork(): Could not open " << stderrFilename << endl;
      cout << "Standard error channel redirection disabled" << endl;
   }
}

void
Application::restoreConsoleOutputStreams()
{
   std::ofstream& fstdout = m_redirectedStreams[IDX_STDOUT];
   std::ofstream& fstderr = m_redirectedStreams[IDX_STDERR];

   if( fstdout.is_open() )
      fstdout.close();
   if( fstderr.is_open() )
      fstderr.close();

   cout.rdbuf( m_origStreams[IDX_STDOUT] );
   cerr.rdbuf( m_origStreams[IDX_STDERR] );
}

void
Application::fillConfigMap()
{
   int   resArgs  = m_argNum;
   char** argIter = m_args;
   string raw;
   CMDLineKey key;
   string value;
   
   for( ; resArgs > 0; --resArgs )
   {
      if( **argIter == '-' ) 
      {
	 bool keyOnly  = true;

	 if( (*argIter)[1] == '-' )
	 { /* gnu style arg */
	    key.setGNUStyle( true );

	    raw = &(*argIter)[2];

	    if( raw.length() == 0 ) continue;

	    string::size_type delimPos = raw.find( '=' );
	    key = raw.substr( 0, delimPos );

	    if( delimPos != string::npos ) 
	    {
	       value = raw.substr( ++delimPos );
	       keyOnly = false;
	    }
	 }
	 else
	 { /* standard style arg */
	    key.setGNUStyle( false );

	    key = &(*argIter)[1];
	    
	    if( key.length() == 0 ) continue;

	    if( --resArgs )
	    {
	       // check next arg is not a new key
	       if( **(++argIter) != '-' )
	       {
		  keyOnly = false;
		  value = *argIter;
	       }
	       else 
	       {
		  // jump back one arg
		  ++resArgs;
		  --argIter;
	       }
	    }

	 }

	 // insert parsed data into map
	 if( keyOnly )
	    value.erase();
#ifdef DEBUG
	 cout << "Application::processArguments(): key = '" << key << "' val = '" << value << '\'' << endl;
#endif
	    m_argMap.insert( std::make_pair( key, value ) );
      }
      ++argIter;
   }
}

bool
Application::writePIDFile( const string& filename )
{
   std::ofstream fileOut( filename.c_str(), std::ios::out | std::ios::trunc );

   if( fileOut.fail() )
      return false;

   fileOut << getPID() << endl;

   cout << "Wrote pid file: " << filename << endl;

   return true;
}

Application::pid_type
Application::readPIDFile( const string& filename )
{
   std::ifstream fileIn( filename.c_str(), std::ios::in );

   if( fileIn.fail() )
      return static_cast<Application::pid_type>( 0 );

   Application::pid_type pid;

   fileIn >> pid;

#ifdef DEBUG
   cout << "PID file found and read: " << pid << endl;
#endif

   return pid;
}

void
Application::erasePIDFile( const string& filename )
{
   unlink( filename.c_str() );
}

void
Application::registerSignal( signal_type signum )
{
  switch( signum )
  {
     case SIG_HUP:
	signal( SIG_HUP, sigHupHandler );
	break;

     case SIG_INT:
	signal( SIG_INT, sigIntHandler );
	break;

     case SIG_QUIT:
	signal( SIG_QUIT, sigQuitHandler );
	break;

     case SIG_ABRT:
	signal( SIG_ABRT, sigAbrtHandler );
	break;

     case SIG_SEGV:
	signal( SIG_SEGV, sigSegvHandler );
	break;

#ifdef DEBUG
     default:
	cout << "Application::registerSignal(): warning: unknown signal type given" << endl;
#endif
  }
}

void
Application::deregisterSignal( signal_type signum )
{
   switch( signum )
   {
      case SIG_HUP:
      case SIG_INT:
      case SIG_QUIT:
      case SIG_ABRT:
      case SIG_SEGV:
	 ::signal( signum, SIG_DFL );
	 break;
#ifdef DEBUG
      default:
	 cout << "Application::deregisterSignal(): warning: unknown signal type given" << endl;
#endif	 
   }
}
