/***************************************************************************
                           filebuf.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: filebuf.cpp,v 1.19 2002/06/02 07:53:59 d3cline Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

extern "C" {
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}
#include <cstdio>
#include "filebuf.hpp"

using std::cout;
using std::cerr;
using std::endl;


/*
 * define it to make this class print
 * what it is printing
 */
// #define VERBOSE_BUFFERCONTENT

FileInBuf::FileInBuf( File& f, len_type len )
   : m_inFd( f ),
     m_readBuffer( new data_type[len] ),
     m_readBufferEnd( m_readBuffer + len ),
     m_readPos( m_readBuffer ),
     m_readMax( m_readBuffer ),
     m_rBufLen( len ),
     m_rStateGood( true )
{}

FileInBuf::~FileInBuf()
{
   delete[] m_readBuffer;
}

FileInBuf::len_type 
FileInBuf::read( data_reference buf,
		 len_type len )
{
   // check stream state _FIRST_
   if( stateEof() )
      return File::eof;
   

   if( readCompletely() )
   {  // buffer is empty
      fillEmptyBuffer();

      if( !stateGood() )
	 return File::eof;
   }

   const len_type        locBufLen  = m_readMax - m_readPos; // maximum number of available chars
   data_reference        rBufIter   = buf; // 'buf' iterator
   const data_reference  readMax    = rBufIter +
      ( ( len >  locBufLen ) ? locBufLen : len ); // read limit
   data_reference        locBufIter = m_readPos; // iterator of local buf

   // copy
   len_type cRead = 0; // number of read bytes
   for( ; rBufIter < readMax; ++cRead )
      *rBufIter++ = *locBufIter++;

   // update read position.
   m_readPos = readMax;

   return cRead;
}

FileInBuf::int_type
FileInBuf::fillEmptyBuffer()
{
   m_readMax = m_readPos = m_readBuffer;

   len_type readBytes = underflow( m_readBuffer, m_rBufLen );
   if( readBytes == File::eof )
   {
      m_rStateGood = false;
      return File::eof;
   }

   m_readMax += readBytes;

   return readBytes;
}

FileInBuf::int_type
FileInBuf::underflow( data_reference bufPos,
		      len_type bytes )
{
   int_type ret = m_inFd.read( bufPos, bytes );
   if( ret < 1 ) 
      return File::eof;
   else 
      return ret;
}

FileInBuf::len_type
FileInBuf::fill()
{
#ifdef DEBUG
   if( stateEof() )
      cerr << "FileInBuf::fill(): 'fill' request but buffer has received 'EOF'" << endl;
#endif

//    if( readCompletely() )
//       return fillEmptyBuffer();
   
   if( !isFull() )
   {
      flush();

      data_reference insertStart = m_readMax;
      const len_type rxLen       = underflow( insertStart, insertStart - m_readBufferEnd );

      if( rxLen == File::eof )
      {
	 m_rStateGood = false;
	 return File::eof;
      }
      else
	 m_readMax += rxLen;

      return rxLen;
   }
   
#ifdef DEBUG
   cerr << "FileInBuf::fill(): useless 'fill' usage: buffer is _full_." << endl;
#endif
   
   return 0;
}

void
FileInBuf::flush()
{
#ifdef DEBUG
   if( stateEof() )
      cerr << "FileInBuf::flush(): warning: flushing buffer while"
	 " stream is in state EOF" << endl;

#endif

   if( readCompletely() )
   {
      m_readPos = m_readMax = m_readBuffer;
   }
   else if( !isFull() && !isEmpty() )
   {
      data_reference srcIter    = m_readPos;
      data_reference srcMax     = m_readMax;
      data_reference destIter   = m_readBuffer;
      len_type       bytesMoved = m_readMax - m_readPos;

      while( srcIter < srcMax )
	 *destIter++ = *srcIter++;

      m_readMax = m_readBuffer + bytesMoved;
      m_readPos = m_readBuffer;
   }
}

// *******************************************************************
// **** FileOutBuf
// *******************************************************************

FileOutBuf::FileOutBuf( File& f, len_type numBytes )
   : std::streambuf(),
     m_outFd( f ),
     m_writeBuffer( new data_type[numBytes] ),
     m_writeBufLen( numBytes )
{
   // as -funsigned-char is used this cast does not matter
   setp( (char*)m_writeBuffer,
	 (char*)m_writeBuffer+numBytes );
}


FileOutBuf::~FileOutBuf()
{
   // write pending data
   doFlush();

   delete[] m_writeBuffer;
}


FileOutBuf::stl_int_type
FileOutBuf::overflow( stl_int_type c )
{
   if( c == File::eof )
      return File::eof;
   
   *pptr() = c;
   pbump( 1 );
   
   if( flushBuffer() == File::eof )
      return File::eof;
   
   return c;
}


FileOutBuf::streamsize 
FileOutBuf::xsputn( const char* buf,
		    streamsize num )
{
   streamsize             bytesLeft     = num; // contains the number of not written characters
   data_reference         rBufIter      = (data_reference)buf; // used to iterate through 'buf'
   data_reference         locBufIter    = (data_reference)pptr();
   const data_reference   locBufIterMax = (data_reference)epptr();
   data_type* const       locBufBase    = (data_reference)pbase();
   int                    ptrDiff       = locBufIterMax - locBufIter;
   const int              writeBufLen   = locBufIterMax - locBufBase;

   while( bytesLeft > ptrDiff )
   {
      /*
       * ok the local buffer is too small to handle the write request
       *
       * copy bytes from bufIter into the writebuffer as
       * long as more bytes then bufferspace is available
       * and write them
       */
      
      // copy the characters
      // --
      // i do not use memcpy because this
      // loop seems to be copying faster
      while( locBufIter < locBufIterMax )
	 *locBufIter++ = *rBufIter++;

      
      // update buffer pointers
      pbump( ptrDiff );

      // write filled buffer and update pointers
      // note: pptr() == pbase() after calling
      //       doFlush()
      doFlush();

      // update remaining characters in 'buf'
      bytesLeft -= ptrDiff;
      
      // reset the local buf iterator
      // as we have printed everything what is in buffer we
      // can safely use the complete buffer...
      locBufIter = locBufBase;

      // update current ptrDiff
      // as the buffer is clean -> use the complete len ...
      ptrDiff    = writeBufLen;
   }

   // advance pptr()
   // note: whenever we get here the buffer is great enough
   //       to contain the remaining characters
   pbump( bytesLeft );

   const streamsize ret = static_cast<streamsize>( bytesLeft );

   // copy remaining characters.
   // they are _always_ less
   // than the buffer could carry
   for( ; bytesLeft; --bytesLeft )
      *locBufIter++ = *rBufIter++;

   return ret;
}


FileOutBuf::len_type
FileOutBuf::flushBuffer()
{
   const len_type wLen = pptr() - pbase();

   assert( wLen >= 0 );

   // update pointers
   pbump( -wLen );

   if( m_outFd.write( (unsigned char*)pbase(), wLen ) != wLen )
   {
#ifdef DEBUG
      cout << "FileOutBuf::flushBuffer(): error " << std::flush;
      perror("");
#endif
      return File::eof;
   }

#ifdef VERBOSE_BUFFERCONTENT
   cout << "FileOutBuf: writing: " << endl;
   cout << "-----" << endl;
   cout.write( pbase(), wLen );
   cout << "-----" << endl;
#endif

   return wLen;
}

FileOutBuf::len_type FileOutBuf::doFlush()
{
  return flushBuffer();
}

int
FileOutBuf::sync()
{
   return 0;
}
