/***************************************************************************
                           hostserv.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: hostserv.cpp,v 1.59 2002/05/23 18:15:57 d3cline Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <algorithm>
#include <cstdio> /* for perror */
#include <ctime> /* for time() */
#include "poller.hpp"
#include "pollerclient.hpp"

#include "debug.hpp"
#include "httpd.hpp"
#include "hostserv.hpp"

const string HostServ::APP_NAME     = "hostserv";
const string HostServ::PID_FILENAME = PID_FILE; /* defined in config.h */

const CMDLineKey HostServ::SA_SERVNAME( "name", false );
const CMDLineKey HostServ::LA_SERVNAME( "server-name", true );
const CMDLineKey HostServ::SA_SERVERDESCR( "descr", false ); 
const CMDLineKey HostServ::LA_SERVERDESCR( "server-description", true );
const CMDLineKey HostServ::SA_REMOTEHOST( "rhost", false );
const CMDLineKey HostServ::LA_REMOTEHOST( "remote-host", true );
const CMDLineKey HostServ::SA_REMOTEPASSWORD( "rpassword", false );
const CMDLineKey HostServ::LA_REMOTEPASSWORD( "remote-password", true );
const CMDLineKey HostServ::SA_REMOTEPORT( "rport", false );
const CMDLineKey HostServ::LA_REMOTEPORT( "remote-port", true );
const CMDLineKey HostServ::SA_HSNICKNAME( "nick", false );
const CMDLineKey HostServ::LA_HSNICKNAME( "hostserv-nick", true );
const CMDLineKey HostServ::SA_HSUSERNAME( "username", false );
const CMDLineKey HostServ::LA_HSUSERNAME( "hostserv-username", true );
const CMDLineKey HostServ::SA_HSADDRESS( "address", false );
const CMDLineKey HostServ::LA_HSADDRESS( "hostserv-address", true );
const CMDLineKey HostServ::SA_HSREALNAME( "realname", false );
const CMDLineKey HostServ::LA_HSREALNAME( "hostserv-realname", true );
const CMDLineKey HostServ::SA_HSMODES( "modes", false );
const CMDLineKey HostServ::LA_HSMODES( "hostserv-modes", true );
const CMDLineKey HostServ::SA_HSCHANNEL( "chan", false );
const CMDLineKey HostServ::LA_HSCHANNEL( "hostserv-channel", true );
const CMDLineKey HostServ::SA_CLIENTDEFAULTPERM( "perms", false );
const CMDLineKey HostServ::LA_CLIENTDEFAULTPERM( "client-defaultpermissions", true );
const CMDLineKey HostServ::SA_VHOSTPUBLICPOLICY( "ppolicy", false );
const CMDLineKey HostServ::LA_VHOSTPUBLICPOLICY( "vhosts-defaultpublicpolicy", true );
const CMDLineKey HostServ::SA_VHOSTSFILENAME( "vhostsfile", false );
const CMDLineKey HostServ::LA_VHOSTSFILENAME( "vhosts-configfilename", true );
const CMDLineKey HostServ::SA_USERSFILENAME( "usersfile", false );
const CMDLineKey HostServ::LA_USERSFILENAME( "clients-configfilename", true );
const CMDLineKey HostServ::SA_HSCONFIGFILENAME( "configfile", false );
const CMDLineKey HostServ::LA_HSCONFIGFILENAME( "hostserv-configfilename", true );
const CMDLineKey HostServ::SA_FOREGROUND( "fg", false );
const CMDLineKey HostServ::LA_FOREGROUND( "foreground", true );
const CMDLineKey HostServ::SA_HELP( "h", false );
const CMDLineKey HostServ::LA_HELP( "help", true );
const CMDLineKey HostServ::SA_ADMINLOGIN( "alogin", false );
const CMDLineKey HostServ::LA_ADMINLOGIN( "admin-login", true );
const CMDLineKey HostServ::SA_ADMINPASSWORD( "apassword", false );
const CMDLineKey HostServ::LA_ADMINPASSWORD( "admin-password", true );
const CMDLineKey HostServ::SA_ADMINVHOST( "avhost", false );
const CMDLineKey HostServ::LA_ADMINVHOST( "admin-vhost", true );
const CMDLineKey HostServ::SA_HTTPDPORT( "httpdport", false );
const CMDLineKey HostServ::LA_HTTPDPORT( "httpd-port", false );
const CMDLineKey HostServ::SA_HTTPDMAXCLIENTS( "maxclients", false );
const CMDLineKey HostServ::LA_HTTPDMAXCLIENTS( "httpd-maxclients", true );
const CMDLineKey HostServ::SA_HTTPDDOCUMENTROOT( "docroot", false );
const CMDLineKey HostServ::LA_HTTPDDOCUMENTROOT( "--httpd-documentroot", true );
const CMDLineKey HostServ::SA_HTTPDDEFDOCUMENT( "httpd-indexfile", false );
const CMDLineKey HostServ::LA_HTTPDDEFDOCUMENT( "httpd-default-document", true );


using httpd::HTTPd;

HostServ::HostServ( arg_type args, int_type argNum )
   : Application( APP_NAME, args, argNum ),
     m_configuration( HSConfiguration::getConfiguration() ),
     m_hostserv( NULL ),
     m_httpd( NULL ),
     m_run( true ),
     m_doIo( true ),
     m_reloadConfig( false ),
     m_timeout( 250 )
{
   if( m_argNum > 0 )
   {
      fillConfigMap();
      HSConfiguration::getConfiguration().processCMDLineArgs( getArgMap() );
   }

   registerSignal( SIG_INT );
   registerSignal( SIG_HUP );
   registerSignal( SIG_SEGV );

#ifdef VERBOSE_MEMORY
  cout << "HostServ] created" << endl;
#endif
}

HostServ::~HostServ()
{
   erasePIDFile( PID_FILENAME );
}

void
HostServ::onStart()
{
   using httpd::HTTPd;
   
   if( HSConfiguration::getConfiguration().getHelp() )
   {
      printHelp();
      return;
   }

   pid_t filePID = readPIDFile( PID_FILENAME );
   if( filePID != 0 )
   {
      // maybe another hostserv instance is already running
      cout << "Another hostserv seems to be running with pid " << filePID << endl;
      cout << "If this is not true delete " << PID_FILENAME << endl;
      return;
   }

   HSConfiguration::getConfiguration().processConfigfile();

   const string* failureDirective = HSConfiguration::getConfiguration().check();
   if( failureDirective != NULL )
   {
      cout << "You have not set '" << *failureDirective << "' in your configuration file." << endl;
      cout << "Because this directive is mandatory I will quit." << endl;
      return;
   }
   else
   {
      cout << "Configuration loaded without errors." << endl;
   }

//    if( !HSConfiguration::getConfiguration().getForeground() )
//    {
//       Application::fork();

//       cout << "PID is:           " << getPID() << endl;
//       cout << "Logfile is:       " << LOG_FILE << endl;
//       cout << "Error logfile is: " << ERR_FILE << endl;

//       // [LOG,ERR]_FILE are defined in config.h
//       Application::redirectConsoleOutputStreams( LOG_FILE, ERR_FILE );

//       // we get here whenever a now child process has been created
//       // ( or fork failed. If this is not the case the mother will exit. 
//       // immediately ).
//       registerSignal( SIG_INT );
//       registerSignal( SIG_HUP );
//       registerSignal( SIG_SEGV );
//    }

   if( !writePIDFile( PID_FILENAME ) )
      cout << "Warning: could not write pidfile." << endl;

   TIME_INIT();

   TIME_START();
   HSConfiguration::getConfiguration().processVHostsConfigfile();
   HSConfiguration::getConfiguration().processUsersConfigfile();
   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "HostServ::onStart(): cfg load duration: ";
   TIME_PRINT(cout);
#endif


   Poller*              poller = new Poller( computePollerSize() ); // space for 1 client
   InetHostAddress      addr   = HSConfiguration::getConfiguration().getHTTPdBindInterface();

   if( addr )
      // if makeHTTPd fails makeHTTPd returns NULL. the following code lines always check
      // m_httpd is non nil before usage. therefore we do not have to check the result value here
      m_httpd = makeHTTPd( addr, poller );
   else
   {
      cout << "Error while resolving " << HSConfiguration::getConfiguration().getHTTPdBindInterface() << endl;
      cout << "HTTPd support deactived" << endl;
   }

   // update member
   m_run          = true;
   m_reloadConfig = false;

   while( m_run )
   {
      m_hostserv = makeHostserv( poller );
      addr       = HSConfiguration::getConfiguration().getRemoteHost();
      
      if( !addr )
      {
	 cout << "Could not resolve remote server name: " << HSConfiguration::getConfiguration().getRemoteHost() << endl;
	 break;
      }

      if( setupHostserv( addr, poller ) )
      {
	 
	 /*
	  * value between [0, HSConfiguration::getConfiguration().getHTTPdMaxClientTimeout]
	  * whenever timeout >= HSConfiguration::getConfiguration().getHTTPdMaxClientTimeout()
	  * we will remove clients from which we did not send / received
	  * anything for the last 30 seconds
	  */
	 long timeout         = 0;
	 long lastMeasurement = time( NULL );

	 // this is the loop where all the work gets done
	 while( m_doIo && m_hostserv->getPollerClient().isRegistered() )
	 {
	    const int pollRet = poller->runSyscallPoll( 1000 );
	    
	    if( m_httpd != NULL )
	    {
	       if( pollRet == 0 )
	       {
		  // no IO
		  ++timeout;
		  ++lastMeasurement;
	       }
	       else if( pollRet > 0 )
	       {
		  /*
		   * at least traffic at one connection
		   */
		  int measurement = time( NULL );
		  timeout += measurement - lastMeasurement;
		  lastMeasurement = measurement;

		  m_httpd->run();
	       }

	       if( timeout >= HSConfiguration::getConfiguration().getHTTPdMaxClientTimeout() )
	       {
		  timeout = 0;
		  m_httpd->runRemoveTimeoutdClients();
	       }
	    }
	 }

	 m_hostserv->disconnect( "Shutdown requested" );
	 m_doIo = true; // to enable io after reconnect() request
	
	 // store current user / vhost database to disk
	 HSConfiguration::getConfiguration().serialize();

	 delete m_hostserv;

	 if( m_reloadConfig )
	 {
	    HSConfiguration::getConfiguration().processConfigfile();

	    m_reloadConfig = false;
	 }
      }

      /*
       * process all waiting http requests before sleeping
       */
      if( m_httpd != NULL )
	 m_httpd->run();	 

      if( !m_run )
	 break;
      
      cout << "Waiting " << DEF_RECONNECT_TIMEOUT << " sec to reconnect" << endl;
     
      sleep( DEF_RECONNECT_TIMEOUT ); // throttle reconnecting
   }

   if( m_httpd )
      delete m_httpd;

   delete poller;
   
   restoreConsoleOutputStreams();
}

void
HostServ::onStop()
{
   m_run = false;
}

void
HostServ::onSigSegv()
{
   if( m_hostserv )
      m_hostserv->disconnect( "Reveived SIG SEGV" );

   onStop();
}

void
HostServ::onSigInt()
{
   if( m_hostserv )
      m_hostserv->disconnect( "Received SIG INT" );

   onStop();

   deregisterSignal( SIG_SEGV );
}

void
HostServ::onSigHup()
{
  // application has not been started before. ignore this event therefore

  if( !m_hostserv ) 
    return;
  
  m_hostserv->disconnect( "Received SIG HUP" );

  m_doIo         = false;
  m_reloadConfig = true;
  
}

void
HostServ::printHelp()
{
  cout 
    << "Available options for HostServ:" << endl
    << endl


    << "TCP options: [MANDATORY]" << endl

    << LA_REMOTEHOST << "=<inet address>" << endl
    << SA_REMOTEHOST << "<inet address>\tset the inet address of the remote irc server" << endl
    << "\t\t\twhich HostServ will connect to establish a" << endl
    << "\t\t\ttcp connection" << endl

    << LA_REMOTEPORT << "=<port>" << endl
    << SA_REMOTEPORT << " <port>\t\tset the port which will be used to" << endl
    << "\t\t\testablish a tcp connection" << endl


    << endl



    << "IRC connection related options: [MANDATORY / OPTIONAL]" << endl

    << LA_SERVNAME << "=<server name>" << endl
    << SA_SERVNAME << " <server name>\tset the irc servername of HostServ. Mandatory." << endl

    << LA_REMOTEPASSWORD << "=<password>" << endl
    << SA_REMOTEPASSWORD << " <password>\tset the password which will be used to establish" << endl
    << "\t\t\ta irc connection with the remote irc server. Mandatory." << endl

    << LA_SERVERDESCR << "=<server description>" << endl
    << SA_SERVERDESCR << " <server description>  set the server description of HostServ" << endl
    << "\t\t\tOptional." << endl

    << endl



    << "IRC representation options: [OPTIONAL]" << endl

    << LA_HSNICKNAME << "=<nickname>" << endl
    << SA_HSNICKNAME << " <nickname>\tsets the nickname of HostServ in your irc network." << endl
    << "\t\t\tDefault is 'HostServ'" << endl
    
    << LA_HSUSERNAME << "=<username>" << endl
    << SA_HSUSERNAME << " <username>\tsets the username of HostServ in your" << endl
    << "\t\t\tirc network. Default is 'hostserv'" << endl

    << LA_HSADDRESS << "=<inet address>" << endl
    << SA_HSADDRESS << " <inet address> sets the inet address which will be displayed" << endl
    << "\t\t\twhen HostServ gets whois'd. Default is 'service'" << endl

    << LA_HSREALNAME << "=<realname>" << endl
    << SA_HSREALNAME << " <realname>\tsets the realname for 'HostServ'." << endl
    << "\t\t\tDefault is 'HostServ <version>'" << endl

    << LA_HSMODES << "=<usermodes>" << endl
    << SA_HSMODES << " <usermodes>\tset the usermodes for HostServ. It is highly" << endl
    << "\t\t\trecommend to give HostServ at least irc operator" << endl
    << "\t\t\tstatus (+o). Default is '+io'" << endl

    << LA_HSCHANNEL << "=<channel name>" << endl
    << SA_HSCHANNEL << " <channel name> makes hostserv join '<channel name>'" << endl
    << "\t\t\tThis option must is not mandatory" << endl


    << endl



    << "HostServ Admin options: [MANDATORY / OPTIONAL]" << endl

    << LA_ADMINLOGIN << "=<login name>" << endl
    << SA_ADMINLOGIN << " <login name>\tsets the login name for the default admin" << endl
    
    << LA_ADMINPASSWORD << "=<password>" << endl
    << SA_ADMINPASSWORD << " <password>\tsets  the login password for the default admin" << endl

    << LA_ADMINVHOST << "=<vhost>" << endl
    << SA_ADMINVHOST << " <vhost>\t\tsets the vhost associated with the admin. Optional." << endl


    << endl



    << "HostServ options: [OPTIONAL]" << endl

    << LA_CLIENTDEFAULTPERM << "=<permission string>" << endl
    << SA_CLIENTDEFAULTPERM << "<permission string> sets the default permissions for" << endl
    << "\t\t\tnewly created clients. Default is '0'" << endl

    << LA_VHOSTPUBLICPOLICY << "=<1|0>" << endl
    << SA_VHOSTPUBLICPOLICY << " <1|0>\t\tsets the default public policy for newly" << endl
    <<"\t\t\tcreated vhosts. Default is '0'" << endl
    
    << LA_VHOSTSFILENAME << "=<filename>" << endl
    << SA_VHOSTSFILENAME << " <filename>\tsets the filename of the vhost db" << endl
    << "\t\t\tDefault value is set by the configure script" << endl
    
    << LA_USERSFILENAME << "=<filename>" << endl
    << SA_USERSFILENAME << " <filename>\tsets the filename of the user db" << endl
    << "\t\t\tDefault value is set by the configure script" << endl

    << LA_HSCONFIGFILENAME << "=<filename>" << endl
    << SA_HSCONFIGFILENAME << " <filename>\tsets the filename of the" << endl
    << "\t\t\thostserv configurationfile." << endl
    << "\t\t\tDefault value is set by the configure script" << endl


    << endl

    << "General: [OPTIONAL]" << endl

    << LA_FOREGROUND << endl
    << SA_FOREGROUND << "\t\t\tmakes hostserv not forking into background" << endl

    << LA_HELP << endl
    << SA_HELP << "\t\t\tprint this help" << endl;
}

int
HostServ::computePollerSize()
{
   if( HSConfiguration::getConfiguration().httpdPortSet() )
   {
      return HSConfiguration::getConfiguration().getHTTPdMaxClientNum() + 2;
   }
   else
      return 1;
}

httpd::HTTPd*
HostServ::makeHTTPd( const InetHostAddress& addr, Poller* poller )
{
   // insert httpd into queue
   if( HSConfiguration::getConfiguration().httpdPortSet() )
   {
      cout << "Opening TCP port for HTTPd at: " << HSConfiguration::getConfiguration().getHTTPdPort() << endl;
      cout << "Document root is:              " << HSConfiguration::getConfiguration().getHTTPdDocumentRoot() << endl;
      
      TCPSocket::port_type httpdPort = 
	 static_cast<TCPSocket::port_type>( HSConfiguration::getConfiguration().getHTTPdPort() );

      return new httpd::HTTPd( const_cast<HostServ*>( this ),
			       poller,
			       addr,
			       httpdPort,
			       HSConfiguration::getConfiguration().getHTTPdMaxClientNum() );
   }

   return NULL;
}

HSIRCServerConnection*
HostServ::makeHostserv( Poller* poller )
{
   return new HSIRCServerConnection( const_cast<HostServ*>( this ),
				     HSConfiguration::getConfiguration().getServerName(),
				     HSConfiguration::getConfiguration().getRemotePassword(),
				     HSConfiguration::getConfiguration().getServerDescription(),
				     HSConfiguration::getConfiguration().getHostservNickname(),
				     HSConfiguration::getConfiguration().getHostservUsername(),
				     HSConfiguration::getConfiguration().getHostservAddress(),
				     HSConfiguration::getConfiguration().getHostservRealname(),
				     HSConfiguration::getConfiguration().getHostservModes(),
				     HSConfiguration::getConfiguration().getHostservChannel(),
				     &HSConfiguration::getConfiguration().getClients(),
				     &HSConfiguration::getConfiguration().getVHosts(),
				     poller,
				     8096,
				     8096 );
}

bool
HostServ::setupHostserv( const InetHostAddress& addr, Poller* poller )
{
   TCPSocket::port_type port   = static_cast<TCPSocket::port_type>( HSConfiguration::getConfiguration().getRemotePort() );

   if( !m_hostserv->registerAt( poller ) )
   {
      cout << "*ouhm* Could not add server connection to the poller. Quitting" << endl;
      return false;
   }   
   
   if( m_hostserv->connect( addr, port ) == File::eof )
   {
      cerr << "Could not connect " << addr << ':' << port << endl;
      return false;
   }
   
   return true;
}
