/***************************************************************************
                           hsircclientlist.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: hsircclientlist.cpp,v 1.11 2002/05/09 19:28:54 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "hsircclientlist.hpp"
#include "hsirctraits.hpp"
#include <iostream>
#include <debug.hpp>
#include <algorithm>

HSIRCClientList::HSIRCClientList() 
  : set<HSIRCClient *, DereferenceLess< HSIRCClient* > > ()
{
#ifdef VERBOSE_MEMORY
  cout << "HSIRCClientList] created" << endl;
#endif
}

HSIRCClientList::~HSIRCClientList()
{
  clear();

#ifdef VERBOSE_MEMORY
  cout << "HSIRCClientList] destroyed" << endl;
#endif
}

bool
HSIRCClientList::insertClient( HSIRCClient *c )
{
   TIME_INIT();
   
   
   TIME_START();
   pair<iterator, bool> p = insert( c );
   TIME_STOP();
   
#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCClientList::insertClient(): duration ";
   TIME_PRINT(cout);
#endif

  return p.second;
}

HSIRCClient*
HSIRCClientList::eraseClient( const IRCToken& nickname )
{
   TIME_INIT();

   TIME_START();
   const_iterator it = getPos( nickname );
   HSIRCClient* ret = NULL;
   TIME_STOP();

   if( it != end() )
   {
      ret = *it;
      erase( it );
   }
   else
     assert( false );

#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCClientList::eraseClient(): duration: ";
   TIME_PRINT(cout);
#endif
   
   return ret;
}

void
HSIRCClientList::updateClientNickname( const IRCToken& oldNickname,
				       const IRCToken& newNickname )
{
   TIME_INIT();
   TIME_START();

   HSIRCClient* client = eraseClient( oldNickname );
   assert( client );
   
   client->setNickname( newNickname );
   insertClient( client );


   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCClientList::updateClientNickname(): duration: ";
   TIME_PRINT(cout);
#endif
}

HSIRCClient* 
HSIRCClientList::getClient( const IRCToken &nickname ) const
{
   const_iterator it = getPos( nickname );
  if( it != end() ) return *it;
  else return NULL;
}

HSIRCClient*
HSIRCClientList::getClient( const HSClient* login ) const
{
   const_iterator it = getPos( login );
   if( it != end() ) return *it;
   else return NULL;
}

inline 
void
myDelete( const HSIRCClient* client )
{ delete client; }

void
HSIRCClientList::clear()
{
  for_each( begin(), end(), &myDelete );

  set<HSIRCClient *, DereferenceLess< HSIRCClient* > >::clear();  
}

HSIRCClientList::const_iterator 
HSIRCClientList::getPos( const IRCToken& nickname ) const
{
   TIME_INIT();
   
   TIME_START();
   HSIRCClientList::const_iterator it = find_if( begin(), end(),
						 HSIRCClientListPredicate( nickname ) );
   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCClientList::getPos(): duration: ";
   TIME_PRINT(cout);
#endif

   return it;
}

HSIRCClientList::const_iterator
HSIRCClientList::getPos( const HSClient* login ) const
{
   return find_if( begin(), end(),
		   HSIRCClientListLoginPredicate( login ) );
}

HSIRCClientListPredicate::HSIRCClientListPredicate( const IRCToken &RefName )
  : refName( RefName )
{
}

bool 
HSIRCClientListPredicate::operator() ( const HSIRCClient *c )
{
  return c->getNickname() == refName; 
}

HSIRCClientListLoginPredicate::HSIRCClientListLoginPredicate( const HSClient *login )
   : m_login( login )
{
}

bool
HSIRCClientListLoginPredicate::operator() ( const HSIRCClient *c )
{
   return c->getHSClient() == m_login;
}

HSIRCClientList&
DistributedHSIRCClientList::getList( int index )
{
   if( index < 0 || index > 32 )
      assert( false );
   
   return m_clientLists[index];
}

const HSIRCClientList&
DistributedHSIRCClientList::getList( int index ) const
{
   if( index < 0 || index > 32 )
      assert( false );
   
   return m_clientLists[index];
}

bool
DistributedHSIRCClientList::insertClient( HSIRCClient* c )
{
   if( c == NULL )
      return false; 
   
   const int        index = computeIndex( c->getNickname() );
   HSIRCClientList& list  = getList( index );
   
   return list.insertClient( c );
}

HSIRCClient*
DistributedHSIRCClientList::eraseClient( const IRCToken& nickname )
{
   const int        index = computeIndex( nickname );
   HSIRCClientList& list  = getList( index );

   return list.eraseClient( nickname );
}

void
DistributedHSIRCClientList::updateClientNickname( const IRCToken& oldNickname,
						  const IRCToken& newNickname )
{
   TIME_INIT();

   TIME_START();

   const int        index = computeIndex( oldNickname );
   HSIRCClientList& list  = getList( index );

   HSIRCClient* client = list.eraseClient( oldNickname );
   assert( client );
   
   client->setNickname( newNickname );

   insertClient( client );

   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "DistributedHSIRCClientList::updateClientNickname(): duration: ";
   TIME_PRINT( cout );
#endif
}

HSIRCClient*
DistributedHSIRCClientList::getClient( const IRCToken& nickname ) const
{
   TIME_INIT();

   TIME_START();

   const int              index = computeIndex( nickname );
   const HSIRCClientList& list  = getList( index );

   HSIRCClient* client = list.getClient( nickname );

   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "DistributedHSIRCClientList::getClient(1): duration: ";
   TIME_PRINT( cout );
#endif

   return client;
}

HSIRCClient*
DistributedHSIRCClientList::getClient( const HSClient* login ) const
{
   TIME_INIT();
   TIME_START();

   const HSIRCClientList* const endIter = &m_clientLists[0] + 26;
   
   HSIRCClient* client = NULL;

   for( const HSIRCClientList* iter = &m_clientLists[0]; iter != endIter; ++iter )
   {
      if( ( client = iter->getClient( login ) ) != NULL )
	 break;
   }


   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "DistributedHSIRCClientList::getClient(2): duraction : ";
   TIME_PRINT( cout );
#endif

   return client;
}

const HSIRCClientList*
DistributedHSIRCClientList::getClients( const IRCToken::data_type c ) const
{
   const int index = computeIndex( c );

   if( index < 0 || index > 32 )
      return NULL;

   return &m_clientLists[index];
}

void
DistributedHSIRCClientList::clear()
{
   const HSIRCClientList* const endIter = &m_clientLists[0] + 26;
   for( HSIRCClientList* iter = &m_clientLists[0]; iter != endIter; ++iter )
      iter->clear();
}


/*
 *   -1 -> error. no valid index
 * > -1 -> valid index. 
 *         As you can see we use some indices twice to reduce the
 *         amount of used memory ( for eg. is index of A == a, etc. )
 */
const char DistributedHSIRCClientList::LISTINDEX_TAB[] = {
   /* 0  */     ~0,
   /* 1  */     ~0,
   /* 2  */     ~0,
   /* 3  */     ~0, 
   /* 4  */     ~0, 
   /* 5  */     ~0, 
   /* 6  */     ~0, 
   /* 7 BEL */  ~0, 
   /* 8  \b */  ~0,
   /* 9  \t */  ~0,
   /* 10 \n */  ~0,
   /* 11 \v */  ~0,
   /* 12 \f */  ~0,
   /* 13 \r */  ~0, 
   /* 14 */     ~0, 
   /* 15 */     ~0,
   /* 16 */     ~0,
   /* 17 */     ~0,
   /* 18 */     ~0, 
   /* 19 */     ~0, 
   /* 20 */     ~0, 
   /* 21 */     ~0, 
   /* 22 */     ~0, 
   /* 23 */     ~0, 
   /* 24 */     ~0,
   /* 25 */     ~0,
   /* 26 */     ~0,
   /* 27 */     ~0,
   /* 28 */     ~0,
   /* 29 */     ~0,
   /* 30 */     ~0,
   /* 31 */     ~0,
   /* SP */     ~0, 
   /* ! */      ~0, 
   /* " */      ~0, 
   /* # */      ~0,
   /* $ */      ~0,
   /* % */      ~0,
   /* & */      ~0,
   /* ' */      ~0,
   /* ( */      ~0,
   /* ) */      ~0,
   /* * */      ~0,
   /* + */      ~0,
   /* , */      ~0,
   /* - */       0, // <-
   /* . */      ~0,
   /* / */      ~0,
   /* 0 */      ~0,
   /* 1 */      ~0,
   /* 2 */      ~0,
   /* 3 */      ~0,
   /* 4 */      ~0,
   /* 5 */      ~0,
   /* 6 */      ~0,
   /* 7 */      ~0,
   /* 8 */      ~0,
   /* 9 */      ~0,
   /* : */      ~0,
   /* ; */      ~0,
   /* < */      ~0,
   /* = */      ~0,
   /* > */      ~0,
   /* ? */      ~0,
   /* @ */      ~0,
   /* A */       1, // <-
   /* B */       2, // <-
   /* C */       3, // <-
   /* D */       4, // <-
   /* E */       5, // <-
   /* F */       6, // <-
   /* G */       7, // <-
   /* H */       8, // <-
   /* I */       9, // <-
   /* J */      10, // <-
   /* K */      11, // <-
   /* L */      12, // <-
   /* M */      13, // <-
   /* N */      14, // <-
   /* O */      15, // <-
   /* P */      16, // <-
   /* Q */      17, // <-
   /* R */      18, // <-
   /* S */      19, // <-
   /* T */      20, // <-
   /* U */      21, // <-
   /* V */      22, // <-
   /* W */      23, // <-
   /* X */      24, // <-
   /* Y */      25, // <-
   /* Z */      26, // <-
   /* [ */      27, // <-
   /* \ */      28, // <-
   /* ] */      29, // <-
   /* ^ */      30, // <-
   /* _ */      31, // <-
   /* ` */      32, // <-
   /* a */       1, // <-
   /* b */       2, // <-
   /* c */       3, // <-
   /* d */       4, // <-
   /* e */       5, // <-
   /* f */       6, // <-
   /* g */       7, // <-
   /* h */       8, // <-
   /* i */       9, // <-
   /* j */      10, // <-
   /* k */      11, // <-
   /* l */      12, // <-
   /* m */      13, // <-
   /* n */      14, // <-
   /* o */      15, // <-
   /* p */      16, // <-
   /* q */      17, // <-
   /* r */      18, // <-
   /* s */      19, // <-
   /* t */      20, // <-
   /* u */      21, // <-
   /* v */      22, // <-
   /* w */      23, // <-
   /* x */      24, // <-
   /* y */      25, // <-
   /* z */      26, // <-
   /* { */      27, // <-
   /* | */      28, // <-
   /* } */      29, // <-
   /* ~ */      ~0,
   /* del  */   ~0,
   /* 0x80 */   ~0,
   /* 0x81 */   ~0,
   /* 0x82 */   ~0,
   /* 0x83 */   ~0,
   /* 0x84 */   ~0,
   /* 0x85 */   ~0,
   /* 0x86 */   ~0,
   /* 0x87 */   ~0,
   /* 0x88 */   ~0,
   /* 0x89 */   ~0,
   /* 0x8A */   ~0,
   /* 0x8B */   ~0,
   /* 0x8C */   ~0,
   /* 0x8D */   ~0,
   /* 0x8E */   ~0,
   /* 0x8F */   ~0,
   /* 0x90 */   ~0,
   /* 0x91 */   ~0,
   /* 0x92 */   ~0,
   /* 0x93 */   ~0,
   /* 0x94 */   ~0,
   /* 0x95 */   ~0,
   /* 0x96 */   ~0,
   /* 0x97 */   ~0,
   /* 0x98 */   ~0,
   /* 0x99 */   ~0,
   /* 0x9A */   ~0,
   /* 0x9B */   ~0,
   /* 0x9C */   ~0,
   /* 0x9D */   ~0,
   /* 0x9E */   ~0,
   /* 0x9F */   ~0,
   /* 0xA0 */   ~0,
   /* 0xA1 */   ~0,
   /* 0xA2 */   ~0,
   /* 0xA3 */   ~0,
   /* 0xA4 */   ~0,
   /* 0xA5 */   ~0,
   /* 0xA6 */   ~0,
   /* 0xA7 */   ~0,
   /* 0xA8 */   ~0,
   /* 0xA9 */   ~0,
   /* 0xAA */   ~0,
   /* 0xAB */   ~0,
   /* 0xAC */   ~0,
   /* 0xAD */   ~0,
   /* 0xAE */   ~0,
   /* 0xAF */   ~0,
   /* 0xB0 */   ~0,
   /* 0xB1 */   ~0,
   /* 0xB2 */   ~0,
   /* 0xB3 */   ~0,
   /* 0xB4 */   ~0,
   /* 0xB5 */   ~0,
   /* 0xB6 */   ~0,
   /* 0xB7 */   ~0,
   /* 0xB8 */   ~0,
   /* 0xB9 */   ~0,
   /* 0xBA */   ~0,
   /* 0xBB */   ~0,
   /* 0xBC */   ~0,
   /* 0xBD */   ~0,
   /* 0xBE */   ~0,
   /* 0xBF */   ~0,
   /* 0xC0 */   ~0,
   /* 0xC1 */   ~0,
   /* 0xC2 */   ~0,
   /* 0xC3 */   ~0,
   /* 0xC4 */   ~0,
   /* 0xC5 */   ~0,
   /* 0xC6 */   ~0,
   /* 0xC7 */   ~0,
   /* 0xC8 */   ~0,
   /* 0xC9 */   ~0,
   /* 0xCA */   ~0,
   /* 0xCB */   ~0,
   /* 0xCC */   ~0,
   /* 0xCD */   ~0,
   /* 0xCE */   ~0,
   /* 0xCF */   ~0,
   /* 0xD0 */   ~0,
   /* 0xD1 */   ~0,
   /* 0xD2 */   ~0,
   /* 0xD3 */   ~0,
   /* 0xD4 */   ~0,
   /* 0xD5 */   ~0,
   /* 0xD6 */   ~0,
   /* 0xD7 */   ~0,
   /* 0xD8 */   ~0,
   /* 0xD9 */   ~0,
   /* 0xDA */   ~0,
   /* 0xDB */   ~0,
   /* 0xDC */   ~0,
   /* 0xDD */   ~0,
   /* 0xDE */   ~0,
   /* 0xDF */   ~0,
   /* 0xE0 */   ~0,
   /* 0xE1 */   ~0,
   /* 0xE2 */   ~0,
   /* 0xE3 */   ~0,
   /* 0xE4 */   ~0,
   /* 0xE5 */   ~0,
   /* 0xE6 */   ~0,
   /* 0xE7 */   ~0,
   /* 0xE8 */   ~0,
   /* 0xE9 */   ~0,
   /* 0xEA */   ~0,
   /* 0xEB */   ~0,
   /* 0xEC */   ~0,
   /* 0xED */   ~0,
   /* 0xEE */   ~0,
   /* 0xEF */   ~0,
   /* 0xF0 */   ~0,
   /* 0xF1 */   ~0,
   /* 0xF2 */   ~0,
   /* 0xF3 */   ~0,
   /* 0xF4 */   ~0,
   /* 0xF5 */   ~0,
   /* 0xF6 */   ~0,
   /* 0xF7 */   ~0,
   /* 0xF8 */   ~0,
   /* 0xF9 */   ~0,
   /* 0xFA */   ~0,
   /* 0xFB */   ~0,
   /* 0xFC */   ~0,
   /* 0xFD */   ~0,
   /* 0xFE */   ~0,
   /* 0xFF */   ~0
};
