/***************************************************************************
                           hsircserverconnection.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: hsircserverconnection.cpp,v 1.31 2002/06/13 16:07:47 d3cline Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <algorithm>
#include "hsircserverconnection.hpp"
#include "application.hpp"
#include "debug.hpp"
#include "irc.hpp"
#include "hostserv.hpp"
#include "hscommands.hpp"
#include "config.h"

const IRCToken HSIRCServerConnection::IRCCMD_SETHOST            = " CHGHOST ";
const IRCToken HSIRCServerConnection::IRCCMD_SETIDENT           = " CHGIDENT ";
const IRCToken HSIRCServerConnection::DEFAULT_KILL_CAUSE        = "This nickname is reserved for a service.";


HSIRCServerConnection::HSIRCServerConnection( HostServ*       mother,
					      const IRCToken& hostservName,
					      const IRCToken& remotePassword, 
					      const IRCToken& hostservDescription,
					      const IRCToken& HostservNick,
					      const IRCToken& HostservIdent,
					      const IRCToken& HostservAddress,
					      const IRCToken& HostservRealname,
					      const IRCToken& HostservModes,
					      const IRCToken& HostservChannel,
					      HSClientList*   Clients,
					      VHostList*      Vhosts,
					      Poller*         poller,
					      len_type        inputBufLen,
					      len_type        outputBufLen )
   : IRCServerConnection ( hostservName,
			   remotePassword,
			   hostservDescription,
			   poller,
			   inputBufLen,
			   outputBufLen ),
     hostservNick( HostservNick ),
     hostservIdent( HostservIdent ),
     hostservAddress( HostservAddress ),
     hostservRealname ( HostservRealname ),
     hostservModes( HostservModes ),
     hostservChannel( HostservChannel ),
     m_mother( mother ),
     hsClients( Clients ),
     ircClients( new HSIRCClientList() ),
     vhosts( Vhosts ),
     m_lastAuthedState( NULL ),
     m_lastUnauthedState( NULL ),
     m_authStatePrintState( PRINT_NOTHING ),
     m_unauthedStatePrintState( PRINT_NOTHING )
{
#ifdef DEBUG
   cout << "HSIRCServerConnection] created" << endl;
#endif
}

HSIRCServerConnection::~HSIRCServerConnection()
{
   for_each( ircClients->begin(), ircClients->end(), mem_fun( &HSIRCClient::logout ) );

   delete ircClients;
   delete m_lastAuthedState;
   delete m_lastUnauthedState;

#ifdef DEBUG
   cout << "HSIRCServerConnection] destroyed" << endl;
#endif
}

void
HSIRCServerConnection::disconnect( const char *cause )
{
   printSquit( hostservNick, cause );
   flushOutput();
   close();
}


// FIXME: !change IRCServerConnection's behaviour to make
// IRCServerConnection emit 'onLogedin()' after nickburst
// only!
void
HSIRCServerConnection::onLogedin()
{
   cout << "logedin" << endl;
   
   // we do not want any fake hostserv nor disconnect
   // by remote server because of already-in-use nickname 'hostservNick'
   printKill( getServername(), hostservNick, DEFAULT_KILL_CAUSE );

   introduceUser( hostservNick, hostservIdent,
		  hostservAddress, hostservModes,
		  hostservRealname );
  
   if( hostservChannel.length() > 0 )
      joinChannel( hostservNick, hostservChannel );
}

void
HSIRCServerConnection::onOutput()
{
   using namespace hostserv::commands;

   IRCServerConnection::onOutput();

   TIME_INIT();
   TIME_START();

   if( m_authStatePrintState == PRINT_NOTHING && m_unauthedStatePrintState == PRINT_NOTHING )
   {
      m_pollerClient.unregisterEvent( PollerFD::EVT_OUT );
      m_pollerClient.registerEvent( PollerFD::EVT_IN );

      return;
   }
   else
   {
      m_pollerClient.unregisterEvent( PollerFD::EVT_IN );
      m_pollerClient.registerEvent( PollerFD::EVT_OUT );
   }

   if( m_authStatePrintState != PRINT_NOTHING )
   {
      assert( m_lastAuthedState );

      if( m_authStatePrintState == PRINT_OLD )
	 m_authStatePrintState = PRINT_STATE;
      else
      {
	 if( ! ( m_lastAuthedState = HSAuthedCommands::getSet()->resume( m_lastAuthedState ) ) )
	    m_authStatePrintState = PRINT_NOTHING;
      }
   } 

   if( m_unauthedStatePrintState != PRINT_NOTHING )
   {
      assert( m_lastUnauthedState );

      if( m_unauthedStatePrintState == PRINT_OLD )
	 m_unauthedStatePrintState = PRINT_STATE;
      else
      {
	 if( ! ( m_lastUnauthedState = HSAuthedCommands::getSet()->resume( m_lastAuthedState ) ) )
	    m_unauthedStatePrintState = PRINT_NOTHING;
      }
   } 

   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCServerConnection::onOutput(): command execution duration: ";
   TIME_PRINT( cout ) << endl;
#endif
}

void
HSIRCServerConnection::onPrivmsg( const IRCToken& from, const IRCToken& to,
				  const IRCToken* msg, int argc )
{
   processMsg( from, to, msg, argc );
}

void
HSIRCServerConnection::onNotice( const IRCToken& from, const IRCToken& to,
				 const IRCToken* msg, int argc )
{
   processMsg( from, to, msg, argc );
}

void
HSIRCServerConnection::onNickChange( const IRCToken &oldNick,
				     const IRCToken &newNick )
{
   ircClients->updateClientNickname( oldNick, newNick );
}

void
HSIRCServerConnection::onNickIntroduce( const IRCToken &nick, 
					const IRCToken &ident,
					const IRCToken &address,
					const IRCToken &realname )
{
   ircClients->insertClient( new HSIRCClient( nick,
					      ident,
					      address,
					      realname,
					      NULL ) );
}

void
HSIRCServerConnection::onQuit( const IRCToken& from, 
			       const IRCToken* msg, 
			       int argc )
{
   processClientQuit( from );
}

void
HSIRCServerConnection::onKill( const IRCToken &from, 
			       const IRCToken &target,
			       const IRCToken* cause,
			       int argc )
{
   if( EqualsIgnoreCase( target, hostservNick ) )
      processForciblyIntroduceHostserv();
   else
      processClientQuit( target );
}

void
HSIRCServerConnection::onWhois( const IRCToken& from,
				const IRCToken& who )
{
   if( EqualsIgnoreCase( who, hostservNick ) )
   {
      tcp() << ':' << getServername() << ' ' << RPL_WHOISUSER << ' '  << from 
	    << ' ' << hostservNick << ' ' << hostservIdent  << ' ' 
	    << hostservAddress << " * :" << hostservRealname << endl
	 
	    << ':' << getServername() << ' ' << RPL_WHOISSERVER << ' ' << from
	    << ' ' << hostservNick << ' ' << getServername()
	    << " :" << getDescription() << endl
	 
	    << ':' << getServername() << ' ' << RPL_ENDOFWHOIS << ' ' << from << ' ' << who 
	    << " :End of /WHOIS list" << endl;
   }
}

void
HSIRCServerConnection::onServerError( const IRCServerMessage &m )
{
   cout << "Fatal error message from remote irc server: " << m.getMessage() << endl;
   cout << "Closing connection ... " << endl;
}

void
HSIRCServerConnection::onNicknameInUse( const IRCToken &nicknameInUse )
{
   if( nicknameInUse == hostservNick )
      processForciblyIntroduceHostserv();
}

void
HSIRCServerConnection::onNicknameCollision( const IRCToken &collidedNick )
{
   if( collidedNick == hostservNick ) 
   { 
      // under normal circumstances this check is overkill
      // however we don't want to produce more useless traffic
      processForciblyIntroduceHostserv();
   }
}

void 
HSIRCServerConnection::processMsg( const IRCToken& sender,
				   const IRCToken& receiver,
				   const IRCToken* tokenArray,
				   int len
   )
{
   if( !EqualsIgnoreCase( receiver, hostservNick ) ) {
#ifdef DEBUG
      cout << "Dropping message for irregular receiver: " << receiver << endl;
#endif
      return;
   }

   if( sender == hostservNick )
   {
#ifdef DEBUG
      cout << "Dropping message from me assigned to me :-)" << endl;
#endif
      return;
   }

   if( len == 0 )
   {
#ifdef DEBUG
      cout << "Ignoring empty message" << endl;
      return;
#endif
   }

   HSIRCClient* senderClient = ircClients->getClient( sender );
   assert( senderClient != NULL );

   IRCToken command = ToLower( *tokenArray );

   using namespace hostserv::commands;

   TIME_INIT();
   TIME_START();

   if( senderClient->isAuthed() )
   {
      m_lastAuthedState = HSAuthedCommands::getSet()->execute( *this,
							       *senderClient,
							       command,
							       tokenArray+1,
							       --len );
      if( m_lastAuthedState )
	 m_authStatePrintState = PRINT_OLD;
   }
   else
   {
      m_lastUnauthedState = HSUnauthedCommands::getSet()->execute( *this,
								   *senderClient,
								   command,
								   tokenArray+1,
								   --len );
      if( m_lastUnauthedState )
	 m_unauthedStatePrintState = PRINT_OLD;
   }

//    TIME_STOP();

//    if( !processCmdHandlerTab( &CMDHANDLERS[0],
// 			      *senderClient,
// 			      command,
// 			      &tokenArray[1],
// 			      --len ) )
//      printHelp( *senderClient );

   TIME_STOP();

#ifdef VERBOSE_TIMESTATS
   cout << "HSIRCServerConnection::processMsg(): command execution duration: ";
   TIME_PRINT( cout );
#endif

}

void
HSIRCServerConnection::changeHost( const IRCToken &target, const IRCToken &newAddress,
				   bool noticeUser )
{
   tcp() << ':' << getServername() << IRCCMD_SETHOST << target << ' ' << newAddress << endl;
#ifdef DEBUG
   cout << "changeHost(): printed: " 
	<< ':' << getServername() << IRCCMD_SETHOST << target << ' ' << newAddress << endl;
#endif

   if( noticeUser )
      printChangedHost( target, newAddress );
}

void
HSIRCServerConnection::changeUsername( const IRCToken &target, const IRCToken &newIdent,
				       bool noticeUser )
{
   tcp() << ':' << getServername() << IRCCMD_SETIDENT << target << ' ' << newIdent << endl;
#ifdef DEBUG
   cout << "changeUserName(): printed: "
	<< ':' << getServername() << IRCCMD_SETIDENT << target << ' ' << newIdent << endl;
#endif

   if( noticeUser )
      printChangedUsername( target, newIdent );
}

void
HSIRCServerConnection::printHelpArray( const IRCToken &target,
                                     const IRCToken *msgArray, int len)
{
   for( int i = 0; i < len; ++i, ++msgArray ) {
      printNotice( hostservNick, target, *msgArray );
   }
}

void
HSIRCServerConnection::printHelpArray( const IRCToken &target,
                                     const unsigned char *msgArray[] )
{
   for( ; *msgArray != NULL; ++msgArray ) {
      printNotice( hostservNick, target, *msgArray );
   }
}

void
HSIRCServerConnection::applyVHost( const HSIRCClient& client, const VHost *const v,
				   bool noticeUser )
{
   // reduce traffic
   if( client.getAddress() == v->getAddress() )
      return;

   changeHost( client.getNickname(), 
	       v->getAddress(), 
	       noticeUser );
  
   if( !v->addressOnly() && client.getHSClient()->updateUsername() )
      changeUsername( client.getNickname(),
		      v->getIdent(),
		      noticeUser );
}

void
HSIRCServerConnection::processForciblyIntroduceHostserv()
{
   printKill( getServername(), hostservNick, DEFAULT_KILL_CAUSE );
   introduceUser( hostservNick, hostservIdent,
                  hostservAddress, hostservModes,
                  hostservRealname );
}

void
HSIRCServerConnection::processClientQuit( const IRCToken &target )
{
#ifdef DEBUG
   cout << "Removing persistent entry for client " << target << endl;
#endif

   HSIRCClient* hsTarget = ircClients->eraseClient( target );
   if( hsTarget )
   {
      if( hsTarget->isAuthed() )
         hsTarget->logout();
      delete hsTarget;
   }
   else
      assert( false );
}

void
HSIRCServerConnection::printChangedHost( const IRCToken& target, 
                                         const IRCToken& newHost )
{
   const static unsigned char* MSG = (const unsigned char*)"Your new hostname is %B%s%B";
   printNotice( hostservNick, target, MSG, &newHost );
}

void
HSIRCServerConnection::printChangedUsername( const IRCToken& target,
                                             const IRCToken& newIdent )
{
   const static unsigned char* MSG1 = (const unsigned char*)"Your new username is %B%s%B";
   printNotice( hostservNick, target, MSG1, &newIdent );
}

bool
HSIRCServerConnection::EqualsIgnoreCase( const IRCToken& lhs, const IRCToken& rhs )
{
   if( lhs.length() == rhs.length() ) {
      IRCToken::reference_type lIter = lhs.c_str();
      IRCToken::reference_type rIter = rhs.c_str();
      IRCToken::reference_type lIterMax = lIter + lhs.length();

      while( lIter < lIterMax )
      {
	 if( tolower( *lIter++ ) != tolower( *rIter++ ) ) return false;
      }

      return true;
   }
   else return false;
}

static void 
toLower( IRCToken::reference_type c )
{
   *c = tolower( *c );
}

IRCToken
HSIRCServerConnection::ToLower( const IRCToken& tok )
{
   IRCToken s( tok );
   IRCToken::reference_type bufIter = s.c_str();
   IRCToken::reference_type bufIterEnd = bufIter+ s.length();
   while( bufIter < bufIterEnd )
      toLower( bufIter++ );

   return s;
}

bool
HSIRCServerConnection::IRCTokenToBool( const IRCToken& s )
{
   return s == "1" ;
}
