/***************************************************************************
                           httpclientconnection.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: httpclientconnection.cpp,v 1.21 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdio> /* for perror */
#include "httpclientconnection.hpp"


namespace httpd
{

HTTPClientConnection::HTTPClientConnection( HTTPd* server,
					    TCPSocket::fd_type Fd,
					    id_type id )
   : TCPClientConnection( Fd ),
     m_inputBuffer( *this, DEF_INBUFLEN ),
     m_outputBuffer( *this, DEF_OUTBUFLEN, http::HTTPCONN_CLOSE ),
     m_ostream( &m_outputBuffer ),
     m_server( server ),
     m_request( NULL ),
     m_id( id ),
     m_printStatus( PRINTING_HEADER ),
     m_recvStatus( RECV_STATUSLINE )
{
#ifdef VERBOSE_MEMORY
   cout << "HTTPClientConnection] created" << endl;
#endif
}

HTTPClientConnection::~HTTPClientConnection()
{
#ifdef VERBOSE_MEMORY
   cout << "HTTPClientConnection] client" << m_id << " destroyed" << endl;
#endif   

   delete m_request;

   m_server->removeClientConnection( const_cast<HTTPClientConnection*>( this ) );
}

void
HTTPClientConnection::writeResponse( const Response* resp )
{
   m_response     = resp;
   m_printStatus  = PRINTING_HEADER;
}

void
HTTPClientConnection::onInput()
{
   if( m_inputBuffer.fill() == File::eof )
   {
      delete this;
      return;
   }
   

   bool processRequest = false;
   bool readError      = false;

   http::MessageHeader            messageHeader;
   RequestFileBuffer::return_type lastReturn    = RequestFileBuffer::READ_OK;
      
   
   while( !m_inputBuffer.readCompletely() )
   {
      if( m_recvStatus == RECV_HEADER )
      {
	 /*
	  * receive request header
	  */
	    
	 // request exists - let's read 'message-header'
	 lastReturn = m_inputBuffer.read( messageHeader );
	    
	 if( lastReturn == RequestFileBuffer::READ_OK )
	 {
	    m_request->appendMessageHeader( messageHeader );
	    continue;
	 }
      }
      else if( m_recvStatus == RECV_STATUSLINE )
      {
	 /*
	  * Receive statusline
	  */
	 http::RequestStartline startLine;

	 lastReturn = m_inputBuffer.read( startLine );
	    
	 if( lastReturn == RequestFileBuffer::READ_OK )
	 {
	    // m_request will be deleted by the dtor
	    m_request = make_request( startLine.getMethod(), this );
	    m_request->setStartline( startLine );
	       
	    if( startLine.getProtocolVersion() == http::HTTPVER_UNKNOWN )
	    {
	       processRequest = true;
	       goto end;
	    }
	       
	    m_recvStatus = RECV_HEADER;
	       
	    continue;
	 }
      }
      else
      {
	 /*
	  * recv request body
	  */
	    
	 RequestBody* body = m_request->getRequestBody();
	    
	 if( body == NULL )
	 {
	    // ok no body though m_request.requiresBody() returned true.
	    // dont stop action totally at this stage as the method may
	    // be able to work without a request body
	    processRequest = true;
	    goto end;
	 }
      
	 len_type retLen = body->read( m_inputBuffer );
	    
	 if( retLen > 0 )
	    continue;
	 else 
	 {
	    if( retLen < 0 )
	    {
	       delete m_request;
	       m_request = new IncompleteRequest( this );
	    }

	    processRequest = true;
	    goto end;
	 }
      }
	 

      switch( lastReturn )
      {
	 case RequestFileBuffer::ERR_MALFORMED:
	    if( m_request ) 
	       delete m_request;
	       
	    // m_request will be deleted by the dtor
	    m_request = new IllegalRequest( this );
	    processRequest = true;
	    goto end;
	       
	 case RequestFileBuffer::ERR_BUFFEREMPTY:
	    if( m_inputBuffer.fill() == File::eof )
	    {
	       readError = true;
	       goto end;
	    }
	    break;
	    
	 case RequestFileBuffer::HEADER_TERMINATE:
	    if( !m_request )
	       // m_request will be deleted by the dtor
	       m_request = new IncompleteRequest( this );
	    else if( m_request->requiresResponseBody() )
	    {
	       m_recvStatus = RECV_BODY;
	       continue;
	    }
	    else 
	    {
	       processRequest = true;
	       goto end;
	    }
      } // EOswitch
   } //EOwhile

   return;

  end:
   if( !readError )
   {
      m_inputBuffer.flush();
      
      if( processRequest )
      {
	 m_server->processRequest( const_cast<HTTPClientConnection*>( this ),
				   m_request );
	 m_recvStatus = RECV_STATUSLINE;
      }

      return;
   }

   delete this;
}

void
HTTPClientConnection::onOutput()
{
   bool changeMode = false;

   assert( m_response );

   if( !m_ostream )
   {
      delete this;
      return;
   }

   if( m_printStatus == PRINTING_BODY )
   {
      int printLen = m_response->getResponseBody()->printJunk( m_ostream );
      
      if( printLen == 0 )
	 // end of body reached
	 m_printStatus = PRINTING_DONE;
   }
   else if( m_printStatus == PRINTING_HEADER )
   {
      m_ostream << m_response->getStatusLine() << http::httpendl;

      const Response::container_type& container =
	 m_response->getMessageHeaders();
      Response::container_type::const_iterator it =
	 container.begin();
      const Response::container_type::const_iterator end =
	 container.end();

      while( it != end )
      {
	 m_ostream << *it << http::httpendl;
	 ++it;
      }

      m_ostream << http::httpendl;

      if( !m_response->getResponseBody() )
	 m_printStatus = PRINTING_DONE;
      else
      {
	 m_printStatus = PRINTING_BODY;
	 changeMode = true;
      }
   }
   // only for debugging purposes
   else
      assert( false );

   // note: even if the last printing returned '0' 
   //       we have to flush the buffer here
   //       to make the buffer print 
   //       rfc 2616 'last-chunk'
   m_outputBuffer.doFlush();


   if( changeMode )
      m_outputBuffer.setBodyTransportEncoding( m_response->getResponseBodyEncoding() );
   else if( m_printStatus == PRINTING_DONE )
      printingDone();
}

void
HTTPClientConnection::onError()
{
#ifdef DEBUG
   cout << "HTTPClientConnection::onError()" << endl;
#endif
   
   delete this;
}

void
HTTPClientConnection::onHangup()
{
#ifdef DEBUG
   cout << "HTTPClientConnection::onHangup()" << endl;
#endif

   delete this;
}

void
HTTPClientConnection::printingDone()
{
   m_server->responsePrinted( const_cast<HTTPClientConnection*>( this ),
			      *m_response );
   
   if( m_response->getConnectionType() == http::HTTPCONN_CLOSE )
      delete this;
   else
   {
      // if we don't set encoding here it may be that
      // response header printing gets done in junked mode.
      m_outputBuffer.setBodyTransportEncoding( http::HTTPCONN_CLOSE );

      // reset member variables
      delete m_request;
      m_request     = NULL;
      m_response    = NULL;
      m_printStatus = PRINTING_DONE;
   }

}

}//EON httpd
