/***************************************************************************
                           httpd.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: httpd.cpp,v 1.14 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <algorithm>
#include <ctime>
#include "httpd.hpp"


//TODO: let httpd use its own poller and do some request queueing, etc.
//      this could be done with a seperate thread *hinthint*
namespace httpd
{

HTTPd::HTTPd( const HostServ* hostserv,
	      Poller* poller, 
	      const InetHostAddress& bind, TCPSocket::port_type port,
	      int maxClientNum,
	      long maxTimeoutTime )
   : m_hostserv( hostserv ),
     m_poller( poller ),
     m_tcpPort( new HTTPServerPort( this, m_poller, bind, port ) ),
     m_tcpPortPollerClient( m_poller, m_tcpPort, false ),
     m_clients( maxClientNum ),
     m_requestQueue(),
     m_clientNum( 0 ),
     m_maxClientNum( maxClientNum ),
     m_maxTimeout( maxTimeoutTime )
{
   m_tcpPortPollerClient.registerClient();
   
   m_tcpPortPollerClient.registerEvent( PollerFD::EVT_IN );
   m_tcpPortPollerClient.registerEvent( PollerFD::EVT_ERR );
   m_tcpPortPollerClient.registerEvent( PollerFD::EVT_HUP );
}

inline void _myDelete( const HTTPd::httpclientlist_type::value_type pc )
{ delete pc; }

HTTPd::~HTTPd()
{
   // shutdown every client
   for_each( m_clients.begin(), m_clients.end(), &_myDelete );

   m_tcpPortPollerClient.unregister();

   delete m_tcpPort;
}

static long getCurrentSecs()
{
   return static_cast<long>( time( NULL ) );
}

int
HTTPd::runRemoveTimeoutdClients()
{
   httpclientlist_type::const_iterator iter = m_clients.begin();
   httpclientlist_type::const_iterator end  = m_clients.end();
   int numClientsRemoved = 0;

   long currTime = getCurrentSecs();

   for( ; numClientsRemoved < m_clientNum && iter != end; ++iter )
   {
      MeasuringPollerClient* client = *iter;

      if( client == NULL )
	 continue;

      long timeDiff = currTime - client->getLastIOEventPointInTime();

      if( timeDiff > m_maxTimeout )
      {
	 delete client->getPollable()->getPollableFile();

	 ++numClientsRemoved;
      }
   }

   return numClientsRemoved;
}

void
HTTPd::addClientConnection( File::fd_type fd )
{
   if( m_clientNum == m_maxClientNum )
      UNIXFile( fd ).close();

   const id_type id = getNextFreeID();

   HTTPClientConnection*  conn   = new HTTPClientConnection( this, fd, id );
   MeasuringPollerClient* client = new MeasuringPollerClient( m_poller, conn, true );

   client->registerEvent( PollerFD::EVT_IN );

   m_clients[id] = client;
   ++m_clientNum;

   if( m_clientNum == m_maxClientNum )
      // makes tcpPort not listen() anymore
      // EVT_IN will be reregistered after
      // the first removeClientConnection call
      m_tcpPortPollerClient.unregisterEvent( PollerFD::EVT_IN );
}

void
HTTPd::removeClientConnection( HTTPClientConnection* conn )
{
#ifdef DEBUG
   cout << "HTTPd::removeClientConnection()" << endl;
#endif
   
   assert( conn );
   assert( conn->getID() < static_cast<id_type>( m_clients.capacity() ) );
   assert( m_clients[conn->getID()] != NULL );

   --m_clientNum;

   MeasuringPollerClient* client = m_clients[conn->getID()];
   m_clients[conn->getID()] = NULL;

   // will call client->unregister too
   delete client;

   // whenever we get here the number of active clients
   // _will_ be less than the maximum number of allowed
   // clients
   // therefore reregister input event
   m_tcpPortPollerClient.registerEvent( PollerFD::EVT_IN );
}

void
HTTPd::processRequest( HTTPClientConnection* conn,
		       Request* request )
{
   assert( request );
   assert( conn );
   assert( conn->getID() < static_cast<id_type>( m_clients.capacity() ) );
   assert( m_clients[conn->getID()] != NULL );

   queueelem_type qe( conn, request );
   m_requestQueue.push( qe );
   
   MeasuringPollerClient* client = m_clients[conn->getID()];
   client->unregisterEvent( PollerFD::EVT_IN );
}

void
HTTPd::responsePrinted( HTTPClientConnection* conn,
			const Response& what )
{
   assert( conn );
   assert( conn->getID() < static_cast<id_type>( m_clients.capacity() ) );
   assert( m_clients[conn->getID()] != NULL );

   MeasuringPollerClient* client = m_clients[conn->getID()];

   client->unregisterEvent( PollerFD::EVT_OUT );
   client->registerEvent( PollerFD::EVT_IN );
}

void
HTTPd::run()
{
   while( !m_requestQueue.empty() )
   {
      queueelem_type elem    = m_requestQueue.pop();
      Request* const request = elem.second;
      
      /*
       * process request
       */
     Method* m = request->getMethod();
     assert( m );
     
     m->processRequest();
     
     const Response* resp = m->getResponse();
     assert( resp );
     
     
     /*
      * forward response
      */
     HTTPClientConnection* const from = elem.first;
     MeasuringPollerClient* client    = m_clients[from->getID()];
     
     client->registerEvent( PollerFD::EVT_OUT );
     // we do not need to unregister PollerFD::EVT_OUT as that was already done by processRequest
     from->writeResponse( resp );
   }  
}

HTTPd::id_type
HTTPd::getNextFreeID()
{
   for( id_type i = 0; i < m_clients.capacity(); ++i )
      if( m_clients[i] == NULL )
	 return i;

   assert( false );

   return 0;
}

}//EON httpd
