/************************************************************************
 *   IRC - Internet Relay Chat, src/match.c
 *   Copyright (C) 1990 Jarkko Oikarinen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Changes:
 * Thomas Helvey <tomh@inxpress.net> June 23, 1999
 * Const correctness changes
 * Cleanup of collapse and match
 * Moved static calls variable to match
 * Added asserts for null pointers
 */

#include "irc.hpp"
#include "hsirctraits.hpp"


/***************************************************************************
                           irc.cpp
                           -------------------
    copyright            : (C) 2001,2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: irc.cpp,v 1.16 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

using std::string;

IRCToken&
collapse( IRCToken& tok )
{ 
   char                             lastChar  = 0;
   IRCToken::data_type*             readIter  = tok.c_str();
   IRCToken::data_type*             writeIter = readIter;
   const IRCToken::data_type* const iterMax   = readIter + tok.length();
   int charsSkipped = 0;


   while( readIter < iterMax )
   {
      switch( *readIter )
      {
	 case '*':
	    if( lastChar != '*' )
	    {
	       *writeIter++ = *readIter++;
	       lastChar = '*';
	    }
	    else 
	    {
	       ++charsSkipped;
	       ++readIter;
	    }

	    break;
	    
	 case '?':
	    if( lastChar != '?' )
	    {
	       *writeIter++ = *readIter++;
	       lastChar = '?';
	    }
	    else
	    {
	       ++charsSkipped;
	       ++readIter;
	    }
	    break;

	 default:
	    lastChar = 0;
	    *writeIter++ = *readIter++;
      }
   }

   tok.erase( tok.length() - charsSkipped );
   
   return tok;
}

bool
match( const IRCToken& criteria, const IRCToken& target )
{
   const IRCToken::data_type*       critIter   = criteria.c_str();
   const IRCToken::data_type*       targetIter = target.c_str();
   const IRCToken::data_type* const critMax    = critIter + criteria.length();
   const IRCToken::data_type* const targetMax  = targetIter + target.length();

   while( targetIter < targetMax )
   {
      switch( *critIter )
      {
	 case '*':
	    if( ++critIter == critMax )
	       return true;
	    
	    while( true )
	    {
	       if( targetIter == targetMax )
		  return false;
	       if( *critIter == *targetIter++ )
	       {
		  ++critIter;
		  break;
	       }
	    }
	    break;

	 case '?':
	    if( ++critIter == critMax )
	       return targetMax - targetIter == 1;

	    ++targetIter;
	    break;

	 default:
	    if( *targetIter++ != *critIter++ )
	       return false;
      }
   }
   
   return true;
}

typedef bool (validator)(unsigned char);

int
validate( validator* valid, IRCToken& raw )
{
   const IRCToken::data_type*       readIter     = raw.c_str();
   IRCToken::data_type*             writeIter    = raw.c_str();
   const IRCToken::data_type* const readMax      = readIter + raw.length();
   int                              skippedChars = 0;
   
   while( readIter < readMax )
   {
      if( (valid)(*readIter) )
	 *writeIter++ = *readIter;
      else
	 ++skippedChars;
      ++readIter;
   }
   
   raw.erase( raw.length()-skippedChars );
   
   return skippedChars;
}

std::string::size_type
validate( validator* valid, string& raw )
{
   string::iterator             writeIter    = raw.begin();
   string::const_iterator       readIter     = raw.begin();
   const string::const_iterator readMax      = raw.end();
   string::size_type            skippedChars = static_cast<string::size_type>( 0 );

   while( readIter != readMax )
   {
      if( (*valid)(*readIter) )
	 *writeIter++ = *readIter;
      else
	 ++skippedChars;
	   
      ++readIter;
   }
   
   raw.erase( writeIter, raw.end() );

   return skippedChars;
}

IRCToken::size_type
toValidNickname( IRCToken& nick )
{
   return validate( &HSIRCTraits::isNickChar, nick );
}

string::size_type
toValidNickname( string& nick )
{
   return validate( &HSIRCTraits::isNickChar, nick );
}

IRCToken::size_type
toValidUsername( IRCToken& username )
{
   return validate( &HSIRCTraits::isUserChar, username );
}

string::size_type
toValidUsername( string& username )
{
   return validate( &HSIRCTraits::isUserChar, username );
}

IRCToken::size_type
toValidHostname( IRCToken& hostname )
{
   return validate( &HSIRCTraits::isHostChar, hostname );
}

string::size_type
toValidHostname( string& hostname )
{
   return validate( &HSIRCTraits::isHostChar, hostname );
}

IRCToken::size_type
toValidChanname( IRCToken& channame )
{
   if( !HSIRCTraits::isChanPrefix( channame[0] ) )
      return channame.length();

   return validate( &HSIRCTraits::isChanChar, channame );
}

string::size_type
toValidChanname( string& channame )
{
   if( !HSIRCTraits::isChanPrefix( channame[0] ) )
      return channame.length();

   return validate( &HSIRCTraits::isChanChar, channame );
}

IRCToken::size_type
toValidHSClientString( IRCToken& str )
{
  return validate( &HSIRCTraits::isHSClientChar, str );
}

string::size_type
toValidHSClientString( string& str )
{
  return validate( &HSIRCTraits::isHSClientChar, str );
}


std::ostream&
bold( std::ostream& o )
{
  o.put( HSIRCTraits::FLAG_BOLD );
  
  return o;
}

std::ostream&
underlined( std::ostream& o )
{
  o.put( HSIRCTraits::FLAG_UNDL );

  return o;
}
