/***************************************************************************
                           irctoken.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: irctoken.cpp,v 1.15 2002/05/26 20:21:27 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <string>
#include <iostream>
#include <exception>
extern "C" {
#include <sys/time.h>
#include <time.h>
}
#include "irctoken.hpp"
#include "irctokenizer.hpp"

using std::string;

IRCToken::IRCToken()
   : m_strbuf( NULL ),
     m_length( 0 ),
     m_authorative( false )
{
}

IRCToken::IRCToken( const char* msg, bool authorative )
  : m_strbuf( NULL ),
    m_length( 0 ),
    m_authorative( authorative )
{
  assign( (unsigned char*) msg, 
	  strlen( (unsigned char*)msg ), 
	  authorative );
}

IRCToken::IRCToken( const reference_type msg )
   : m_strbuf( NULL ),
     m_length( 0 ),
     m_authorative( false )
{
   assign( msg, strlen( msg ), true );
}

IRCToken::IRCToken( const reference_type msg,
		    const size_type len,
		    bool authorative )
   : m_strbuf( 0 ),
     m_length( 0 ),
     m_authorative( authorative )
{
   assign( msg, len, authorative );
}

IRCToken::IRCToken( const string& msg )
   : m_strbuf( 0 ),
     m_length( 0 ),
     m_authorative( true )
{
   assign( (reference_type)msg.c_str(), msg.length(), m_authorative );
}

IRCToken::IRCToken( const IRCToken& token, bool authorative )
   : m_strbuf( 0 ),
     m_length( 0 ),
     m_authorative( authorative )
{
   assign( token.m_strbuf, token.m_length, authorative );
}

const IRCToken&
IRCToken::operator=( const IRCToken& tok)
{
   if( this == &tok ) return *this;

   // create deep copies
   assign( tok.m_strbuf, tok.m_length, true );
   
   return *this;
}

const IRCToken&
IRCToken::operator=( const reference_type msg )
{
   cleanup();
   
   m_authorative = true;
   m_length      = IRCToken::strlen( msg );
   m_strbuf      = new data_type[m_length+1];

   fast_strcpy( m_strbuf, msg, m_length );
   // we must not terminate because every operation on this
   // string gets autodelimited by 'm_length'. 
   // but terminating + allocating space for the terminator
   // improves c_str() performance...
   m_strbuf[m_length] = '\0';

   return *this;
}

const IRCToken&
IRCToken::operator=( const string& str )
{
   assign( (reference_type)str.c_str(), str.length(), true );

   return *this;
}

const IRCToken&
IRCToken::operator=( const char* msg )
{
   return this->operator=( (unsigned char*)msg );
}

IRCToken::data_type
IRCToken::operator[]( size_type pos ) const
{
   assert( pos < m_length );
   
   return m_strbuf[pos];
}

const IRCToken&
IRCToken::assign( reference_type buf, 
		  size_type Length )
{
   cleanup();
   
   m_strbuf = buf;
   m_length = Length;
   
   return *this;
}

const IRCToken&
IRCToken::assign( const IRCToken& tok )
{
   if( &tok != this )
      assign( tok.m_strbuf, tok.m_length, tok.m_authorative );
   
   return *this;
}

const IRCToken&
IRCToken::assign( const reference_type buf, 
		  size_type Length,
		  bool authorative )
{
   cleanup(); // implicit m_authorative = false

   if( !authorative ) 
   {
      m_strbuf = buf;
      m_length = Length;      
   }
   else
   {
      // ok. reference counting would perform better here
      m_strbuf      = new data_type[Length+1];
      m_length      = Length;
      m_authorative = true;

      fast_strcpy( m_strbuf, buf, m_length );

      m_strbuf[Length] = '\0';
      // terminate
   }

   return *this;
}

void
IRCToken::cleanup()
{
   if( m_authorative )
   {
     if( m_strbuf )
	delete[] m_strbuf;

      m_length      = 0;
      m_authorative = false;
   }
}

IRCToken
IRCToken::substr( size_type begin, 
		  size_type len ) const
{
   if( len == IRCToken::npos )
   {
      len = m_length-begin;

   }

   return IRCToken( &m_strbuf[begin], len , m_authorative );
}

IRCToken
IRCToken::substr( size_type begin ) const
{
   assert( begin <= m_length );

   return IRCToken( &m_strbuf[begin], m_length-begin, m_authorative );
}


void
IRCToken::fast_strcpy( reference_type dest, 
		       reference_type src, 
		       size_type Length )
{
   // this performed best
   // (even better than asm version, memcpy, etc.)
   reference_type end = src+Length;
   while( src < end )
   {
      *dest++ = *src++;
   }
}

IRCToken::size_type
IRCToken::strlen( const unsigned char* str )
{
   assert( str != NULL );
   
   const unsigned char* strIter = str;

   size_type len = 0;
   for( ;*strIter != '\0'; ++len, ++strIter );
   
   return len;
}

const IRCToken& 
IRCToken::operator+=( const reference_type str )
{
   this->strcat( str, strlen( str ) );
   
   return *this;
}

const IRCToken&
IRCToken::operator+=( const IRCToken& tok )
{
   this->strcat( tok.m_strbuf, tok.m_length );
   
   return *this;
}

const IRCToken&
IRCToken::operator+=( const string& str )
{
   this->strcat( (reference_type)str.c_str(), str.length() );
   
   return *this;
}

void
IRCToken::strcat( const reference_type addChars,
		  size_type len )
{
   const size_type newLen = m_length + len;
   
   // allocate new memory ( we must be memory authorative )
   reference_type newStrBuf = new data_type[newLen];
   
   fast_strcpy( newStrBuf, m_strbuf, m_length );
   fast_strcpy( m_strbuf+m_length, addChars, len );
   
   // free authorative memory
   cleanup();

   m_strbuf = newStrBuf;
   m_length = newLen;
}

bool
IRCToken::operator==( const IRCToken& tok ) const
{
   if( tok.m_length != this->m_length ) return false;
   else if( &tok == this ) return true;
   else 
   {
      // strings have equally length and do not share location
      reference_type locIter = m_strbuf;
      reference_type argIter = tok.m_strbuf;
      const reference_type locIterMax = m_strbuf + m_length;

      while( locIter < locIterMax  )
      {
	 if( *locIter++ != *argIter++ ) return false;
      }

      return true;
   }
}

bool
IRCToken::operator==( const reference_type str ) const
{
   if( str == m_strbuf ) return true;
   else {
      reference_type strIter = const_cast<reference_type>( str );
      reference_type locIter = m_strbuf;
      reference_type locIterMax = locIter + m_length;

      while( locIter < locIterMax )
      {
	 if( *strIter == '\0' || *locIter++ != *strIter++ ) return false;
      }

      return *strIter == '\0';
   }
}

bool
IRCToken::operator==( const string& str ) const
{
   return this->operator==( IRCToken( (reference_type)str.c_str(), str.length(), false ) );
}

bool
IRCToken::operator==( const char* str ) const
{
  return this->operator==( (const reference_type)str );
}

IRCToken
IRCToken::operator+( const reference_type str )
{
   return this->operator+( IRCToken( str, strlen( str ), false ) );
}

IRCToken
IRCToken::operator+( const IRCToken& arg )
{
   IRCToken tok;

   const size_type newLen =  arg.m_length + m_length;

   tok.m_strbuf = new data_type[newLen];
   tok.fast_strcpy( tok.m_strbuf, m_strbuf, m_length );
   tok.fast_strcpy( tok.m_strbuf+arg.m_length, arg.m_strbuf, arg.m_length );

   tok.m_authorative = true;
   tok.m_length = newLen;

   return tok;   
}

IRCToken
IRCToken::operator+( const string& str )
{
   return this->operator+( IRCToken( (reference_type)str.c_str(), str.length(), false ) );
}

bool
IRCToken::operator<( const IRCToken& tok ) const
{
   if( length() > tok.length() )
      return false;
   else if( length() < tok.length() )
      return true;
   else 
      return strcmp( m_strbuf, tok.m_strbuf, length() ) < 0;
}

bool
IRCToken::operator<( const reference_type str ) const
{
   size_type strLen = strlen( str );

   if( length() > strLen )
      return false;
   else if( length() < strLen )
      return true;
   else 
      return strcmp( m_strbuf, str, length() ) < 0;
}

bool
IRCToken::operator<( const string& str ) const
{
   if( length() > str.length() )
      return false;
   else if( length() < str.length() )
      return true;
   else
      return strcmp( m_strbuf, (unsigned char*) str.c_str(), length() ) < 0;
}

bool
IRCToken::operator>( const IRCToken& tok ) const
{
   if( length() < tok.length() )
      return false;
   else if( length() > tok.length() )
      return true;
   else
      return strcmp( m_strbuf, 
		     tok.m_strbuf, 
		     length() ) > 0 ;
}

bool
IRCToken::operator>( const reference_type str ) const
{
   size_type strLen = strlen( str );

   if( length() < strLen )
      return false;
   else if( length() > strLen )
      return true;
   else
      return strcmp( m_strbuf,
		     str, 
		     length() ) > 0;
}

bool
IRCToken::operator>( const string& str ) const
{
   if( length() < str.length() )
      return false;
   else if( length() > str.length() )
      return true;
   else 
      return strcmp( m_strbuf, (unsigned char*) str.c_str(), length() ) > 0;
}

int
IRCToken::strcmp( const reference_type l, const reference_type r,
		  size_type len )
{
   const unsigned char* lIter = l;
   const unsigned char* rIter = r;
   int ret, n = len;
   
   while( (ret = toupper( *lIter ) - toupper( *rIter ) ) == 0 )
   {
      ++lIter;
      ++rIter;
      --n;
      
      if( n == 0 || (*lIter == '\0' && *rIter == '\0' ) )
	 return 0;
   }
   
   return ret;
}

const IRCToken&
IRCToken::erase( size_type from )
{

   if( from < m_length && from > 0 )
   {
      // ok we could also want to modify m_length only. But we want strings
      // that are like java.lang.String 

      reference_type newBuf = new data_type[from+1];
      fast_strcpy( newBuf, m_strbuf, from );
      newBuf[from] = '\0';

      cleanup();
      
      m_authorative = true;
      m_strbuf      = newBuf;
      m_length      = from;
   }
   return *this;
}

bool
IRCToken::equalsIgnoreCase( const IRCToken& rhs ) const
{
   if( this->length() == rhs.length() ) {
      IRCToken::reference_type lIter = this->c_str();
      IRCToken::reference_type rIter = rhs.c_str();
      IRCToken::reference_type lIterMax = lIter + this->length();

      while( lIter < lIterMax )
      {
	 if( tolower( *lIter++ ) != tolower( *rIter++ ) ) return false;
      }

      return true;
   }
   else return false;
}
