#include "method.hpp"
#include "request.hpp"
#include "hostserv.hpp"
#include "cgi.hpp"
#include "config.h"


namespace httpd
{

bool
IncompleteMethod::onProcessRequest()
{ 
  m_response = make_response( http::ResponseStatusline::RESP_BAD_REQUEST, 
			      http::HTTPVER_110, 
			      m_request ); 
  return true;
}

bool
UnimplementedMethod::onProcessRequest()
{
  m_response = make_response( http::ResponseStatusline::RESP_NOT_IMPLEMENTED, 
			      http::HTTPVER_110,
			      m_request ); 
  return true;
}

bool
HEADMethod::onProcessRequest()
{
   using namespace http;

   const RequestStartline& startline       = m_request->getRequestStartline();
   const uri::URL&         url             = m_request->getRequestStartline().getURL();
   const HSConfiguration&  hsConfiguration = m_request->getFrom()->getServer()->getApplication()->getConfiguration();
      

   if(   url.isIncomplete() || 
       ( startline.getProtocolVersion() == HTTPVER_110 && !m_request->hostSet() ) )
   {
#ifdef DEBUG
      cout << "GETMethod::processRequest(): warning: client send malformed uri" << endl;
#endif
      m_response = make_response( ResponseStatusline::RESP_BAD_REQUEST,
				  HTTPVER_110,
				  m_request );
      
      return false;
   }


   // We ignore the host setting. OK this is _NOT_ rfc compatible
   // but I do not intend to implement an 'apache' here ...
   
   HTTPString requestDocument = url.getDocument();
   if( requestDocument.length() == 1 && requestDocument[0] == '/' )
      // received default path
      requestDocument = hsConfiguration.getHTTPdDefaultDocument();

   const int directoryJumps = util::countDirectoryJumps( requestDocument );

   if( directoryJumps < 0 )
   {
      // bad guy. wants to break out of the httpd jail
#ifdef DEBUG
      cout << "GETMethod::processRequest(): trapped bad bad jail braker" << endl;
#endif
      m_response = make_response( ResponseStatusline::RESP_NOT_FOUND,
				  HTTPVER_110,
				  m_request );

      m_response->setConnectionType( HTTPCONN_CLOSE );

      return true;
   }


   HTTPString requestFile  =  hsConfiguration.getHTTPdDocumentRoot();
   requestFile            +=  requestDocument;

#ifdef DEBUG
   cout << "HEADMethod:: processRequest(): GET request for: " << requestFile << endl;
#endif

   util::FileStats stats( requestFile.c_str() );
   
   if( !stats.exists() || !stats.canRead() )
   {
      m_response = make_response( ResponseStatusline::RESP_NOT_FOUND,
				  HTTPVER_110,
				  m_request );

      m_response->setConnectionType( HTTPCONN_CLOSE );
      return true;
   }

   
   m_response = make_response( ResponseStatusline::RESP_OK,
			       HTTPVER_110,
			       m_request );


   /*
    * get response setting and apply it to the current Response object
    */
   bool defaultSettingUsed = false;
   const FileResponseSetting& responseSetting = 
     FileResponseSetting::getFileResponseSetting( requestFile,
						  startline.getProtocolVersion(),
						  defaultSettingUsed );

   responseSetting.applySettingsTo( m_response );

   addFileDocumentHeaders( requestFile,
			   stats,
			   responseSetting,
			   defaultSettingUsed );

   return onSuccessfullRequest( requestFile, 
				stats,
				responseSetting );
}

void
HEADMethod::addFileDocumentHeaders( const http::HTTPString& filename,
				    const util::FileStats& stats,
				    const FileResponseSetting& responseSetting,
				    const bool defaultSetting )
{
   using namespace http;

   // do not append any headers for clients with suspicious
   // http version
   if( m_request->getRequestStartline().getProtocolVersion() == HTTPVER_UNKNOWN )
   {
      m_response->setConnectionType( http::HTTPCONN_CLOSE );
      m_response->setResponseBodyEncoding( http::HTTPCONN_CLOSE );

      return;
   }

   addHTTP1xDocumentHeaders( filename,
			     stats,
			     responseSetting,
			     defaultSetting );
}

void
HEADMethod::addHTTP1xDocumentHeaders( const http::HTTPString& filename,
				      const util::FileStats& stats,
				      const FileResponseSetting& responseSetting,
				      const bool defaultSetting )
{
   using namespace http;

   // append Date header
   string           date;
   util::timeToString( date );
   m_response->appendMessageHeader( MessageHeader( GENERAL_DATE, HTTPString( date ) ) );

   // append Server header 
   m_response->appendMessageHeader( http::MessageHeader::Server );


   if( !responseSetting.isHSSScript() )
   {
      // append the 'Last-Modified' header
      string          modiTime;
      util::timeToString( modiTime, stats.getLastModificationTime() );
      m_response->appendMessageHeader( MessageHeader( ENT_LAST_MODIFIED, HTTPString( modiTime ) ) );
   }

   if( responseSetting.getBodyEncodingType() != HTTPCONN_CHUNKED )
   {
      // append Accept-Ranges header
      m_response->appendMessageHeader( MessageHeader::AcceptRanges );
      
      // append content length header
      if( !responseSetting.isHSSScript() )
      {
	 const HTTPString len       = util::typeToString( stats.getFileSize() );
	 m_response->appendMessageHeader( MessageHeader( ENT_CONTENT_LENGTH, len ) );
      }
   }
   else
      // append Transfer-Encoding header
      m_response->appendMessageHeader( MessageHeader::ChunkedTransferEncoding );
   
   // append Content-Type header
   if( !defaultSetting )
      m_response->appendMessageHeader( responseSetting.getContentTypeHeader() );

   // append Connection header
   m_response->appendMessageHeader( http::httpconnectionTypeToHeader( responseSetting.getConnectionType() ) );
}

bool
GETMethod::onSuccessfullRequest( const http::HTTPString& requestDocument,
				 const util::FileStats& stats,
				 const FileResponseSetting& responseSetting )
{
   using namespace http;

   if( responseSetting.isHSSScript() )
   {
      const HTTPd* const                 httpd = m_response->getRequest()->getFrom()->getServer();
      const HSIRCServerConnection* const conn  = httpd->getApplication()->getHSIRCServerConnection();
      
      m_responseBody = new DynamicFileResponseBody( &httpd->getApplication()->getConfiguration(),
						    conn->getHSIRCClients(),
						    conn->getHSClients(),
						    conn->getVHosts(),
						    requestDocument );
   }
   else
      m_responseBody = new FileResponseBody( requestDocument );
   
   m_response->setResponseBody( m_responseBody );

   return true;
}

bool
IllegalMethod::onProcessRequest()
{ 
  m_response = make_response( http::ResponseStatusline::RESP_BAD_REQUEST, 
			      http::HTTPVER_110, 
			      m_request ); 

  return true;
}

bool
NotImplementedMethod::onProcessRequest()
{ 
  m_response = make_response( http::ResponseStatusline::RESP_NOT_IMPLEMENTED,
			      http::HTTPVER_110,
			      m_request );
  return true;
}



POSTCGIScriptHandlerKit* POSTMethod::m_handlerKit = NULL;

POSTMethod::~POSTMethod()
{
#ifdef VERBOSE_MEMORY
   cout << "POSTMethod] destroyed" << endl;
#endif

   /*
    * m_script is authorative for the instance m_response points to.
    * therefore we do only have to free m_script
    */
   delete m_script;
}

void
POSTMethod::makeHandlerKit()
{
   if( m_handlerKit != NULL )
      // kit has been created already
      return;

   /*
    * get the last created application object.
    * as this function gets called whenever hostserv is already running 
    * we can be sure an application instance exists.
    */
   HostServ* hostserv = dynamic_cast< HostServ* >( Application::getLastApplication() );
   assert( hostserv );

   m_handlerKit = new POSTCGIScriptHandlerKit( hostserv );
}

bool
POSTMethod::onProcessRequest()
{
   const RequestBody* const genericBody = m_request->getRequestBody();

   if( genericBody == NULL )
   {
      m_response = make_response( http::ResponseStatusline::RESP_BAD_REQUEST, 
				  http::HTTPVER_110, 
				  m_request ); 
      return true;
   }


   const PlainTextPOSTRequestBody* const body = dynamic_cast<const PlainTextPOSTRequestBody*>( m_request->getRequestBody() );

   // currently we do not support request bodys not plaintext encoded
   assert( body );

   HostServ* hostserv = dynamic_cast< HostServ* >( Application::getLastApplication() );
   assert( hostserv );


   /*
    * note: m_script will be authorative for the given response
    */
   m_script = new SimpleCGIScript( m_request, *m_handlerKit );

   m_response = const_cast<Response*>( m_script->makeResponse( body->getDataMap() ) );

   return true;
}

}; //EON httpd
