#include "requestfilebuffer.hpp"
#include "config.h"

/***************************************************************************
                           requestfilebuffer.cpp
                           -------------------
    copyright            : (C) 2002 by Martin Scholl
    email                : ms@deridiot.net

    $Id: requestfilebuffer.cpp,v 1.11 2002/05/25 16:15:36 d3cline Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

namespace httpd
{

using namespace http;
using std::cout;
using std::endl;

RequestFileBuffer::RequestFileBuffer( File& f, len_type bufLen )
  : FileInBuf( f, bufLen )
{
}

RequestFileBuffer::return_type
RequestFileBuffer::read( RequestStartline& startLine )
{
   // rfc: Request-Line: Method SP Request-URI SP HTTP-Version CRLF
   HTTPString method;
   HTTPString requestUri;
   HTTPString protocolVersion;
   

   if( FileInBuf::isEmpty() || FileInBuf::readCompletely() )
	 return ERR_BUFFEREMPTY;

   
   data_reference       start = getReadPos();
   const data_reference max   = getReadMax();

   // skip leading whitespaces
   while( *start == HTTPTraits::WS )
   {
      if( ++start == max )
      {
	 FileInBuf::setReadPos( max );
	 return ERR_MALFORMED;
      }
   }
   
   // rfc says it is a good idea to skip a first empty line. 
   if( *start == HTTPTraits::CR || *start == HTTPTraits::LF )
   {
      // read further line
      while( *start == HTTPTraits::WS )
      {
	 if( ++start == max )
	 {
	    FileInBuf::setReadPos( max );
	    return ERR_MALFORMED;
	 }
      }
   }
   
   
   /*
    * let's do real processing...
    */

   // first: find newline / carriage return
   data_reference startCopy = start;
   do
   {
      if( *startCopy == HTTPTraits::LF )
	 break;
   } while( ++startCopy < max );
   
   if( startCopy == max )
   {
      FileInBuf::setReadPos( max );
      return ERR_MALFORMED;
   }
   // update read position
   setReadPos( startCopy + 1 );

   // we want startCopy to point the last character before newline + 1
   if( *(startCopy-1) == HTTPTraits::CR )
      // we have to include this dirty thing here
      // as the rfc urges one to be not very
      // CRLF pedantic
      --startCopy;

   const data_reference end = startCopy;

   // we do _NOT_ need to update read position after here 
   // as it was done before

   // read 'method' token
   const data_reference methodEnd = readToken( method, start, end );
   if( methodEnd == end || *methodEnd != HTTPTraits::WS )
      return ERR_MALFORMED;
   
   /*
    * read uri token
    */
   data_reference uriStart = methodEnd + 1;
   // skip whitespaces
   while( *uriStart == HTTPTraits::WS && ++uriStart < end );

   if( uriStart == end )
      // premature end
      return ERR_MALFORMED;

   data_reference uriEnd = uriStart + 1;

   // skip chars until end or whitespace
   while( *uriEnd != HTTPTraits::WS && ++uriEnd < end );

   requestUri.assign( uriStart, uriEnd - uriStart );
   
   if( uriEnd != end )
   {
      /*
       * read version token
       */
      
      // skip WS
      // we have to skip WS here as the 'READTEXT' function
      // won't skip ws as they can be part of a 'TEXT' object
      data_reference protVersionBegin = const_cast<data_reference>( uriEnd );
      while( *protVersionBegin == HTTPTraits::WS && protVersionBegin++ < end );
      
      if( uriEnd != end )
      {
	 // version token received
	 readText( protocolVersion, protVersionBegin, end );
	 
	 if( !protocolVersion.empty() )
	 {
	    startLine.update( method,
			      requestUri,
			      protocolVersion );
#ifdef DEBUG
	    cout << "RequestFileBuffer: read start-line: Method: " << method << endl
		 << "request-uri: " << requestUri << endl
		 << "prot-version: " << protocolVersion << endl;
#endif
	    
	    return READ_OK;
	 }
      }
   }


   // did not receive any http version token (-> HTTP/1.0 )   
   startLine.update( method, 
		     requestUri );
   
   return READ_OK;
}

RequestFileBuffer::return_type
RequestFileBuffer::read( MessageHeader& messageHeader )
{
   // rfc: message-header: field-name ':' [field-value]
   // field-name: token
   // field-value: TEXT or combination of TEXT and separators


   HTTPString           fieldName;
   FileInBuf::data_reference       readPos = getReadPos();
   const FileInBuf::data_reference readMax = getReadMax();   
   // points the character before CRLF
   FileInBuf::data_reference       readEnd = readPos;


   if( FileInBuf::isEmpty() || FileInBuf::readCompletely() )
	 return ERR_BUFFEREMPTY;


   // check whether we have received an empty line 
   if( *readPos == HTTPTraits::CR && readPos+1 < readMax && *(readPos+1) == HTTPTraits::LF )
   {
      setReadPos( readPos + 2 );
      return HEADER_TERMINATE;
   }
   if( *readPos == HTTPTraits::LF )
   {
      setReadPos( readPos+1 );
   }

   
   /*
    * seek readEnd to the first occurence of 'HTTPTraits::LF'
    * and update buffer pointers
    */
   do {
     if( *readEnd == HTTPTraits::LF )
       break;
   } while( ++readEnd < readMax );

   if( readEnd == readMax )
   {
      FileInBuf::setReadPos( readMax );
      return ERR_MALFORMED;
   }
   
   FileInBuf::setReadPos( readEnd + 1 );

   if( *( readEnd - 1) == HTTPTraits::CR )
      --readEnd;


   /*
    * we do not need to care about buffer pointers from here on
    */
   

   // read field-name token
   readPos = readToken( fieldName, readPos, readEnd );
   if( readPos == readEnd )
      return ERR_MALFORMED;

#ifdef DEBUG
      cout << "RequestFileBuffer::read(): MessageHeader fieldname = '" << fieldName << '\'' << endl;
#endif

   /*
    * read field-value
    */


   // skip whitespaces between field-name and ':'
   while( *readPos == HTTPTraits::WS && ++readPos < readEnd );
   if( readPos == readEnd || *readPos != ':' )
      return ERR_MALFORMED;


   // omit whitespaces between ':' and <field-value>
   while( ++readPos < readEnd && *readPos == HTTPTraits::WS );
   if( readPos == readEnd )
   {
      // found field-name only header
      messageHeader.assign( fieldName );

      return READ_OK;
   }


   /*
    * receive field-value
    */
   HTTPString fieldValue;
   
   readPos = readText( fieldValue, readPos, readEnd );
   if( fieldValue.empty() )
      // receive an field-name only messageheader
      messageHeader.assign( fieldName );
   else
   {
      messageHeader.assign( fieldName, fieldValue );
#ifdef DEBUG
      cout << "RequestFileBuffer::read(): MessageHeader fieldvalue = '" << fieldValue << '\'' << endl;
#endif

   }

   return READ_OK;
}


FileInBuf::data_reference
RequestFileBuffer::readToken( HTTPString& token,
			      data_reference start,
			      data_reference end )
{
   // skip leading whitespaces
   
   while( !HTTPTraits::isTokenChar( *start++ ) && start < end );
   if( start == end )
      return start;
   
   return readGeneric( token,
		       &HTTPTraits::isTokenChar,
		       --start,
		       end );
}

FileInBuf::data_reference
RequestFileBuffer::readText( HTTPString& text,
			     data_reference start,
			     data_reference end )
{
   return readGeneric( text,
		       &HTTPTraits::isTextChar,
		       start,
		       end );
}

FileInBuf::data_reference 
RequestFileBuffer::readGeneric( HTTPString& target, 
				genericChecker* checkFunc,
				data_reference start,
				data_reference end )
{
  while( start < end && (checkFunc)( *start ) )
     target += *start++;

  return start;
}

}// EON http
