#include "util.hpp"
#include "responsefilebuffer.hpp"
#include "config.h"

#define VERBOSE_TIMESTATS

#define STREAMBUF_UPDATE_PPTR

using std::cout;
using std::endl;

namespace httpd
{

ResponseFileBuffer::ResponseFileBuffer( File& f,
					len_type buflen,
					buffer_type type )
   : FileOutBuf( f, buflen ),
     m_bufferType( type ),
     
     // _FIRST_: make a copy of buffer's original base.
     // we need this variable when switching to 'close' transfer mode
     m_bufferBase( (data_reference)pbase() ),
     m_bufferLen( buflen ),
     m_reservedBytes( 0 ),
     m_chunkHeaderBuffer( NULL )
{
   setupBuffer();
}

ResponseFileBuffer::~ResponseFileBuffer()
{
   if( m_chunkHeaderBuffer ) 
      delete[] m_chunkHeaderBuffer; 
}
   
void
ResponseFileBuffer::setBodyTransportEncoding( buffer_type type )
{
   if( type != m_bufferType )
   {
      m_bufferType = type;
      setupBuffer();
   }
}

FileOutBuf::len_type
ResponseFileBuffer::flushBuffer()
{
   len_type retLen;
   if( m_bufferType == http::HTTPCONN_CHUNKED )
   {
      /*
       * when we get here points pbase() to the 
       * region where the actual data has been
       * written.
       * Means we have to move pbase() to the point,
       * where we have written the buffer len integer.
       * When  the data is written we will move the 
       * pointers back to original state.
       */
      char* const      oldPBase    = pbase();
      data_type* const writeStart  = addChunkHeader();
#ifdef STREAMBUF_UPDATE_PPTR
      data_type* const origPptrPos = (data_type*)pptr();
#endif

      /*
       * we encountered a very bad problem here
       * pptr() gets updated by this function call
       * (pptr()==pbase()). 
       */
      setp( (char*)writeStart, epptr() );

#ifdef STREAMBUF_UPDATE_PPTR
      // move pptr() back to the original position therefore
      pbump( origPptrPos - writeStart );
#endif
      
      retLen = FileOutBuf::flushBuffer();

      if( retLen != File::eof )
	 // we don't want to the caller to see
	 // the real amount of transmitted data.
	 // omit the connection encoding part therefore
	 retLen -= ( (unsigned char*)oldPBase ) - writeStart;

      // here we do not have to update pptr().
      // the whole buffer has been transmitted.
      setp( oldPBase, epptr() );
   }
   else
      retLen = FileOutBuf::flushBuffer();


   return retLen;
}

void
ResponseFileBuffer::setupBuffer()
{
   if( m_bufferType == http::HTTPCONN_CHUNKED )
   {
      /*
       * we have to reserve a little space at the beginning.
       * this space will able us to write the junk meta information
       */
      
      // count the number of reserved chars 
      len_type numberLen  = util::countDigits( m_bufferLen, 16 );
      m_chunkHeaderBuffer = new char[numberLen];


      /*
       * we have to reserve at least 5 bytes to make sure we
       * have got enough space for '0\r\n\r\n'
       */
      if( numberLen < 3 )
	 numberLen = 3;

      // we have to add 2 here because of '\r\n' (see below)
      m_reservedBytes = numberLen + 2;

      
      assert( m_bufferLen > m_reservedBytes );

      // update buffer pointers
      setp( (char*)( m_bufferBase + m_reservedBytes ), epptr() );

      /*
       * and finally write \r\n into buffer
       * we do it here so we do not have to do it
       * every time flushBuffer gets called
       */
      data_reference writeIter =
	 (data_reference)( m_bufferBase + numberLen );

      *writeIter   = '\r';
      *++writeIter = '\n';

#ifdef DEBUG
      cout << "ResponseFileBuffer::setupBuffer(): switched to chunked transfer encoding" << endl;
      cout << "ResponseFileBuffer::setupBuffer(): reserved bytes: " << m_reservedBytes << endl;
#endif
   }
   else
   {      
      // make the complete buffer available to the caller
      setp( (char*)m_bufferBase, epptr() );
      m_reservedBytes     = 0;
      if( m_chunkHeaderBuffer )
      {
	 delete[] m_chunkHeaderBuffer;
	 m_chunkHeaderBuffer = NULL;
      }
      
#ifdef DEBUG
      cout << "ResponseFileBuffer::setupBuffer(): switched to closed transfer encoding" << endl;
#endif
   }
}

FileOutBuf::data_reference
ResponseFileBuffer::addChunkHeader()
{
   // the value of m_resservedBytes must be at least 5
   // whenever this is not the case we have a problem :-)
   assert( m_reservedBytes > 3 );
   
   
   const int  usedLen     = pptr() - pbase();
   const bool lastChunk   = usedLen == 0;

   if( !lastChunk )
   {

      const int  lenDigitNum = util::numToHexString( m_chunkHeaderBuffer, usedLen );
      // we have to reserve space for at least '0\r\n\r\n' ( -> rfc 2616: 'last-chunk' )
      const int  unusedBytes = m_reservedBytes - lenDigitNum - 2;

      assert( unusedBytes >= 0 );

      /*
       * copy buffer's length in hex system into buffer
       */
      
      data_reference const writeStart = m_bufferBase + unusedBytes;
      data_reference       writeIter  = writeStart;
      data_reference const writeMax   = writeStart + lenDigitNum;
      char*                cacheIter  = m_chunkHeaderBuffer;
   
      // the actual junk header is in m_chunkHeaderBuffer. 
      // copy it into the transmit buffer
      while( writeIter < writeMax )
	 *writeIter++ = *cacheIter++;

      // note: we don't have to add '\r\n' into the buffer 
      // after the region where the bufferlen has been stored.
      // that was done by 'setupBuffer()'
      return writeStart;
   }

   /*
    * here: fill resserved buffer region
    * with 5 chars: '0\r\n\r\n'
    * the last '\r\n' substring is already part
    * of the buffer
    */

   data_reference const writeStart = m_bufferBase + m_reservedBytes - 3;
   data_reference       writeIter  = writeStart;

   *writeIter++ = '0';
   *writeIter++ = '\r';
   *writeIter++ = '\n';
//    *writeIter++ = '\r';
//    *writeIter++ = '\n';
//    *writeIter++ = '0';
//    *writeIter++ = '\r';
//    *writeIter++ = '\n';

   setp( (char*)writeStart, epptr() );

#ifdef STREAMBUF_UPDATE_PPTR
   pbump( writeIter - writeStart );
#endif

   return writeStart;
}

}//EON httpd
