#include "uri.hpp"

namespace uri
{

using http::HTTPString;
using std::string;

const HTTPString URI::PROT_DIV  = ":/";
const HTTPString URI::PORT_DIV  = ":";
const HTTPString URI::QUERY_DIV = "?";
const HTTPString URI::DOC_DIV   = "/";

URI::URI()
   : m_protocol(),
     m_host(),
     m_port(),
     m_document(),
     m_query(),
     m_portSet( false ),
     m_invalidURI( true )
{
}

URI::URI( const HTTPString& urlString )
   : m_protocol(),
     m_host(),
     m_port(),
     m_document(),
     m_query(),
     m_portSet( false ),
     m_invalidURI( true )
{
  parse( urlString );
}

URI::URI( const URI& uri )
   : m_protocol( uri.m_protocol ),
     m_host( uri.m_host ),
     m_port( uri.m_port ),
     m_document( uri.m_document ),
     m_query( uri.m_query ),
     m_portSet( uri.m_portSet ),
     m_invalidURI( true )
{
}

const URI& 
URI::operator=( const URI& uri )
{
  if( this != &uri )
  {
     m_protocol   = uri.m_protocol;
     m_host       = uri.m_host;
     m_port       = uri.m_port;
     m_document   = uri.m_document;
     m_query      = uri.m_query;
     m_portSet    = uri.m_portSet;
     m_invalidURI = uri.m_invalidURI;
  }
  return *this;
}

void
URI::parse( const HTTPString& urlString )
{
   /*
    * <'protocol'>://<host>[:<port>]/<document>[?<query>]
    */

   // _FIRST_: reset members ( and true m_invalidURI )
   reset();

   HTTPString::size_type documentStart = static_cast<HTTPString::size_type>( 0 );

   // extrapolate protocol
   const HTTPString::size_type protocolEnd = urlString.find( PROT_DIV );
   if( protocolEnd != HTTPString::npos )
   {
      if( protocolEnd == 0 )
	 // empty protocol: invalid
	 // returning here means m_invalidURI is still true
	 return;
      
      const HTTPString::size_type hostStart = urlString.find_first_not_of( (const unsigned char*)"/", 
									   protocolEnd + PROT_DIV.length() );
      documentStart                         = urlString.find( DOC_DIV,
							      hostStart );

         // .- no document defined       no host defined -.
	 // v                                             v
      if( documentStart == HTTPString::npos || hostStart == urlString.length() )
	 // returning here means m_invalidURI is still true
	 return;

      m_protocol = urlString.substr( 0, protocolEnd );

      HTTPString::size_type portStart = urlString.find( PORT_DIV, hostStart );

      if( portStart == HTTPString::npos )
      {
	 /*
	  * extrapolate hostname without trailing port
	  */
      	 const HTTPString::size_type hostLen = documentStart - hostStart;

	 if( hostLen )
	    m_host = urlString.substr( hostStart, documentStart - hostStart );
	 else
	    return;
      }
      else if( portStart < documentStart )
      {
	 /*
	  * extrapolate hostname with trailing port
	  */
	 const HTTPString::size_type hostLen = portStart - hostStart;

	 portStart                           += PORT_DIV.length();

	 const HTTPString::size_type portLen =  documentStart - portStart;

	 if( hostLen && portLen )
	 {
	    m_host                      = urlString.substr( hostStart, hostLen );
	    const HTTPString portString = urlString.substr( portStart, portLen );
	    m_port                      = portString; // calls HTTPString::operator unsigned int()
	    m_portSet                   = true;
	 }
	 else
	    return;
      }
   }

   /*
    * extrapolate document
    */
   const HTTPString::size_type queryStart = urlString.find( QUERY_DIV, documentStart );
   
   if( documentStart == urlString.length() )
      return;
   if( queryStart == HTTPString::npos )
      // found no '?' in document string: -> no query assigned
      m_document = urlString.substr( documentStart );
   else
   {
      m_document = urlString.substr( documentStart, queryStart - documentStart );
      m_query    = urlString.substr( queryStart + QUERY_DIV.length() );
   }
   
   m_invalidURI = false;
}

void
URI::reset()
{
   m_protocol.erase();
   m_host.erase();
   m_document.erase();
   m_query.erase();
   m_portSet  = false;
   m_invalidURI = true;
}

std::ostream& operator<<( std::ostream& out, const URI& uri )
{
   using namespace http;
   
   if( uri.isIncomplete() )
      out << "(uri: incomplete)";
   else
   {
      if( uri.hostSet() && uri.protocolSet() )
	 out << uri.getProtocol() << "://" << uri.getHost();

      out << uri.getDocument();
      
      if( uri.querySet() )
	 out << '?' << uri.getQuery();
   }
   return out;
}

} //EON uri
